package sos.ambulance_v2;

import java.security.InvalidParameterException;
import java.util.Arrays;

import sos.ambulance_v2.base.AmbulanceConstants.ATstates;
import sos.ambulance_v2.decision.AmbulanceDecision;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.VirtualCivilian;
import sos.base.move.Path;
import sos.base.move.types.StandardMove;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.base.worldGraph.WorldGraphEdge;
import sos.tools.GraphEdge;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class AmbulanceUtils {

	public static boolean isBurning(Human hm) {
		if (!hm.isPositionDefined())
			return false;
		if (!(hm.getPositionArea() instanceof Building))
			return false;
		Building b = (Building) hm.getPositionArea();
		if (b.isOnFire())
			return true;
		else
			return false;
	}

	public static ATstates convertStateIndexToState(int stateIndex) {
		switch (stateIndex) {
		case 0:
			return ATstates.MOVE_TO_TARGET;
		case 1:
			return ATstates.RESCUE;
		case 2:
			return ATstates.MOVE_TO_REFUGE;
		case 3:
			return ATstates.SEARCH;
		default:
			return null;
		}
	}

	/**
	 * using in refugeLess strategy
	 * 
	 * @param hm
	 * @return boolean
	 */
	public static boolean isHumaninFireBuilding(Human hm) {
		if (hm.getPositionArea() instanceof Building && isBuildingOnFire((Building) hm.getPositionArea()))
			return true;

		return false;
	}

	public static boolean isVirtualCivilianInFireBuilding(VirtualCivilian vc) {
		if (vc.getPosition() instanceof Building && isBuildingOnFire((Building) vc.getPosition()))
			return true;

		return false;
	}

	private static boolean isBuildingOnFire(Building building) {
		if (building.isEitherFieryOrBurnt())
			return true;
		return false;
	}

	public static boolean isValidToDecide(Human hm, SOSLoggerSystem log) {

		if (!hm.isPositionDefined()) {
			log.trace(hm + " Position has not Defined==> it is invalid to decide");
			return false;
		}
		if (!hm.isBuriednessDefined()) {
			log.trace(hm + " Buriedness is not Defined==> it is invalid to decide");
			return false;
		}
		if (!hm.isHPDefined()) {
			log.trace(hm + " HP is not Defined==> it is invalid to decide");
			return false;
		}
		if (hm.getHP() == 0) {
			log.trace(hm + " id dead:( ==> it is invalid to decide");
			return false;
		}
		if (hm.getPosition() instanceof AmbulanceTeam) {
			log.trace(hm + " is in Ambulance==> it is invalid to decide");
			return false;
		}
		if (hm.getPosition() instanceof Refuge) {
			log.trace(hm + " is in Refuge==> it is invalid to decide");
			return false;
		}
		
		if (hm.getRescueInfo().isIgnored() && hm.getRescueInfo().getIgnoredUntil() > hm.model().time()) { //sinash bug fixed, changed "<" to ">" in the comparison
			log.trace(hm + "is ignored till " + hm.getRescueInfo().getIgnoredUntil() + " because " + hm.getRescueInfo().getIgnoreReason() + " ==> it is invalid to decide");
			return false;
		}
		if (hm instanceof Civilian) {

			if ((hm.getAreaPosition() instanceof Road) && hm.model().refuges().isEmpty()) {
				log.trace(hm + " is in road and there is no refuge in map ==> it is invalid to decide");
				return false;
			}

			if (!(hm.getPosition() instanceof Building) && (hm.getDamage() == 0 || hm.getHP() == 10000)) {//todo should better
				log.trace(hm + " is not in Building and damage==0 or hp==10000 ==> it is invalid to decide");
				return false;
			}

			if (hm.getBuriedness() == 0 && hm.getDamage() == 0 && hm.getHP() == 10000) {//TODO if near fire??????
				log.trace(hm + "has no buriedness and damge and it is healthy==> it is invalid to decide");
				return false;
			}
			/*
			 * if (hm.getDamage() == 0 || hm.getHP() == 10000) {//TODO should solve!!!why???
			 * return false;
			 * }
			 */

		} else { // Human is an agent
			//			if (!(hm.getPosition() instanceof Building))
			//				return false;//Comment by Ali because some time an agent may have buriedness but it was in road!!!!!
			if (hm.getBuriedness() == 0) {
				log.trace(hm + "has no buriedness ==> it is invalid to decide");
				return false;
			}

		}

		if (!(hm.isReallyReachable(true))) {
			log.trace(hm + "is not reachable ==> it is invalid to decide");
			return false;
		}

		log.trace(hm + " is valid to decide");
		return true;
	}

	public static boolean isValidToDecide(VirtualCivilian virtualcivilian, AmbulanceTeam AT, SOSLoggerSystem log) {

		if(virtualcivilian.isIgnored()){
			log.trace(virtualcivilian + " is ignored");
			return false;
	}
		if (virtualcivilian.getDeathTime()+30 <= AT.model().time()){
			log.trace(virtualcivilian + " is dead");
			return false;
		}

		if (!virtualcivilian.isPositionDefined()) {
			log.trace(virtualcivilian + " Position has not Defined==> it is invalid to decide");
			return false;
		}
		if (virtualcivilian.getPosition() instanceof Refuge) {
			log.trace(virtualcivilian + " is in Refuge==> it is invalid to decide");
			return false;
		}

		if ((virtualcivilian.getPosition() instanceof Road) && AT.model().refuges().isEmpty()) {
			log.trace(virtualcivilian + " is in road and there is no refuge in map ==> it is invalid to decide");
			return false;
		}

		if (virtualcivilian.getBuridness() == 0) {
			log.trace(virtualcivilian + "has no buriedness ==> it is invalid to decide");
			return false;
		}

		if (!(virtualcivilian.isReallyReachable())) {
			log.trace(virtualcivilian + "is not reachable ==> it is invalid to decide");
			return false;
		}

		log.trace(virtualcivilian + " is valid to decide");
		return true;
	}


	/*
	 * it returns the expireTime for task assigning(The max sycle that we can start the task)
	 */
	public static int taskAssigningExpireTime(Human hm, int deathTime) {
		int expire = 0;
		expire += Math.ceil(hm.getBuriedness() / (float) hm.getRescueInfo().getATneedToBeRescued());
		expire += (hm.getRescueInfo().getATneedToBeRescued() - 1);
		expire++; //for loading
		expire += hm.getRescueInfo().getTimeToRefuge();
		expire += AmbulanceDecision.AVERAGE_MOVE_TO_TARGET;
		expire += getCommunicationDelay(hm);
		expire++;//for unloading
		return deathTime - expire;
	}

	public static int taskAssigningExpireTime(VirtualCivilian virtualCivilian, int deathTime) {
		int expire = 0;
		expire += Math.ceil(virtualCivilian.getBuridness() / (float) virtualCivilian.getATneedToBeRescued());
		expire += (virtualCivilian.getATneedToBeRescued() - 1);
		expire++; //for loading
		expire += virtualCivilian.getTimeToRefuge();
		expire += AmbulanceDecision.AVERAGE_MOVE_TO_TARGET;
		expire += 1; //getCommunicationDelay(AT);
		expire++;//for unloading
		return deathTime - expire;
	}


	public static int getCommunicationDelay(Human hm) {
		int cycle = 0;
		//			if (agent instanceof CenterAgent) {
		cycle += hm.model().sosAgent().messageSystem.getNormalMessageDelay() > 0 ? hm.model().sosAgent().messageSystem.getNormalMessageDelay() : 0;
		//			}
		return cycle;
	}



	public static Area getRoadNeighbour(Area positionArea) {
		if (positionArea instanceof Road)
			return positionArea;
		for (Area r : positionArea.getNeighbours()) {
			return getRoadNeighbour(r);
		}
		return null;
	}

	//-------------------------------------------------------------------------------------
	public static void updateATtarget(Human target, AmbulanceTeam agent, SOSIState state) {
		if (target == null || state == null)
			throw new InvalidParameterException("Target: " + target + " State: " + state);
		if (agent.getWork().getTarget() != null)
			agent.getWork().getTarget().getRescueInfo().removeAT(agent);

		agent.getWork().setTarget(target, state);
		((AmbulanceTeamAgent) agent.getAgent()).setUnderMissionTarget(target);
		//		target.getRescueInfo().addAT(agent);

	}

	public static boolean isReachableForAT(Human target, boolean check) {

		if (target.isReallyReachable(check))
			return true;
		if (target.getAgent().move.isReallyUnreachableXY(target.getAreaPosition(), target.getX(), target.getY()))
			return false;

		Path path = target.getAgent().move.getPathTo(Arrays.asList(target.getAreaPosition()), StandardMove.class);
		int cost = 0;

		for (GraphEdge ge : path.getEdges())
			if (ge instanceof WorldGraphEdge)
				switch (ge.getState()) {
				case FoggyOpen:
					cost++;
					break;
				case FoggyBlock:
					cost += 3;
					break;
				default:
					break;
				}
		if (cost > 20)
			return false;

		return true;

	}



}