package sos.ambulance_v2.decision;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.AmbulanceTeamAgent;
import sos.ambulance_v2.base.RescueInfo.IgnoreReason;
import sos.ambulance_v2.tools.GraphUsage;
import sos.base.SOSAgent;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.entities.VirtualCivilian;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.message.structure.MessageXmlConstant;
import sos.base.message.structure.blocks.MessageBlock;
import sos.base.move.Path;
import sos.base.move.types.StandardMove;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.base.worldGraph.WorldGraphEdge;
import sos.search_v2.tools.cluster.ClusterData;
import sos.tools.GraphEdge;

/**
 * @author Saina Ghadiany
 */

public class AmbUtile {
	static SOSLoggerSystem logUtile;
	static int lastInfoSent = 2;
	static boolean stateChanged = true;
	static public GraphUsage gu;
	static ArrayList<Pair<Human, Integer>> humansByTime;
	static ArrayList<Pair<Area, Integer>> refugesByTime;
	static boolean acknowledgeMsgSent = false;
	static float clusterScore;
	static int clusterBurriedness = 0;
	static int mapBurriedness = 0;
	static int numberOfHumans = 0;

	public AmbUtile() {

		humansByTime = new ArrayList<Pair<Human, Integer>>();
		refugesByTime = new ArrayList<Pair<Area, Integer>>();
	}

	public static boolean isValiadCivilianInMap(Human civ) {
		if (!civ.isAlive())
			return false;
		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined())
			return false;
		if (!isReallyReachable(civ, true))
			return false;
		if (civ.getBuriedness() == 0 && civ.getHP() == 10000)
			return false;
		if (civ.getPositionArea() instanceof Refuge)
			return false;
		return true;

	}

	// isLoadingCivilian \\
	public static boolean isLoadingCivilian(AmbulanceTeamAgent ambAgent) {
		for (Human civilian : ambAgent.model().civilians()) {
			if (ambAgent.me().getPositionPoint() == civilian.getPositionPoint())
				return true;
		}
		return false;
	}

	public static ArrayList<Pair<ClusterData, Float>> ClustersPriority(AmbulanceTeamAgent amb){
		 ArrayList<Pair<ClusterData, Float>> mapClusters = new ArrayList<Pair<ClusterData,Float>>();
		mapBurriedness = 0;
		for(Human civ : amb.model().civilians())
		{
			mapBurriedness += civ.getBuriedness();
				
		}	
		
		for(ClusterData cluster : amb.model().searchWorldModel.getAllClusters()){
			clusterBurriedness = 0;
			numberOfHumans = 0;
			for(Human civ : amb.model().civilians())
			{
				if(cluster.getClusterArea().contains(civ.getX(), civ.getY())){
					clusterBurriedness += civ.getBuriedness();
					numberOfHumans++;
				}
			}
			
			clusterScore = (numberOfHumans) / ((float)(clusterBurriedness/(mapBurriedness+1))+1);
			mapClusters.add(new Pair<ClusterData, Float>(cluster, clusterScore));

		}
		
		Collections.sort(mapClusters, new clusterComparator());	
		
		return mapClusters;
	}

	public static void sendInfoMsgIfRequire(int status, int busy, AmbulanceTeamAgent self) {
		//		if (self.time() - lastInfoSent > 7 || stateChanged) {
		sendInfoMsg(self.target.getID().getValue(), status, busy, self);
		self.log().info("@@@@@ message send @@@@@@");
		//		}
	}

	public static void sendTaskAckMsg(int type, AmbulanceTeamAgent self, Human target) {
		self.messageBlock = new MessageBlock(MessageXmlConstant.HEADER_AMBULANCE_TASK_ACK);
		self.messageBlock.addData(MessageXmlConstant.DATA_AMBULANCE_INDEX, self.me().getAmbIndex());
		self.messageBlock.addData(MessageXmlConstant.DATA_ID, target.getID().getValue());
		self.messageBlock.addData(MessageXmlConstant.DATA_ACK_TYPE, type);
		self.messageBlock.setResendOnNoise(false);
		self.messages.add(self.messageBlock);
		self.log().info("sending Task Ack Message" + target + "==>" + ((type == 0) ? "accepted" : (type == 1) ? "finished" : (type == 2) ? "rejected" : "unknown " + type));
		self.logToAmbDecision("sending Task Ack Message" + target + "==>" + ((type == 0) ? "accepted" : (type == 1) ? "finished" : (type == 2) ? "rejected" : "unknown " + type));
	}

	public static int getMyBusyCycles(Human target, Pair<? extends Area, Point2D> place, ATTask currentState, AmbulanceTeamAgent self) {
		self.log().log("[Trace] calculating my busy time for " + target);
		int time = 0;
		switch (currentState) {
		case SearchTask:
			// for msg delay
			time++;
			self.log().log("\t" + currentState + " current time=" + time);
			break;
		case MoveTask:
			if (target == null) {
				self.log().info("how target is null??");
				time++;
			} else {
				time += self.timeOfReachingMeToPlace(place);
				self.log().log("\t" + currentState + " time=" + time);
			}
		case RescueTask:
			if (target == null) {
				self.log().error("how target is null??");
				time++;
			} else {
				if (target.getBuriedness() != 0) {
					if (target.getRescueInfo().getNowWorkingOnMe().size() != 0) {
						time += target.getBuriedness() / target.getRescueInfo().getNowWorkingOnMe().size();
					} else {
						time += target.getBuriedness();
					}
					self.log().log("\t" + currentState + " time=" + time);
				}

				if (target instanceof Civilian) { // for load
					time++;
					self.log().log("\t" + currentState + "load time=" + time);
				}

			}

		case GoToRefugeTask:
			if (target == null) {
				self.log().error("how target is null??");
				time++;
			} else if (target instanceof Civilian) {
				boolean loader = AmITheLoader1(target.getRescueInfo().getNowWorkingOnMe(), self);
				if (loader) {
					time += target.getRescueInfo().getTimeToRefuge();
					// for unload
					time++;
				}

				self.log().log("\t" + currentState + " time=" + time);
			}

		}
		return time;
	}

	public static void sendRescueMsgBySay(Human target, AmbulanceTeamAgent agent) {
		boolean needHelp = false;
		if (target.getRescueInfo().getATneedToBeRescued() - target.getRescueInfo().getNowWorkingOnMe().size() > 0)
			needHelp = true;
		sendStatueMsgBySay(1, false, agent, agent.last_task);
	}

	public static boolean AmITheLoader1(Collection<Integer> ids, AmbulanceTeamAgent self) {
		int age = 0;
		for (Integer at : ids) {
			if (at < self.getID().getValue())
				age++;
		}
		if (age == 0)
			return true;
		return false;
	}

	// is valid to decide for center \\

	public static boolean isValiadToDecideForCenter(Human civ, int time, boolean isCenter) {
		if (civ == null)
			return false;
		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined())
			return false;
		if (!isCenter && !AmbUtile.isReallyReachable(civ, true))
			return false;
		if (civ.getBuriedness() == 0 && civ.getHP() == 10000)
			return false;
		if (civ.getHP() == 0)
			return false;
		if (civ.getPositionArea() instanceof Refuge)
			return false;
		if (civ.getPosition() instanceof AmbulanceTeam)
			return false;
		if (civ.getRescueInfo().getIgnoredUntil() > time)
			return false;
		if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
			return false;
		if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
			return false;
		return true;

	}

	// is valid to decide \\

	public static Area getRoadNeighbour(Area positionArea) {
		if (positionArea instanceof Road)
			return positionArea;
		for (Area r : positionArea.getNeighbours()) {
			return getRoadNeighbour(r);
		}
		return null;
	}
	static boolean hasEnoughAT;
	public static int numberOfAmbNeeded(int timeToRefuge, Human civ, int deathTime) {
		
		hasEnoughAT = false;
		int numberOfAmb = 0;
		for (int i = 1; !hasEnoughAT; i++) {
			if (timeToRefuge + 7 + civ.getBuriedness() / i < deathTime) {
				hasEnoughAT = true;
			}
			numberOfAmb = i;
		}

		return numberOfAmb;
	}

	public static int numberOfFreeAmbs(AmbulanceTeam amb) {
		int frees = 0;
		for (AmbulanceTeam freeAmb : amb.model().ambulanceTeams())
			if (freeAmb.getWork() == null && freeAmb.getWork().getTarget() == null)
				frees++;
		return frees;
	}

	
	public static int numberOfFreeAmbs(SOSAgent<? extends StandardEntity> amb) {
		int frees = 0;
		for (AmbulanceTeam freeAmb : amb.model().ambulanceTeams())
			if (freeAmb.getWork() == null && freeAmb.getWork().getTarget() == null)
				frees++;
		return frees;
	}
	
	public static boolean checkCLuster(AmbulanceTeamAgent amb) {
		int buildingSize = amb.getMyClusterData().getBuildings().size();
		int checkedBuildingsNumber = 0;
		
		for (Building build : amb.getMyClusterData().getBuildings()) {
			if (build.getChecked())
				checkedBuildingsNumber++;
		}

		if (checkedBuildingsNumber > buildingSize / 3)
			return true;
		else
			return false;

	}

	public static boolean validFreeAgent(AmbulanceTeam ambulance) {

		if (ambulance == null || ambulance.getBuriedness() != 0 || !ambulance.isAlive())
			return false;

		if (ambulance.getWork() != null && ambulance.getWork().getTarget() != null && (ambulance.getWork().getMyState() != ATTask.MoveTask))
			return false;

		return true;

	}

	public static boolean isValiadToDecide(Human civ, int time, boolean isCenter, AmbulanceTeam amb) {

		if (civ == null) {
			//	logUtile.log("civ was not valid null");
			return false;
		}

		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined() || !civ.isAlive()) {
			//	logUtile.log("civ was not valid damage or hp or pos");
			return false;
		}
		if (!isCenter && !isReallyReachable(civ, true)) {
			//	logUtile.log("civ was not valid not reachable" + civ.getID());
			return false;
		}

		if (civ.getBuriedness() == 0 && civ.getHP() == 10000) {
			return false;
		}
		if (civ.getAreaPosition() instanceof Refuge) {
			//logUtile.log("civ was not valid instanceof refuge" + civ.getID());
			return false;
		}

		if (civ.getHP() == 0) {
			//	logUtile.log("civ was not valid hp = 0" + civ.getID());
			return false;
		}

		if (!isCenter && civ.getRescueInfo().getIgnoredUntil() > time)
			return false;

		if (!isCenter && time - civ.updatedtime() < 3)
			return false;

		if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
			return false;

		if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
			return false;
		if(civ.getAreaPosition() instanceof Building && ((Building)civ.getAreaPosition()).getFieryness() > 0 && ((Building)civ.getAreaPosition()).getFieryness() < 4)
			return false;
		return true;
	}

	public static boolean isValiadVirtualToDecide(VirtualCivilian civil, int time, AmbulanceTeam amb) {

		if (civil == null)
			return false;

		if (!civil.isPositionDefined())
			return false;

		if (civil.isIgnored())
			return false;
		

		if (civil.getBuridness() == 0)
			return false;

		if (civil.getDeathTime() < time)
			return false;

		if (civil.getPosition() instanceof Refuge)
			return false;

		return true;
	}

	public static boolean isValiadToDecideLow(Human civ, int time, boolean isCenter, AmbulanceTeam amb) {

		if (civ == null) {
			//	logUtile.log("civ was not valid null");
			return false;
		}

		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
			//	logUtile.log("civ was not valid damage or hp or pos");
			return false;
		}
		if (!isCenter && !isReallyReachable(civ, true)) {
			//	logUtile.log("civ was not valid not reachable" + civ.getID());
			return false;
		}
		if (civ.getBuriedness() == 0 && civ.getHP() == 10000) {
			//	logUtile.log("civ was not valid buridness 0 and hp 1000" + civ.getID());
			return false;
		}
		if (civ.getAreaPosition() instanceof Refuge) {
			//logUtile.log("civ was not valid instanceof refuge" + civ.getID());
			return false;
		}

		if (civ.getHP() == 0) {
			//	logUtile.log("civ was not valid hp = 0" + civ.getID());
			return false;
		}

		if (civ.getRescueInfo().getIgnoredUntil() > time && civ.getRescueInfo().getIgnoreReason() == IgnoreReason.WillDie) {
			//	logUtile.log("civ was not valid ignored" + civ.getID());
			return false;
		}

		return true;
	}

	public static boolean isValiadToDecideForCheck(Human civ, int time, boolean isCenter, AmbulanceTeam amb) {

		if (civ == null) {
			//	logUtile.log("civ was not valid null");
			return false;
		}
		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
			//	logUtile.log("civ was not valid damage or hp or pos");
			return false;
		}

		if (!isCenter && !isReallyReachable(civ, true)) {
			//	logUtile.log("civ was not valid not reachable" + civ.getID());
			return false;
		}

		if (civ.getDamage() == 0) {
			//	logUtile.log("civ was not valid buridness 0 and hp 1000" + civ.getID());
			return false;
		}

		if (civ.getAreaPosition() instanceof Refuge) {
			//logUtile.log("civ was not valid instanceof refuge" + civ.getID());
			return false;
		}

		if (civ.getHP() == 0) {
			//	logUtile.log("civ was not valid hp = 0" + civ.getID());
			return false;
		}

		if (civ.getRescueInfo().getIgnoredUntil() > time) {
			//	logUtile.log("civ was not valid ignored" + civ.getID());
			return false;
		}

		if (civ.getAreaPosition() instanceof Building && ((Building) civ.getAreaPosition()).isEitherFieryOrBurnt() && civ.getBuriedness() > 30)
			return false;

		return true;
	}

	public static boolean isValiadToDecideUndermission(Human civ, int time, AmbulanceTeam amb) {

		if (civ == null) {
			return false;
		}

		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
			return false;
		}
		if (!isReallyReachable(civ, true)) {
			return false;
		}
		if (civ.getHP() == 0) {
			return false;
		}
		if (civ.getPosition() instanceof AmbulanceTeam) {
			if (civ.getPosition().getID().getValue() == amb.getPosition().getID().getValue())
				return true;
			else {
				return false;

			}
		}
		if (civ.getPositionArea() instanceof Refuge) {
			return false;
		}
		
		if (civ.getRescueInfo().getIgnoredUntil() > time) {
			return false;
		}

		if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
			return false;

		if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
			return false;
		return true;
	}

	// finish task not rescue that human again
	public static void finish(AmbulanceTeamAgent ambulance) {
		ambulance.log().info("FINISH");
		ambulance.setUnderMissionTarget(null);
		ambulance.target = null;
		ambulance.log().info("id= " + ambulance.getID() + " target =" + ambulance.target);
		ambulance.centerRecommendedTarget = null;
	}

	// send message \\

	public static boolean isReallyReachable(Human target, boolean check) {
		if (target.isReallyReachable(check))
			return true;
		if (target.getAgent().move.isReallyUnreachableXY(target.getAreaPosition(), target.getX(), target.getY()))
			return false;
		Path path = target.getAgent().move.getPathTo(Arrays.asList(target.getAreaPosition()), StandardMove.class);
		if (path == null)
			return false;
		int cost = 0;
		if (path.getEdges() == null)
			return false;
		for (GraphEdge g : path.getEdges())
		{
			if (g instanceof WorldGraphEdge)
				switch (g.getState()) {
				case FoggyOpen:
					cost++;
					break;
				case FoggyBlock:
					cost += 3;
					break;

				default:
					break;
				}
		}

		if (cost < 20)
			return true;
		else
			return false;

	}

	public static boolean isReallyReachableBuilding(Area target, boolean check) {
		if (target.isReallyReachable(check))
			return true;
		if (target.getAgent().move.isReallyUnreachableXY(target.getAreaPosition(), target.getX(), target.getY()))
			return false;
		Path path = target.getAgent().move.getPathTo(Arrays.asList(target.getAreaPosition()), StandardMove.class);

		int cost = 0;
		for (GraphEdge g : path.getEdges())
		{
			if (g instanceof WorldGraphEdge)
				switch (g.getState()) {
				case FoggyOpen:
					cost++;
					break;
				case FoggyBlock:
					cost += 3;
					break;

				default:
					break;
				}
		}

		if (cost < 20)
			return true;
		else
			return false;

	}

	public static void sendMessage(int type, AmbulanceTeamAgent ambulance, Human target) {

		if (target == null)
			return;
		ambulance.messageBlock = new MessageBlock(MessageXmlConstant.HEADER_AMBULANCE_TASK_ACK);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AMBULANCE_INDEX, ambulance.me().getAmbIndex());
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_ID, target.getID().getValue());
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_ACK_TYPE, type);
		ambulance.messageBlock.setResendOnNoise(false);
		ambulance.messages.add(ambulance.messageBlock);
		ambulance.log().info("sending Task Ack Message" + target + "==>" + ((type == 0) ? "accepted" : (type == 1) ? "finished" : (type == 2) ? "rejected" : "unknown " + type));
		ambulance.logToAmbDecision("sending Task Ack Message" + target + "==>" + ((type == 0) ? "accepted" : (type == 1) ? "finished" : (type == 2) ? "rejected" : "unknown " + type));
	}

	public static void sendStatueMsgBySay(int state, boolean needHelp, AmbulanceTeamAgent ambulance, ATTask lastTask) {
		if ((ambulance.messageSystem.type != Type.NoComunication) || (ambulance.messageSystem.type != Type.LowComunication))
			return;
		int mlposIndex;
		int id;

		int help = needHelp ? 1 : 0;
		if (lastTask == ATTask.SearchTask) { //search mode
			id = 0;
		} else {
			id = ambulance.target.getID().getValue();
		}
		mlposIndex = ambulance.me().getPositionArea().getAreaIndex();
		ambulance.messageBlock = new MessageBlock(MessageXmlConstant.HEADER_AMBULANCE_STATUS);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AMBULANCE_INDEX, ambulance.me().getAmbIndex());
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AT_STATE, state);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_ID, id);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AREA_INDEX, mlposIndex);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_NEED_HELP, help);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_TIME, ambulance.time());
		ambulance.messageBlock.setResendOnNoise(false);
		ambulance.sayMessages.add(ambulance.messageBlock);

	}

	public static void sendInfoMsg(int id, int state, int finishTime, AmbulanceTeamAgent ambulance) {

		ambulance.messageBlock = new MessageBlock(MessageXmlConstant.HEADER_AMBULANCE_INFO);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AMBULANCE_INDEX, ambulance.me().getAmbIndex());
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_ID, id);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_AT_STATE, state);
		ambulance.messageBlock.addData(MessageXmlConstant.DATA_FINISH_TIME, ambulance.time() + finishTime);
		ambulance.messageBlock.setResendOnNoise(false);
		ambulance.messages.add(ambulance.messageBlock);
		ambulance.sayMessages.add(ambulance.messageBlock);

		lastInfoSent = ambulance.time();
		stateChanged = false;
		ambulance.log().info("sending info Message==>target=" + ambulance.model().getEntity(id) + " state=" + state + " ft:" + finishTime);
		ambulance.logToAmbDecision("sending info Message==>target=" + ambulance.model().getEntity(id) + " state=" + state + " ft:" + finishTime);

	}

	// refuges by time\\

	public static ArrayList<Pair<Area, Integer>> refugesAndTimes(AmbulanceInformationModel infoModel) {

		refugesByTime.clear();
		if (infoModel == null)
			logUtile.error("null infomodel");

		if (infoModel.getModel().refuges().size() > 0)
		{

			for (Area refuge : infoModel.getModel().refuges()) {
				if (!(infoModel.getAgent().move.isReachable(infoModel.getATEntity()
						.getAreaPosition(), refuge)))
					continue;
				sos.base.move.Path path = infoModel.getAgent().move.getPathFromTo(
						infoModel.getATEntity().getAreaPosition(), refuge,
						StandardMove.class);
				int time = infoModel.getAgent().move.getMovingTime(path);
				refugesByTime.add(new Pair<Area, Integer>(refuge, time));
			}

			Collections.sort(refugesByTime, new RefugeComparator());
			return refugesByTime;
		}
		
		
		return null;
	}

	
	// ////////////////////////////////

	// humans by time\\
	public static ArrayList<Pair<Human, Integer>> humansAndTimes(AmbulanceInformationModel infoModel) {
		gu = new GraphUsage(infoModel.getAgent());

		humansByTime.clear();

		for (Human human : infoModel.getModel().humans()) {
			if (!human.isReallyReachable(true))
				continue;
			int time = gu.getFoolMoveTimeFromTo(infoModel.getATEntity().getAreaPosition(), human.getAreaPosition());

			humansByTime.add(new Pair<Human, Integer>(human, time));
		}

		Collections.sort(humansByTime, new humanComparator());
		return humansByTime;
	}
}
