package sos.ambulance_v2.decision;

import java.util.ArrayList;
import java.util.Collections;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.ambulance_v2.AmbulanceUtils;
import sos.ambulance_v2.tools.GraphUsage;
import sos.ambulance_v2.tools.MultiDitinctSourceCostInMM;
import sos.base.SOSAgent;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.move.MoveConstants;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.base.util.sosLogger.SOSLoggerSystem.OutputType;


public class AmbulanceDecision {

	private SOSAgent<? extends StandardEntity> agent = null;
	public GraphUsage gu;
	public final SOSLoggerSystem dclog;

	/**************** center assigning map ******************/
	/******************* needed lists ***********************/
	ArrayList<AmbulanceTeam> ambulances;
	ArrayList<Human> targets;


	int lastTimeUpdated = 0;
	int lastCycleIwasHere = 0;
	int lastCycleISetNumberOfATs = 0;

	public static int AVERAGE_MOVE_TO_TARGET = 4;
	public boolean hugeMap = false;
	public SOSLoggerSystem humanUpdateLog;
	public MultiDitinctSourceCostInMM costTable;

	public AmbulanceDecision(SOSAgent<? extends StandardEntity> agent) {
		this.agent = agent;
		AVERAGE_MOVE_TO_TARGET = (agent.model().roads().size() < 1200 ? 6 : (agent.model().roads().size() < 2000 ? 9 : 12));
		dclog = new SOSLoggerSystem(agent.me(), "Agent/AmbulanceDecision", true, OutputType.File);
		dclog.setFullLoggingLevel();
		agent.sosLogger.addToAllLogType(dclog);

		humanUpdateLog = new SOSLoggerSystem(agent.me(), "Agent/ATHumanUpdate", true, OutputType.File, true);
		agent.sosLogger.addToAllLogType(humanUpdateLog);

		gu = new GraphUsage(agent);
		if (agent.getMapInfo().isBigMap()) {
			hugeMap = true;
		}
		dclog.logln("AVERAGE_MOVE_TO_TARGET=" + AVERAGE_MOVE_TO_TARGET + "  hugeMap=" + hugeMap);
	}

	/**
	 * @r@mik updating death and critical and average times and time to transport to nearest open refuge
	 */
	/***************************************************************************************/
	/******************************** updateHumansInfo **************************************/

	public void updateHumansInfo() {
		dclog.info("************* updateHumansInfo *******************");
		dclog.info("IN UPDATE_HUMANS_INFO .....time=" + agent.time() + " to more information take a look at human update log");
		if (lastTimeUpdated == agent.time())
			return;
		ArrayList<Area> srcs = new ArrayList<Area>();

		if (agent.getCenterActivities().isEmpty())
			srcs.add(agent.me().getAreaPosition());
		else {
			for (AmbulanceTeam at : agent.model().ambulanceTeams())
				srcs.add(at.getAreaPosition());
		}

		for (Refuge ref : agent.model().refuges()) {
			srcs.add(ref);
		}

		costTable = new MultiDitinctSourceCostInMM(agent.model(), srcs);
		lastTimeUpdated = agent.time();

		for (Human human : agent.model().humans()) {
			boolean result = updateHuman(human);
			if (!result)
				continue;

		}
		dclog.info("*******************************************");
	}

	/***************************************************************************************/

	public boolean updateHuman(Human hm) {
		humanUpdateLog.log(hm);
		//initial checks
		if (hm == null) {
			humanUpdateLog.logln(" --> NULL Human!");
			return false;
		}
		if (!hm.isBuriednessDefined()) {
			humanUpdateLog.logln(" --> Unknown State!");
			return false;
		}
		if (!(hm instanceof Civilian) && hm.getBuriedness() == 0) {
			humanUpdateLog.logln(" --> Not burried Agent!");
			return false;
		}
		if (!hm.isPositionDefined()) {
			humanUpdateLog.logln(" --> No position!");
			return false;
		}

		if (hm.getPosition() instanceof Refuge) {
			humanUpdateLog.logln(" --> in refuge!");
			return false;
		}
		if (hm.getPosition() instanceof AmbulanceTeam) {
			humanUpdateLog.logln(" --> in AT!");
			return false;
		}
		if (hm instanceof Civilian && hm.getPosition() instanceof Road && hm.getDamage() == 0 && hm.getHP() == 10000) {// chera hp==10000?
			humanUpdateLog.logln(" --> in road Healthy Civilian");
			return false;
		}


		if (agent.model().refuges().isEmpty()) {
			hm.getRescueInfo().setTimeToRefuge(1);
			hm.getRescueInfo().setBestRefuge(null);
		} else {
			calculateRefugeInformation(hm);
		}

		return true;
	}

	/***********************************************************************************************/
	

	



	
	/************************* calculateRefugeInformation ***********************************/
	public void calculateRefugeInformation(Human human) {
		if (!human.isPositionDefined() || human.getPosition() instanceof Refuge)
			return;
		if (!(human instanceof Civilian))
			return;
		if (agent.model().refuges().isEmpty() || human.getRescueInfo().getRefugeCalculatedTime() == agent.time())
			return;
		if (costTable == null) {//if is from hear
			human.getRescueInfo().setTimeToRefuge(AVERAGE_MOVE_TO_TARGET);
			return;
		}
		long minCost = Long.MAX_VALUE;
		Refuge best = null;
		for (Refuge refuge : agent.model().refuges()) {
			long cost = costTable.getCostFromTo(refuge, human.getAreaPosition());
			if (cost < minCost) {
				best = refuge;
				minCost = cost;
			}
		}

		if (best != null && minCost < MoveConstants.UNREACHABLE_COST_FOR_GRAPH_WEIGTHING) {
			//TODO else nabayad null mikard?
			human.getRescueInfo().setBestRefuge(best);
			human.getRescueInfo().setTimeToRefuge(gu.getFoolMoveTime(minCost));
		}

		human.getRescueInfo().setRefugeCalculatedTime(agent.time());

	}

	/***********************************************************************************************/
	/****************************** calculateRefugeInformation_old *********************************/
	/**
	 * @param hm
	 * @r@mik update nearest reachable refuge of Human and time to reaching that refuge
	 */
	public void calculateRefugeInformation_old(Human hm) {
		if (!hm.isPositionDefined() || hm.getPosition() instanceof Refuge)
			return;
		if (!(hm instanceof Civilian))
			return;
		if (agent.model().refuges().isEmpty() || hm.getRescueInfo().getRefugeCalculatedTime() == agent.time())
			return;

		ArrayList<Pair<? extends Area, Point2D>> Positions = new ArrayList<Pair<? extends Area, Point2D>>(agent.model().ambulanceTeams().size());
		for (Refuge rf : agent.model().refuges())
			Positions.add(new Pair<Area, Point2D>(rf, new Point2D(rf.getX(), rf.getY())));
	}
	/***********//**********getBestAmbulanceForr************//****************/
	//ezafe shode...//
	public AmbulanceTeam getBestAmbulanceForr(Human first) {
		dclog.info("//getBestAmbulance//");
		ArrayList<AmbulanceTeam> ATs = new ArrayList<AmbulanceTeam>();
		ATs.addAll(ambulances);
		Collections.sort(ATs, new clusterComparitor(first));
		for (AmbulanceTeam AT : ATs) {
			if (AT.getWork() != null && AT.getWork().getTarget() != null && AmbulanceUtils.isReachableForAT(AT, true)) {
				dclog.info("skipped from AT =" + AT);
				continue;
			}
			return AT;
		}
		return null;
	}
	/***********************************************************************************************/
	/************************************** movingTime *********************************************/
	public int[] movingTime(Pair<? extends Area, Point2D> from, ArrayList<Pair<? extends Area, Point2D>> targets) {
		int[] times = new int[targets.size()];
		if (hugeMap) {
			for (int i = 0; i < times.length; i++) {
				times[i] = gu.getFoolMoveTimeFromTo(from.first(), targets.get(i).first());
			}
			return times;
		}
		long[] lenghts = agent.move.getMMLenToTargets_notImportantPoint(from, targets);
		for (int i = 0; i < times.length; i++) {
			times[i] = (lenghts[i] >= MoveConstants.UNREACHABLE_COST_FOR_GRAPH_WEIGTHING) ? Integer.MAX_VALUE / 2 : gu.getFoolMoveTime(lenghts[i]);
		}
		return times;
	}

	//private AmbulanceTeam getBestAmbulanceFor(Human first) {
		//dclog.info("/////getBestAmbulanceFor////");
		//ArrayList<AmbulanceTeam> ATs = new ArrayList<AmbulanceTeam>();
		//ATs.addAll(ambulances);
		//Collections.sort(ATs, new clusterComparitor(first));
		//for (AmbulanceTeam AT : ATs) {

			//if (AT.getWork() != null && AT.getWork().getTarget() != null && AmbulanceUtils.isReachableForAT(AT, true)) {
				//dclog.info("skipped from AT =" + AT);
				//continue;
		//	}

			//return AT;
		//}

		//return null;//
	//}


	}









	
	





