package sos.ambulance_v2.decision.controller;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import rescuecore2.misc.Pair;
import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.base.AmbulanceConstants.ATstates;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.tools.GraphUsage;
import sos.ambulance_v2.tools.MultiDitinctSourceCostInMM;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.SOSAgent;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Civilian;
import sos.base.entities.FireBrigade;
import sos.base.entities.Human;
import sos.base.entities.PoliceForce;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageXmlConstant;
import sos.base.message.structure.blocks.MessageBlock;
import sos.base.move.MoveConstants;
import sos.base.move.types.StandardMove;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.base.util.sosLogger.SOSLoggerSystem.OutputType;
import sos.search_v2.tools.cluster.ClusterData;

/**
 * @author Saina Ghadiany
 */

public class CenterDecision {

	public GraphUsage gu;
	public SOSLoggerSystem centerUpdate;
	public SOSAgent<? extends StandardEntity> centerAmb;
	int AverageTimeToTarget;
	public ArrayList<AmbulanceTeam> ambAgents = new ArrayList<AmbulanceTeam>();
	ArrayList<Refuge> refuges = new ArrayList<Refuge>();
	public ArrayList<Human> civilians = new ArrayList<Human>();
	public ArrayList<Pair<AmbulanceTeam, Integer>> myAmbs;
	public ArrayList<Pair<AmbulanceTeam, Integer>> myBestAmbs;
	public ArrayList<Pair<CiviliansAndMeasurments, Integer>> CiviliansAndMeasur;
	public ArrayList<AmbulancesAndMeasurments> AmbulancesAndMeasur;
	ArrayList<AmbulancesAndMeasurments> AmbulancesAndMeasurUsedForCal;
	private ArrayList<Pair<Human, Integer>> civilianCosts;
	public MultiDitinctSourceCostInMM costTable;
	ArrayList<Human> ALL;
	ArrayList<AmbulanceTeam> ambulances;
	public static int AVERAGE_MOVE_TO_TARGET = 4;
	ArrayList<AmbulancesAndMeasurments> remove;
	ArrayList<Human> removeciv;
	ArrayList<Pair<ClusterData, Float>> mapClusters = new ArrayList<Pair<ClusterData, Float>>();
	int numberOfAmb;
	int moveDistance;
	int timeToRefuge;
	ArrayList<Human> noAmbCivilian;
	public int IWillBeRes;
	int lastTimeUpdated = 0;
	int mapBurriedness = 0;
	int clusterBurriedness = 0;
	int numberOfHumans = 0;
	float clusterScore = 0;
	int civClusterIndex = 0;

	public CenterDecision(SOSAgent<? extends StandardEntity> agent) {

		gu = new GraphUsage(agent);
		this.centerAmb = agent;

		centerUpdate = new SOSLoggerSystem(agent.me(), "Agent/SainaCenter",
				true, OutputType.File, true);
		agent.sosLogger.addToAllLogType(centerUpdate);
		AverageTimeToTarget = (centerAmb.model().buildings().size() > 1600 ? 20
				: (centerAmb.model().buildings().size() > 1300 ? 15 : 12));
		civilians = new ArrayList<Human>();
		ambAgents = new ArrayList<AmbulanceTeam>();
		ALL = new ArrayList<Human>();
		ambulances = new ArrayList<AmbulanceTeam>();
		AmbulancesAndMeasur = new ArrayList<AmbulancesAndMeasurments>();
		AmbulancesAndMeasurUsedForCal = new ArrayList<AmbulancesAndMeasurments>();
		refuges = new ArrayList<Refuge>();
		myAmbs = new ArrayList<Pair<AmbulanceTeam, Integer>>();
		myBestAmbs = new ArrayList<Pair<AmbulanceTeam, Integer>>();
		remove = new ArrayList<AmbulancesAndMeasurments>();
		removeciv = new ArrayList<Human>();
		civilianCosts = new ArrayList<Pair<Human, Integer>>();
		CiviliansAndMeasur = new ArrayList<Pair<CiviliansAndMeasurments, Integer>>();

	}

	public ArrayList<Pair<Human, Integer>> getcivilianCosts() {
		return civilianCosts;
	}

	public void setcivilianCosts(ArrayList<Pair<Human, Integer>> civilianCosts) {
		this.civilianCosts = civilianCosts;
	}

	public void findValidAmbulancesAndCivilians(boolean isCenter) {
		centerUpdate.logln("|| find Valid amb and civs ||");

		ambAgents.clear();
		civilians.clear();
		int numberOfAmb = 0;

		for (AmbulanceTeam ambulance : centerAmb.model().ambulanceTeams()) {
			if (ambulance == null)
				continue;
			if (ambulance.getBuriedness() != 0)
				continue;
			if (!ambulance.isAlive())
				continue;
			if (ambulance.getWork() != null && ambulance.getWork().getTarget() != null && (ambulance.getWork().getMyState() != ATTask.MoveTask))
				continue;
			ambAgents.add(ambulance);
			centerUpdate.logln("in findvalid adding amb" + ambulance.getID());
		}

		ALL.clear();

		ALL.addAll(centerAmb.model().civilians());
		if (centerAmb.model().time() < 90 || (centerAmb.model().getLastAfterShockTime() > 1 && centerAmb.model().time() < 210))
			ALL.addAll(centerAmb.model().agents());
		centerUpdate.logln("all size: " + ALL.size());

		for (Human civ : ALL) {
			ambulances.clear();

			if (civ == null) {
				centerUpdate.logln("civ was not valid null");
				continue;
			}
			if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
				centerUpdate.logln("civ was not valid damage or hp or pos " + civ.getID());
				continue;
			}
			if (!isCenter && !AmbUtile.isReallyReachable(civ, true)) {
				centerUpdate.logln("civ was not valid not reachable" + civ.getID());
				continue;
			}
		
			if (civ.getDamage() == 0) {
				centerUpdate.logln("civ was not valid damage" + civ.getID());

				continue;
			}
			if (civ.getAreaPosition() instanceof Refuge) {
				centerUpdate.logln("civ was not valid instanceof refuge" + civ.getID());
				continue;
			}
			

			if (civ.getHP() == 0) {
				centerUpdate.logln("civ was not valid hp = 0" + civ.getID());
				continue;
			}

			if (civ.getRescueInfo().getIgnoredUntil() > centerAmb.model().time()) {
				centerUpdate.logln("civ was not valid ignored" + civ.getID());
				continue;
			}
			if (!isCenter && civ.updatedtime() > centerAmb.model().time() - 3) {
				centerUpdate.logln("updated time" + civ.getID());
				continue;
			}
			if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
				continue;

			if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
				continue;


			if (civ.getRescueInfo().getDeathTime() > 330)
				continue;

			for (AmbulanceTeam amb : centerAmb.model().ambulanceTeams()) {
				if (amb.getWork() == null || amb.getWork().getTarget() == null) {
					continue;
				}
				else if (amb.getWork().getTarget().getID().getValue() == civ.getID().getValue() && amb.getWork().getCurrentState() == ATstates.RESCUE) {
					{

						centerUpdate.logln("amb" + amb.getID() + "is woking on " + civ.getID());
						ambulances.add(amb);
					}
				}
			}

			int timeToRefuge;
			if (centerAmb.model().refuges().size() == 0)
				timeToRefuge = 0;
			else
				timeToRefuge = civ.getRescueInfo().getTimeToRefuge();

			int deathTime = civ.getRescueInfo().getDeathTime();

			boolean enough = false;
			for (int i = 1; !enough; i++) {
				if (timeToRefuge + 7 + civ.getBuriedness() / i < deathTime) {
					enough = true;
					numberOfAmb = i;
				}
			}

			centerUpdate.logln("$$$ number : " + numberOfAmb);

			//			if ((deathTime - centerAmb.time()) > 15 + 2 + timeToRefuge + civ.getBuriedness()) {
			//				numberOfAmb = 1;
			//			}
			//
			//			else if ((deathTime - centerAmb.time()) > (15 + 2 + timeToRefuge) / 2) {
			//				numberOfAmb = 2;
			//			}
			//
			//			else
			//				numberOfAmb = 3;

			if (ambulances.size() >= numberOfAmb) {
				continue;
			}

			civilians.add(civ);
			centerUpdate.logln("this civ needs center to choose an amb for it" + civ.getID());
		}

	}

	public void ambsMeasure() {

		AmbulancesAndMeasur.clear();
		AmbulancesAndMeasurUsedForCal.clear();

		centerUpdate.logln("|| ambs measures ||");

		for (AmbulanceTeam amb : ambAgents) {
			AmbulancesAndMeasur.add(new AmbulancesAndMeasurments(amb));
			AmbulancesAndMeasurUsedForCal.add(new AmbulancesAndMeasurments(amb));
		}

	}

	public void printAmbAndMeasures() {
		centerUpdate.logln("number of ambulances in amb measure:" + AmbulancesAndMeasur.size());
		//		for (AmbulancesAndMeasurments amb : AmbulancesAndMeasur)
		//			//centerUpdate.logln("amb measures : " + amb.getAmbulance().getID());
	}

	public void refuges() {
		centerUpdate.logln("|| refuges ||");

		refuges.clear();
		for (Refuge ref : centerAmb.model().refuges())
			if ((!ref.isBurning()) && ref.isReallyReachable(true)) {
				refuges.add(ref);
			}
	}

	// error in number of ambs
	public void numberOfAmbNeeded() {
		int timeToRefuge;
		int nowWorking;
		centerUpdate.logln("|| number of AT needed ||");
		for (Pair<CiviliansAndMeasurments, Integer> p : CiviliansAndMeasur) {

			CiviliansAndMeasurments civil = p.first();
			if (civil == null || p.first().getTarget().getHP() == 0 || p.first().getTarget().getAreaPosition() instanceof Refuge || p.first().getTarget().getDamage() == 0)
				continue;
			nowWorking = civil.getTarget().getRescueInfo().getNowWorkingOnMe().size();
			centerUpdate.logln("now working on civ " + civil.getTarget().getID() + " number" + nowWorking);

			if (centerAmb.model().refuges().size() == 0)
				timeToRefuge = 0;

			else
				timeToRefuge = civil.getTarget().getRescueInfo().getTimeToRefuge();

			int deathTime = SimpleDeathTime.getEasyLifeTime(civil.getTarget().getHP(), civil.getTarget().getDamage(), centerAmb.time());

			boolean enough = false;
			for (int i = 1; !enough; i++) {
				if (timeToRefuge + 7 + civil.getTarget().getBuriedness() / i < deathTime) {
					enough = true;
					if (i - nowWorking < 0)
						civil.setNumberOfAmbNeeded(0);
					else
						civil.setNumberOfAmbNeeded(i - nowWorking);

				}
			}

		}
	}

	public void setAmbAndCiv() {

		centerUpdate.logln("|| set ambs and civs ||");
		// // send ambulance position to center
		int timeToResMe = 0;
		for (Pair<CiviliansAndMeasurments, Integer> civil : CiviliansAndMeasur) {

			Human me = civil.first().getTarget();
			if (me.getAreaPosition() instanceof Refuge || me.getHP() == 0 || me.getDamage() == 0)
				continue;
			myAmbs.clear();
			myBestAmbs.clear();
			centerUpdate.info("|| civil that is being decided" + civil.first().getTarget().getID());
			for (AmbulancesAndMeasurments ambulance : AmbulancesAndMeasurUsedForCal) {

				// change it with better strategy
				if (me.getRescueInfo().getNowWorkingOnMe().contains(ambulance.getAmbulance().getID().getValue())) {
					centerUpdate.logln("this amb" + ambulance.getAmbulance().getID() + "is woking on civ " + me.getID());
					continue;
				}

				if (ambulance.getMyTarget() != null) {
					centerUpdate.logln("this amb has a target amb:" + ambulance.getAmbulance().getID() + "civ" + ambulance.getMyTarget().getID());
					continue;
				}

				int moveDistance = centerAmb.model().getDistance(ambulance.getAmbulance().getAreaPosition(), me.getAreaPosition());
				int moveTime = gu.getFoolMoveTimeFromTo(ambulance.getAmbulance().getAreaPosition(), me.getAreaPosition());
				if (centerAmb.model().refuges().size() == 0)
					timeToResMe = moveDistance + me.getBuriedness();
				else {
					sos.base.move.Path path = centerAmb.move.getPathFromTo(java.util.Collections.singleton(me.getAreaPosition()), centerAmb.model().refuges(), StandardMove.class);
					int timeToRefuge = sos.tools.Utils.getSampleTimeToTarget(path);
					timeToResMe = moveDistance + me.getBuriedness() + timeToRefuge;
				}

				IWillBeRes = ambulance.getFreeTime() + timeToResMe;
				myAmbs.add(new Pair<AmbulanceTeam, Integer>(ambulance.getAmbulance(), IWillBeRes));
			}

			Collections.sort(myAmbs, new AmbComparator());

			remove.clear();
			for (int i = 0; i < civil.first().getNumberOfAmbNeeded(); i++) {
				if (!(myAmbs.size() < civil.first().getNumberOfAmbNeeded())) {
					myBestAmbs.add(new Pair<AmbulanceTeam, Integer>(myAmbs.get(i).first(), myAmbs.get(i).second()));
					civil.first().setMyAmbs(myBestAmbs);
					for (int j = 0; j < AmbulancesAndMeasur.size(); j++) {
						AmbulancesAndMeasurments amb = AmbulancesAndMeasur.get(j);
						if (amb.getAmbulance().getID().getValue() == civil.first().getMyAmbs().get(i).first().getID().getValue())
						{
							amb.setMyTarget(civil.first().getTarget());
							amb.setFreeTime(amb.getAmbulance().model().time() + civil.first().getMyAmbs().get(i).second());
							centerUpdate.logln("||ambulance that is choosed:" + amb.getAmbulance().getID() + "for the target" + civil.first().getTarget());
							for (int k = 0; k < AmbulancesAndMeasurUsedForCal.size(); k++)
								if (AmbulancesAndMeasurUsedForCal.get(k).getAmbulance().getID().getValue() == amb.getAmbulance().getID().getValue())
									remove.add(AmbulancesAndMeasurUsedForCal.get(k));
						}
					}
					AmbulancesAndMeasurUsedForCal.removeAll(remove);
				}
				else
					for (int o = 0; o < myAmbs.size(); o++)
					{
						centerUpdate.logln("my ambs size < number of at needed");
						myBestAmbs.add(new Pair<AmbulanceTeam, Integer>(myAmbs.get(o).first(), myAmbs.get(o).second()));
					}
			}

			//			for (AmbulancesAndMeasurments ambi : AmbulancesAndMeasurUsedForCal)
			//				centerUpdate.logln(" ambulances and measure after delet ambulance: " + ambi.getAmbulance().getID());
			//
			//			for (int i = 0; i < myBestAmbs.size(); i++) {
			//				centerUpdate.logln("printing best ambs and civ : " + civil.first().getTarget().getID() + "ambulance: " + myBestAmbs.get(i).first() + "rescue time : " + myBestAmbs.get(i).second());
			//			}
		}
	}

	public void printAssigned() {
		for (AmbulancesAndMeasurments AT : AmbulancesAndMeasur) {
			centerUpdate.logln("number of ambulances and Measur" + AmbulancesAndMeasur.size() + "ambulance " + AT.getAmbulance() + "assigned to" + AT.getMyTarget());
		}

	}

	public void checkCivList() {
		centerUpdate.logln("||check civs list ||");
		removeciv.clear();

		for (Human civilian : civilians) {
			int deathTime = SimpleDeathTime.getEasyLifeTime(civilian.getHP(), civilian.getDamage(), centerAmb.time());

			if ((civilian.getBuriedness() == 0 && civilian.getDamage() == 0)
					|| (deathTime < centerAmb.time() + 5)) {
				removeciv.add(civilian);
			}
			civilians.removeAll(removeciv);
		}
	}

	public void priorityOfTargets() {

		civilianCosts.clear();
		centerUpdate.logln("|| priority of targets ||");
		int targetCost;

		int frees = 0;
		for (AmbulanceTeam freeAmb : centerAmb.model().ambulanceTeams())
			if (freeAmb.getWork() == null && freeAmb.getWork().getTarget() == null)
				frees++;

		for (Human civ : civilians) {
			if (civ == null) {
				centerUpdate.logln("civ was not valid null");
				continue;
			}
			if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
				centerUpdate.logln("civ was not valid damage or hp or pos " + civ.getID());
				continue;
			}
		
			if (civ.getDamage() == 0) {
				centerUpdate.logln("civ was not valid damage" + civ.getID());

				continue;
			}
			if (civ.getAreaPosition() instanceof Refuge) {
				centerUpdate.logln("civ was not valid instanceof refuge" + civ.getID());
				continue;
			}
			

			if (civ.getHP() == 0) {
				centerUpdate.logln("civ was not valid hp = 0" + civ.getID());
				continue;
			}

			if (civ.getRescueInfo().getIgnoredUntil() > centerAmb.model().time()) {
				centerUpdate.logln("civ was not valid ignored" + civ.getID());
				continue;
			}

			if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
				continue;

			if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
				continue;

			if (centerAmb.model().refuges().size() == 0) {
				timeToRefuge = 0;
			}
			else

				timeToRefuge = civ.getRescueInfo().getTimeToRefuge();

			numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, civ, civ.getRescueInfo().getDeathTime());

			//TODO UBUYUBBUCDEFEWFEWFGERGRE
			targetCost = numberOfAmb / (frees + 1) + civ.getBuriedness() + (civ.getHP() * civ.getRescueInfo().getDeathTime());

			if (civ instanceof FireBrigade || civ instanceof AmbulanceTeam || civ instanceof PoliceForce) {
				//				targetCost = (((civil.getBuriedness() * civil.getHP()) / civil
				//						.getDamage())) / 2;
				targetCost /= 2;
				centerUpdate.logln("civ agent" + civ.getID() + "cost" + targetCost);
			}

			civClusterIndex = mapClusters.indexOf(civ.model().searchWorldModel.getClusterData(civ));
			targetCost *= civClusterIndex;
			civilianCosts.add(new Pair<Human, Integer>(civ, targetCost));

		}
		Collections.sort(civilianCosts, new CostComparator());
		centerUpdate.logln("CivilianCosts size" + civilianCosts.size());
	}

	public void civiliansMeasure(ArrayList<Pair<Human, Integer>> civilianCosts, boolean isCenter, AmbulanceTeam agent) {
		centerUpdate.logln("|| civilians measures ||");

		CiviliansAndMeasur.clear();
		for (Pair<Human, Integer> p : civilianCosts) {
			if (p.first() == null || p.first().getHP() == 0 || p.first().getAreaPosition() instanceof Refuge || p.first().getDamage() == 0)
				continue;
			Human civil = p.first();
			CiviliansAndMeasur.add(new Pair<CiviliansAndMeasurments, Integer>(
					new CiviliansAndMeasurments(civil, ambAgents, refuges),
					civil.getID().getValue()));
		}
	}

	public void timeToRescue(Civilian civil) {
		centerUpdate.logln("|| time to rescue ||");
		sos.base.move.Path path = centerAmb.move.getPathFromTo(java.util.Collections.singleton(civil.getAreaPosition()), centerAmb.model().refuges(), StandardMove.class);
		int timeToRefuge = AverageTimeToTarget + sos.tools.Utils.getSampleTimeToTarget(path);

	}

	public void sendAssignMessages() {
		centerUpdate.logln("|| send assign message ||");

		for (AmbulancesAndMeasurments AT : AmbulancesAndMeasur) {
			Human target = AT.getMyTarget();

			centerUpdate.logln("entered the for of send assign message");

			//centerUpdate.logln("Ambulance" + AT.getAmbulance().getID() + " target : " + AT.getMyTarget());

			if (target == null) {
				centerUpdate.logln("ambualnce " + AT.getAmbulance().getID() + "target is null");
				continue;
			}

			int mlposIndex = target.getPositionArea().getAreaIndex();
			int LongLife = 0;
			if (target.getRescueInfo().longLife())
				LongLife = 1;

			centerUpdate.info("Agent" + AT.getAmbulance().getID() + "Target" + target);
			centerAmb.messageBlock = new MessageBlock(MessageXmlConstant.HEADER_AMBULANCE_ASSIGN);
			centerAmb.messageBlock.addData(MessageXmlConstant.DATA_AMBULANCE_INDEX, AT.getAmbulance().getAmbIndex());
			centerAmb.messageBlock.addData(MessageXmlConstant.DATA_ID, target.getID().getValue());
			centerAmb.messageBlock.addData(MessageXmlConstant.DATA_AREA_INDEX, mlposIndex);
			centerAmb.messageBlock.addData(MessageXmlConstant.DATA_LONG_LIFE, LongLife);
			centerAmb.messageBlock.setResendOnNoise(false);
			centerAmb.messages.add(centerAmb.messageBlock);
		}

	}

	public void updateHumansInfo() {
		if (lastTimeUpdated == centerAmb.time())
			return;
		ArrayList<Area> srcs = new ArrayList<Area>();

		if (centerAmb.getCenterActivities().isEmpty())
			srcs.add(centerAmb.me().getAreaPosition());
		else {
			for (AmbulanceTeam at : centerAmb.model().ambulanceTeams())
				srcs.add(at.getAreaPosition());
		}

		for (Refuge ref : centerAmb.model().refuges()) {
			srcs.add(ref);
		}

		costTable = new MultiDitinctSourceCostInMM(centerAmb.model(), srcs);
		lastTimeUpdated = centerAmb.time();

		for (Human human : centerAmb.model().humans()) {
			boolean result = updateHuman(human);
			if (!result)
				continue;

		}
	}

	public boolean updateHuman(Human hm) {
		//initial checks
		if (hm == null) {
			return false;
		}
		if (!hm.isBuriednessDefined()) {
			return false;
		}
		if (!(hm instanceof Civilian) && hm.getBuriedness() == 0) {
			return false;
		}
		if (!hm.isPositionDefined()) {
			return false;
		}

		if (hm.getPosition() instanceof Refuge) {
			return false;
		}
		if (hm.getPosition() instanceof AmbulanceTeam) {
			return false;
		}
		if (hm instanceof Civilian && hm.getPosition() instanceof Road && hm.getDamage() == 0 && hm.getHP() == 10000) {// chera hp==10000?
			return false;
		}

		if (centerAmb.model().refuges().isEmpty()) {
			hm.getRescueInfo().setTimeToRefuge(1);
			hm.getRescueInfo().setBestRefuge(null);
		}
		else
		{
			calculateRefugeInformation(hm);
		}

		return true;
	}

	public void calculateRefugeInformation(Human human) {
		if (!human.isPositionDefined() || human.getPosition() instanceof Refuge)
			return;
		if (!(human instanceof Civilian))
			return;
		if (centerAmb.model().refuges().isEmpty() || human.getRescueInfo().getRefugeCalculatedTime() == centerAmb.time())
			return;
		if (costTable == null) {//if is from hear
			human.getRescueInfo().setTimeToRefuge(AVERAGE_MOVE_TO_TARGET);
			return;
		}
		long minCost = Long.MAX_VALUE;
		Refuge best = null;
		for (Refuge refuge : centerAmb.model().refuges()) {
			long cost = costTable.getCostFromTo(refuge, human.getAreaPosition());
			if (cost < minCost) {
				best = refuge;
				minCost = cost;
			}
		}

		if (best != null && minCost < MoveConstants.UNREACHABLE_COST_FOR_GRAPH_WEIGTHING) {
			//TODO else nabayad null mikard?
			human.getRescueInfo().setBestRefuge(best);
			human.getRescueInfo().setTimeToRefuge(gu.getFoolMoveTime(minCost));
		}

		human.getRescueInfo().setRefugeCalculatedTime(centerAmb.time());

	}

	public boolean isValiadToDecide(Human civ, int time, boolean isCenter, AmbulanceTeam amb) {

		if (civ == null) {
			centerUpdate.logln("civ was not valid null");
			return false;
		}
		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
			centerUpdate.logln("civ was not valid damage or hp or pos " + civ.getID());
			return false;
		}
		if (!isCenter && !AmbUtile.isReallyReachable(civ, true)) {
			centerUpdate.logln("civ was not valid not reachable" + civ.getID());
			return false;
		}
		//		if (civ.getBuriedness() == 0 && civ.getHP() == 10000) {
		//			centerUpdate.logln("civ was not valid buridness 0 and hp 1000" + civ.getID());
		//			return false;
		//		}

		if (civ.getDamage() == 0) {
			centerUpdate.logln("civ was not valid damage" + civ.getID());

			return false;
		}
		if (civ.getAreaPosition() instanceof Refuge) {
			centerUpdate.logln("civ was not valid instanceof refuge" + civ.getID());
			return false;
		}
		//		if (!isCenter && civ.getPosition() instanceof AmbulanceTeam) {
		//			("civ was not valid instanceof ambulance" + civ.getID());
		//			return false;
		//		}

		if (civ.getHP() == 0) {
			centerUpdate.logln("civ was not valid hp = 0" + civ.getID());
			return false;
		}

		if (civ.getRescueInfo().getIgnoredUntil() > time) {
			centerUpdate.logln("civ was not valid ignored" + civ.getID());
			return false;
		}
		if (!isCenter && civ.updatedtime() > time - 3) {
			centerUpdate.logln("updated time" + civ.getID());
			return false;
		}
		if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty())
			return false;

		if (!(civ instanceof Civilian) && civ.getBuriedness() == 0)
			return false;

		return true;
	}

	public void sortingMapClusters() {

		mapBurriedness = 0;

		for (Human civ : centerAmb.model().humans())
		{
			mapBurriedness += civ.getBuriedness();

		}

		for (ClusterData cluster : centerAmb.model().searchWorldModel.getAllClusters()) {
			clusterBurriedness = 0;
			numberOfHumans = 0;
			for (Human civ : centerAmb.model().civilians())
			{

				if (cluster.getClusterArea().contains(civ.getX(), civ.getY())) {

					clusterBurriedness += civ.getBuriedness();
					numberOfHumans++;
				}
			}

			clusterScore = (numberOfHumans) / ((float) (clusterBurriedness / (mapBurriedness + 1)) + 1);
			mapClusters.add(new Pair<ClusterData, Float>(cluster, clusterScore));

		}

		Collections.sort(mapClusters, new clusterComparator());

		// TODO Auto-generated method stub

	}

}

class AmbComparator implements Comparator<Pair<AmbulanceTeam, Integer>> {

	// return -1 : go to the top. 

	@Override
	public int compare(Pair<AmbulanceTeam, Integer> o1,
			Pair<AmbulanceTeam, Integer> o2) {
		if (o1.second() < o2.second())
			return -1;
		else if (o1.second() > o2.second())
			return 1;
		return 0;
	}

}

class CostComparator implements Comparator<Pair<Human, Integer>> {
	@Override
	public int compare(Pair<Human, Integer> h1, Pair<Human, Integer> h2) {// Target
																		  // cost

		if (h1.second() < h2.second())
			return -1;
		else if (h1.second() > h2.second())
			return 1;
		else if (h1.second() == h2.second() && (h1.first().getID().getValue() > h2.first().getID().getValue()))
			return -1;
		else
			return 1;

	}
}

class clusterComparator implements Comparator<Pair<ClusterData, Float>> {

	@Override
	public int compare(Pair<ClusterData, Float> o1, Pair<ClusterData, Float> o2) {
		// TODO Auto-generated method stub
		if (o1.second() > o2.second())
			return -1;
		else if (o1.second() < o2.second())
			return 1;
		else if (o1.first().getY() > o2.first().getY())
			return -1;
		else
			return 1;

	}

}
