package sos.ambulance_v2.decision.states;

import java.util.ArrayList;
import java.util.List;

import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.base.RescueInfo.IgnoreReason;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.decision.controller.CenterDecision;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.Building;
import sos.base.entities.Human;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.util.SOSActionException;
import sos.search_v2.tools.cluster.ClusterData;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class ExcessiveSearchState extends SOSIState<AmbulanceInformationModel> {

	CenterDecision centerDecide = new CenterDecision(infoModel.getAmbulance());
	ArrayList<Human> nearest;
	boolean MyClusterIsDone;
	ClusterData myCluster;

	public ExcessiveSearchState(AmbulanceInformationModel infoModel) {
		super(infoModel);

		// TODO Auto-generated constructor stub
	}

	//TODO
	// add cluster percentage search instead of 10 cycle
	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {

		if (infoModel.getAmbulance().last_task != ATTask.SearchTask)
			return null;
		infoModel.getLog().info("//Excessive search state\\");
		nearest = new ArrayList<Human>();
		infoModel.getLog().info("searchTImes: " + infoModel.getAmbulance().getSearchtimes());
		checkCLuster();
		if (infoModel.getAmbulance().getSearchtimes() > 10 && MyClusterIsDone)
		{
			infoModel.getLog().info("excessive search");
			if ((infoModel.getAgent().messageSystem.type == Type.LowComunication || infoModel.getAgent().messageSystem.type == Type.NoComunication))
			{
				if (MyClusterIsDone)
					LowNoStrategy();
			}

			else if (!(infoModel.getAgent().messageSystem.type == Type.LowComunication || infoModel.getAgent().messageSystem.type == Type.NoComunication))
				chooseNearest();
		}
		// TODO Auto-generated method stub
		return null;
	}

	private void checkCLuster() {
		int buildingSize = infoModel.getAmbulance().getMyClusterData().getBuildings().size();
		int checkedBuildingsNumber = 0;
		for (Building build : infoModel.getAmbulance().getMyClusterData().getBuildings()) {
			if (build.getChecked())
				checkedBuildingsNumber++;
		}
		infoModel.getLog().info("buildingsSize = " + buildingSize + "checked = " + checkedBuildingsNumber);
		if (checkedBuildingsNumber > buildingSize / 3)
			MyClusterIsDone = true;
		else
			MyClusterIsDone = false;
		// TODO Auto-generated method stub

	}

	private void LowNoStrategy() {
		if (infoModel.getAmbulance().underMissionTarget() != null)
			infoModel.getAmbulance().underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, 10000);
		AmbUtile.finish(infoModel.getAmbulance());
		infoModel.getATEntity().getWork().setTarget(null, null);

		for (Human civil : infoModel.getModel().civilians()) {
//			if (myCluster.getBuildings().contains(civil)) {
//				if (AmbUtile.isValiadToDecide(civil, infoModel.getTime(), false, infoModel.getATEntity())) {
//					infoModel.getAmbulance().target = civil;
//					infoModel.getATEntity().getWork().setTarget(civil, this);
//					return;
//				}
//			}
//
			int deathTime = SimpleDeathTime.getEasyLifeTime(infoModel.getATEntity().getHP(), infoModel.getATEntity().getDamage(), infoModel.getTime());

			if (AmbUtile.isValiadToDecide(civil, infoModel.getTime(), false, infoModel.getATEntity())) {
				infoModel.getAmbulance().target = civil;
				infoModel.getATEntity().getWork().setTarget(civil, this);
				return;
			}
		}

		// TODO Auto-generated method stub

	}

	private void chooseNearest() {
		int min = 10000000;
		nearest.clear();
		for (Human civil : infoModel.getATEntity().model().civilians())
		{
			if (!AmbUtile.isValiadToDecide(civil, infoModel.getTime(), false, infoModel.getATEntity()))
				continue;
			if (infoModel.getATEntity().getAreaPosition() == null || civil.getAreaPosition() == null)
				continue;
			int moveTime = centerDecide.gu.getFoolMoveTimeFromTo(infoModel.getATEntity().getAreaPosition(), civil.getAreaPosition());
			if (moveTime < min) {
				min = moveTime;
				nearest.clear();
				nearest.add(civil);
			}

		}

		if (nearest.size() == 0)
			return;
		
		infoModel.getLog().info("I choosed nearest target " + nearest.get(0).getID());
		infoModel.getAmbulance().target = nearest.get(0);
		infoModel.getATEntity().getWork().setTarget(nearest.get(0), this);
		infoModel.getAmbulance().setUnderMissionTarget(null);

		// TODO Auto-generated method stub

	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub

	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}
