package sos.ambulance_v2.decision.states;

import java.util.List;

import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.tools.GraphUsage;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.Refuge;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.types.StandardMove;
import sos.base.util.SOSActionException;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.decisionMaker.implementations.targets.EmptyTarget;
import sos.tools.decisionMaker.implementations.tasks.GoToRefugeTask;
import sos.tools.decisionMaker.implementations.tasks.RestTask;

public class IHaveDamage extends SOSIState<AmbulanceInformationModel> {

	public GraphUsage gu;
	int ATDeathTime;
	int timeToRefuge;

	public IHaveDamage(AmbulanceInformationModel infoModel) {
		super(infoModel);
		gu = new GraphUsage(infoModel.getAgent());
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {

		if (infoModel.getATEntity().getDamage() == 0)
			return null;
		
		infoModel.getLog().info("^^ DAMAGE State ^^");
		
		ATDeathTime = SimpleDeathTime.getEasyLifeTime(infoModel.getATEntity().getHP(), infoModel.getATEntity().getDamage(), infoModel.getTime());
		sos.base.move.Path path = infoModel.getAgent().move.getPathFromTo(java.util.Collections.singleton(infoModel.getATEntity().getAreaPosition()), infoModel.getModel().refuges(), StandardMove.class);
		timeToRefuge = sos.tools.Utils.getSampleTimeToTarget(path);
		
		infoModel.getLog().info("** DEATH T : " + ATDeathTime + "T to REF:" + timeToRefuge);
		
		// TODO Is "20" ok?
		
		if (ATDeathTime < infoModel.getTime() + timeToRefuge + 20) {

			if (infoModel.getATEntity().getAreaPosition() instanceof Refuge) {
				
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				infoModel.getAmbulance().last_task = ATTask.RestTask;
				
				infoModel.getATEntity().getWork().setMyState(ATTask.RestTask);
				return new RestTask(new EmptyTarget(), infoModel.getTime(), "I have damage");
			}

			else {
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				infoModel.getAmbulance().last_task = ATTask.GoToRefugeTask;
				
				infoModel.getATEntity().getWork().setMyState(ATTask.GoToRefugeTask);
				return new GoToRefugeTask(null, infoModel.getTime());
			}
		}

		return null;
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub

	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}
