package sos.ambulance_v2.decision.states;

import java.util.ArrayList;
import java.util.List;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.AmbulanceTeamAgent;
import sos.ambulance_v2.base.RescueInfo.IgnoreReason;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.Area;
import sos.base.entities.Human;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.util.SOSActionException;
import sos.base.util.SOSGeometryTools;
import sos.base.util.information_stacker.CycleInformations;
import sos.base.util.information_stacker.act.MoveAction;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.decisionMaker.implementations.targets.EmptyTarget;
import sos.tools.decisionMaker.implementations.tasks.DummyWalkTask;



public class ImStuck extends SOSIState<AmbulanceInformationModel> {
	int numberOfNearPosition = 0;
	public ImStuck(AmbulanceInformationModel infoModel) {
		super(infoModel);
		infoModel.getLog().log("IMSTUCK");
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		
		// $$$$$$$$$
		
		if (infoModel.getTime() < 5)
			return null;
		if(infoModel.getAmbulance().last_task == ATTask.RescueTask)
			return null;
		if (!IAmStuck(infoModel))
			return null;
		
		infoModel.getLog().info("//Stuck State\\");
		
		
		if (AmbUtile.isLoadingCivilian((AmbulanceTeamAgent)infoModel.getAgent())){
				infoModel.getAmbulance().unload();
				infoModel.getAmbulance().underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.UnloadInRoad, infoModel.getTime() + 10);
				return null;
		}

		
		if(infoModel.getAmbulance().last_task == ATTask.GoingToCivilian){
			infoModel.getATEntity().getWork().getRealTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.NotReachableToTarget, infoModel.getTime()+5);
			return null;
		}
		
		
		if(infoModel.getAmbulance().last_task == ATTask.SearchTask){
			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.MoveTask;
			infoModel.getAmbulance().last_task = ATTask.MoveTask;

			return new DummyWalkTask(new EmptyTarget(), infoModel.getTime());
		}
			
		return null;

	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub

	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}

	@Override
	public void hear(String header, DataArrayList data,
			SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

	public boolean IAmStuck(AmbulanceInformationModel infomodel) {
		List<Point2D> positions = new ArrayList<Point2D>();
		numberOfNearPosition = 0;
		if( infomodel.getTime() > 10){
		for (int i = 1; i <= 5; i++) {
			CycleInformations act = infomodel.getAgent().informationStacker
					.getInformations(i);
			if (act.getAct() instanceof MoveAction)
				positions.add(act.getPositionPair().second());
		}
		}
		
		for (int i = 0; i < positions.size(); i++) {
			int minDistanceToThere = getdistanceToHere(positions.get(i));
			if (minDistanceToThere < 1000)
				numberOfNearPosition++;
		}

		if (numberOfNearPosition >= 3)
			return true;
		
		return false;

	}

	// saina
	private int getdistanceToHere(Point2D point2d) {
		int minDistance = SOSGeometryTools.distance(point2d, infoModel
				.getATEntity().getPositionPoint());
		return minDistance;
	}

	// if map was no refuge AT was not carring target to refuge. !!!! we should
	// think about it.

;

	ArrayList<Pair<Area, Integer>> refuges = new ArrayList<Pair<Area,Integer>>();
	
	private boolean shouldIignoreIt() {
		refuges = AmbUtile.refugesAndTimes(infoModel);
		if (decideToIgnore())
			return true;
		return false;
	}

	Pair<Area, Integer> secondBest;

	private boolean decideToIgnore() {
		if(!AmbUtile.isValiadToDecideUndermission(infoModel.getAmbulance().underMissionTarget(), infoModel.getTime(), infoModel.getATEntity()))
			return true;
		Human target = infoModel.getATEntity().getWork().getTarget();
		int targetDeathTime = SimpleDeathTime.getEasyLifeTime(target.getHP(),
				target.getDamage(), infoModel.getModel().time());
		
		if(refuges.size() > 0){
			secondBest = refuges.get(1);
			if (secondBest.second() < targetDeathTime - 10)
			return false;
		}
		else
			return true;
		return true;

	}
}



