package sos.ambulance_v2.decision.states;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Vector;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.AmbulanceTeamAgent;
import sos.ambulance_v2.base.AmbulanceConstants.ATstates;
import sos.ambulance_v2.base.RescueInfo.IgnoreReason;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.util.SOSActionException;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.decisionMaker.implementations.targets.AreaTarget;
import sos.tools.decisionMaker.implementations.tasks.GoToRefugeTask;
import sos.tools.decisionMaker.implementations.tasks.GoToRoadTask;
import sos.tools.decisionMaker.implementations.tasks.LoadTask;
import sos.tools.decisionMaker.implementations.tasks.RescueTask;
import sos.tools.decisionMaker.implementations.tasks.StandardMoveToAreaTask;
import sos.tools.decisionMaker.implementations.tasks.UnloadTask;

//** Undermission better than Center 
/**
 * 
 * @author Saina Ghadiany
 *
 */

public class RescueState extends SOSIState<AmbulanceInformationModel> {

	private Human target;
	private AmbulanceTeamAgent ambulance;
	ArrayList<AmbulanceTeam> ambulances;
	Pair<? extends Area, Point2D> place;
	private SOSTask<?> task = null;
	boolean hasEnoughAT = false;
	boolean letItGo = false;
	private int numberOfExtremeAmbs;
	ArrayList<AmbulanceTeam> ambulancesAndIds;
	ArrayList<Building> firee;
	Vector<AmbulanceTeam> nowWorking;

	public RescueState(AmbulanceInformationModel infoModel) {
		super(infoModel);
		ambulances = new ArrayList<AmbulanceTeam>();
		ambulancesAndIds = new ArrayList<AmbulanceTeam>();
		nowWorking = new Vector<AmbulanceTeam>();

	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {

		if (infoModel.getTime() < 5)
			return null;

		infoModel.getLog().info("//Rescue State\\");
		ambulance = infoModel.getAmbulance();
		if (ambulance.underMissionTarget()!=null && ambulance.underMissionTarget().getRescueInfo().getIgnoredUntil() > infoModel.getTime()) {
			infoModel.getLog().info("hiii");
		}
		
		if ((ambulance.isLoadingInjured() && ambulance.underMissionTarget() != null && ambulance.underMissionTarget().getHP() != 0) ||
				AmbUtile.isValiadToDecideUndermission(ambulance.underMissionTarget(), infoModel.getTime(), infoModel.getATEntity())) {
			if (infoModel.getAmbulance().isLoadingInjured()) {
				if (infoModel.getAmbulance().loadingInjured().getID().getValue() != infoModel.getAmbulance().underMissionTarget().getID().getValue())
					infoModel.getAmbulance().setUnderMissionTarget(infoModel.getAmbulance().loadingInjured());
				infoModel.getLog().info("entered check for undermission target and is loading injured");
			}
			ambulance.log().info("^^ entered check for undermissiontarget");
			target = ambulance.underMissionTarget();
			check(target);
			return task;

		}

		else if ((infoModel.getAmbulance().isLoadingInjured() && ambulance.underMissionTarget() != null)) {
			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.UnloadTask;
			infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);
			task = new UnloadTask(target, infoModel.getTime());
			infoModel.getLog().info("setting ignored because target is dead");
			ambulance.underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.DeadHuman, 1000);
			AmbUtile.finish(ambulance);
			infoModel.getLog().info("My target is dead I am unloading it");
			return task;
		}

		if (ambulance.underMissionTarget() != null) {
			infoModel.getLog().info("Civ is gone Iam changing my undermission");
			infoModel.getLog().info("setting ignored because civ is gone");
			//TODO 
			//WILL DIE?
			ambulance.underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.InvalidOldTarget, infoModel.getTime() + 10);
			AmbUtile.finish(ambulance);
			
			if (infoModel.getAmbulance().target != null && ambulance.underMissionTarget().getID() == infoModel.getAmbulance().target.getID())
			{
				AmbUtile.sendTaskAckMsg(2, infoModel.getAmbulance(), ambulance.underMissionTarget());
				infoModel.getAmbulance().target = null;
			}

			ambulance.setUnderMissionTarget(null);
			infoModel.getATEntity().getWork().setTarget(null, null);

		}

		target = infoModel.getAmbulance().target;

		if (!AmbUtile.isValiadToDecide(target, infoModel.getTime(), false, infoModel.getATEntity()))
		{
			infoModel.getLog().info("^^ undermission target has been null and target is not valid");
			ambulance.setUnderMissionTarget(null);
			ambulance.target = null;
			ambulance.centerRecommendedTarget = null;
			infoModel.getATEntity().getWork().setTarget(null, null);
			target = null;
			//if(infoModel.getTime() < 50)
			return null;
			//TODO
			//			else
			//				chooseTargetEmerg();
		}

		else {
			ambulance.log().info("^^ entered check for target");
			check(target);
		}

		ambulance.setUnderMissionTarget(target);

		if (target != null)
			infoModel.getATEntity().getWork().setTarget(target, this);
		else
			infoModel.getATEntity().getWork().setTarget(null, null);
		return task;

	}

//	private void chooseTargetEmerg() {
//		// TODO Auto-generated method stub
//
//	}

	private void check(Human tar) {
		if (tar == null)
			return;
		infoModel.getLog().info("^^ target != null and is" + tar.getID() + "^^");
		if (!(tar.getPositionArea().equals(infoModel.getATEntity().getAreaPosition()))) {

			if (infoModel.getAmbulance().last_task == ATTask.MoveTask && infoModel.getTime() < 140) {
				checkIfMoveContinue(tar);
//				if(infoModel.getTime() > 140)
//					hasEnoughAT = false;
				if (!hasEnoughAT || letItGo) {
					infoModel.getLog().info("has enough at I should not go in checkmove ");
					ambulance.setUnderMissionTarget(null);
					ambulance.target = null;
					if (letItGo && tar != null && !hasEnoughAT)
						tar.getRescueInfo().setIgnoredUntil(IgnoreReason.WillDie, 1000);

					target = null;

					infoModel.getATEntity().getWork().setTarget(null, null);
					task = null;
					infoModel.getLog().info("hiiii");
					return;
				}
			}

			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.MoveTask;
			infoModel.getATEntity().getWork().setMyState(ATTask.MoveTask);

			task = new StandardMoveToAreaTask(new AreaTarget(tar.getAreaPosition()), infoModel.getTime());
			infoModel.getLog().info("^^ my task is standard move to area task now ^^");
			int busy = AmbUtile.getMyBusyCycles(target, place, ATTask.MoveTask, infoModel.getAmbulance());
			AmbUtile.sendInfoMsgIfRequire(0, busy, infoModel.getAmbulance());
			AmbUtile.sendStatueMsgBySay(0, false, infoModel.getAmbulance(), infoModel.getAmbulance().last_task);
			AmbUtile.sendMessage(0, ambulance, tar);
			if (infoModel.getTime() % 2 == 0)
				AmbUtile.sendTaskAckMsg(0, infoModel.getAmbulance(), tar);

			return;
		}

		//move zadam vali hich kas unja nabod
		if (infoModel.getAmbulance().getVisibleEntities(Human.class) == null || !infoModel.getAmbulance().getVisibleEntities(Human.class).contains(tar))
		{
			if (ambulance.underMissionTarget() != null) {
				ambulance.underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.InvalidOldTarget, infoModel.getTime() + 1000);
				AmbUtile.sendTaskAckMsg(2, infoModel.getAmbulance(), ambulance.underMissionTarget());
			}
			if (infoModel.getAmbulance().target != null && ambulance.underMissionTarget() != null && ambulance.underMissionTarget().getID() == infoModel.getAmbulance().target.getID())
			{
				infoModel.getAmbulance().target = null;
			}

			ambulance.setUnderMissionTarget(null);
			infoModel.getATEntity().getWork().setTarget(null, null);
			return;
		}
		if (tar.getBuriedness() == 0) {
			if (tar.getPositionArea() instanceof Refuge) {
				tar.getRescueInfo().setIgnoredUntil(IgnoreReason.InRefuge, 1000);
				AmbUtile.sendMessage(2, ambulance, tar);
				AmbUtile.finish(ambulance);
			}

			infoModel.getLog().info("^^ I have target but it does not have buriedness. target is" + tar.getID() + "^^");
			dontHaveBurridness();
		}

		else {
			infoModel.getLog().info(" ^^ I have target, it has buriedness, I am sending rescue task, target is" + tar.getID() + "^^");

			checkIfResContinue(tar);

			if (hasEnoughAT || letItGo)
			{
				infoModel.getLog().info("I SHOULD NOT CONTINUE RESCUING");
				ambulance.setUnderMissionTarget(null);
				ambulance.target = null;
				ambulance.centerRecommendedTarget = null;
				target = null;
				if (letItGo) {
					infoModel.getLog().info("let it go was true");
					infoModel.getATEntity().getWork().getTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.WillDie, 10000);
				}

				infoModel.getATEntity().getWork().setTarget(null, null);
				return;

			}

			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.RescueTask;
			infoModel.getATEntity().getWork().setMyState(ATTask.RescueTask);

			infoModel.getATEntity().getWork().setTarget(tar, this);
			place = tar.getPositionPair();
			int busy = AmbUtile.getMyBusyCycles(tar, place, ATTask.RescueTask, infoModel.getAmbulance());
			task = new RescueTask(tar, infoModel.getTime());
			AmbUtile.sendRescueMsgBySay(tar, infoModel.getAmbulance());
			AmbUtile.sendInfoMsg(tar.getID().getValue(), 1, busy, infoModel.getAmbulance());

		}

	}

	private void checkIfMoveContinue(Human tar) {
		infoModel.getLog().info("check if move continue" + tar.getID());
		
		
		if (tar == null || tar.getDamage() == 0 || tar.getHP() == 0 || tar.getRescueInfo().getIgnoredUntil() > infoModel.getTime()){
			letItGo = true;
			return;
		}
		else
			letItGo = false;
		
		int numberOfAmb = 0;

		ambulances.clear();

		for (AmbulanceTeam amb : infoModel.getModel().ambulanceTeams()) {
			if (!amb.isAlive() || amb.getHP() == 0 || amb.getBuriedness() != 0 || amb.getWork() == null || amb.getWork().getTarget() == null)
				continue;
			else if (amb.getWork().getTarget().getID().getValue() == tar.getID().getValue() && amb.getID().getValue() != infoModel.getAmbulance().getID().getValue() && amb.getWork().getCurrentState() == ATstates.RESCUE) {
				ambulances.add(amb);
				infoModel.getLog().info("amb target is as same as mine and it is rescuing the civ" + amb.getID());
			}
		}

		int timeToRefuge;
		if (infoModel.getModel().refuges().size() == 0)
			timeToRefuge = 0;
		else
			timeToRefuge = tar.getRescueInfo().getTimeToRefuge();


		int deathTime = SimpleDeathTime.getEasyLifeTime(tar.getHP(), tar.getDamage(), infoModel.getTime());

		boolean enough = false;
		for (int i = 1; !enough; i++) {
			if(i> infoModel.getModel().ambulanceTeams().size()){
				hasEnoughAT = false;
				numberOfExtremeAmbs = 0;
				//TODO letitgo true
				break;
			}
			if (timeToRefuge + 7 + tar.getBuriedness() / i < deathTime) {
				enough = true;
				hasEnoughAT = true;
				numberOfAmb = i;
				numberOfExtremeAmbs = ambulances.size() - numberOfAmb;
			}
		}

		int timeToRescue = tar.getBuriedness() / Math.max(1, numberOfAmb) + timeToRefuge + infoModel.getTime() + 2;

		if (tar.getHP() < 150) {
			infoModel.getLog().info("hp");
			letItGo = true;
		}

		else if (timeToRescue > deathTime + 2) {
			int withoutWork = 0;
			for (AmbulanceTeam amb : infoModel.getModel().ambulanceTeams()) {
				if (amb.getWork().getTarget() == null)
					withoutWork++;
			}
			if (numberOfAmb > withoutWork - 3) {

				infoModel.getLog().info("Time to rescue");
				letItGo = true;
			}
		}
		else
			letItGo = false;

	}

	private void checkIfResContinue(Human tar) {
		letItGo = false;
		hasEnoughAT = false;
		if (tar == null || tar.getDamage() == 0 || tar.getHP() == 0)
		{
			letItGo = true;
			return;
		}
		if (infoModel.getTime() < 140)
			theyAreEnough(tar);
		civIsNoMoreGood(tar);

	}

	private void civIsNoMoreGood(Human tar) {
		ambulances.clear();
		for (AmbulanceTeam amb : infoModel.getAmbulance().getVisibleEntities(AmbulanceTeam.class)) {
			if(amb.getBuriedness()!=0)
				continue;
			if (amb.getWork() == null || amb.getWork().getTarget() == null)
				continue;
			else if (amb.getWork().getTarget().getID().getValue() == tar.getID().getValue() && amb.getID().getValue() != infoModel.getAmbulance().getID().getValue() && amb.getWork().getMyState() == ATTask.RescueTask) {
				ambulances.add(amb);
				infoModel.getLog().info("amb target is as same as mine" + amb.getID());
			}
		}

		int timeToRescue = tar.getBuriedness() / (ambulances.size() + 1) + tar.getRescueInfo().getTimeToRefuge() + infoModel.getTime() + 2;

		infoModel.getLog().info("time to res: " + timeToRescue + "death time : " + tar.getRescueInfo().getDeathTime());
		if (tar.getRescueInfo().getDeathTime() + infoModel.getTime() < timeToRescue)
			letItGo = true;

		else
			letItGo = false;

	}

	private void theyAreEnough(Human tar) {
		int numberOfAmb = 0;
		ambulances.clear();
		for (AmbulanceTeam amb : infoModel.getAmbulance().getVisibleEntities(AmbulanceTeam.class)) {
			if(amb.getBuriedness()!=0)
				continue;
			if (amb.getWork() == null || amb.getWork().getTarget() == null) {
				infoModel.getLog().info("amb does not have any work or target");
				continue;
			}
			else if (amb.getWork().getTarget().getID().getValue() == tar.getID().getValue() && amb.getID().getValue() != infoModel.getAmbulance().getID().getValue() && amb.getWork().getCurrentState() == ATstates.RESCUE) {
				ambulances.add(amb);
				infoModel.getLog().info("amb target is as same as mine" + amb.getID());
			}
		}

		int timeToRefuge;
		if (infoModel.getModel().refuges().size() == 0)
			timeToRefuge = 0;
		else
			timeToRefuge = tar.getRescueInfo().getTimeToRefuge();

		int deathTime = SimpleDeathTime.getEasyLifeTime(tar.getHP(), tar.getDamage(), infoModel.getTime());

		boolean enough = false;
		for (int i = 1; !enough; i++) {
			if (timeToRefuge + 7 + tar.getBuriedness() / i < deathTime) {
				enough = true;
				numberOfAmb = i;
			}
		}

		if (numberOfAmb <= ambulances.size()) {
			infoModel.getLog().info("has enough AT:" + tar.getID());
			numberOfExtremeAmbs = ambulances.size() + 1 - numberOfAmb;

			for (AmbulanceTeam amb : ambulances)
				ambulancesAndIds.add(amb);
			ambulancesAndIds.add(infoModel.getATEntity());
			Collections.sort(ambulancesAndIds, new IdComparator());
			for (int i = 0; i < numberOfExtremeAmbs - 1; i++)
				if (ambulancesAndIds.get(i).getID().getValue() == infoModel.getAmbulance().getID().getValue())
				{
					infoModel.getLog().info("I am extreme. I should go");
					hasEnoughAT = true;
				}
		}

		else {
			hasEnoughAT = false;
			numberOfExtremeAmbs = 0;
		}

	}

	private void dontHaveBurridness() {

		nowWorking.clear();
		if (ambulance.isLoadingInjured()){
			loadingInjured();
		return;
		}
		
		if(!ambulance.getVisibleEntities(Civilian.class).contains(target)){
			infoModel.getLog().info("^^ I can not see the target ^^");
			AmbUtile.finish(ambulance);
			target.getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, infoModel.getTime() + 1000);
			target = null;
			infoModel.getATEntity().getWork().setTarget(null, null);
			infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);
			target.getRescueInfo().setIgnoredUntil(IgnoreReason.FinishedWorkOnTarget, 1000);
			AmbUtile.sendMessage(1, ambulance, target);
			AmbUtile.finish(ambulance);
			return;
		}

		else {

			if (!AmbUtile.isValiadToDecideUndermission(target, infoModel.getTime(), infoModel.getATEntity())) {
				infoModel.getAmbulance().setUnderMissionTarget(null);
				target = null;
				return;
			}
			if (!(target instanceof Civilian)) {
				infoModel.getLog().info("^^ target is not civilian, I should not load it^^");
				AmbUtile.finish(ambulance);
				target.getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, infoModel.getTime() + 3);
				target = null;
				infoModel.getATEntity().getWork().setTarget(null, null);
				infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);
				target.getRescueInfo().setIgnoredUntil(IgnoreReason.FinishedWorkOnTarget, 1000);
				AmbUtile.sendMessage(1, ambulance, target);
				AmbUtile.finish(ambulance);
				return;
			}

			for (AmbulanceTeam amb : infoModel.getAmbulance().getVisibleEntities(AmbulanceTeam.class))
			{
				if(amb.getBuriedness()!=0)
					continue;
				infoModel.getLog().info("checking ambulances near" + amb.getID());

				if (amb.getWork() == null || amb.getWork().getTarget() == null)
					continue;
				else if (amb.getWork().getTarget().getID().getValue() == infoModel.getATEntity().getWork().getTarget().getID().getValue()) {
					nowWorking.add(amb);
					infoModel.getLog().info("now working more than me is : " + amb.getID());
				}
			}

			int max = -1;
			if (nowWorking.size() == 1) {

				infoModel.getLog().info(" I am loading target " + target.getID() + "number of ambulances near target =" + nowWorking.size());

				if (infoModel.getAmbulance().last_task == ATTask.LoadTask)
					infoModel.getAmbulance().setLoadtimes(infoModel.getAmbulance().getLoadtimes() + 1);
				else
					infoModel.getAmbulance().setLoadtimes(1);

				infoModel.getLog().info("loadNum = " + infoModel.getAmbulance().getLoadtimes());
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				if (infoModel.getAmbulance().getLoadtimes() > 1) {
					if (ambulance.underMissionTarget() != null)
						ambulance.underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, infoModel.getTime() + 60);
					AmbUtile.finish(ambulance);
				}

				else {
					int busy = AmbUtile.getMyBusyCycles(target, place, ATTask.LoadTask, infoModel.getAmbulance());
					AmbUtile.sendInfoMsgIfRequire(2, busy, infoModel.getAmbulance());
					AmbUtile.sendStatueMsgBySay(2, false, infoModel.getAmbulance(), ATTask.LoadTask);
					infoModel.getATEntity().getWork().setMyState(ATTask.LoadTask);
					infoModel.getAmbulance().last_task = ATTask.LoadTask;
					task = new LoadTask(target, infoModel.getTime());
				}
			}

			else {
				infoModel.getLog().info("^^ more than one ambulance were working on target^^");
				for (AmbulanceTeam amb : nowWorking) {
					if (amb.getID().getValue() > max)
						max = amb.getID().getValue();
				}

				if (max == infoModel.getAgent().getID().getValue()) {

					if (infoModel.getAmbulance().last_task == ATTask.LoadTask)
						infoModel.getAmbulance().setLoadtimes(infoModel.getAmbulance().getLoadtimes() + 1);
					else
						infoModel.getAmbulance().setLoadtimes(1);

					infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
					infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
					infoModel.getAmbulance().last_task = ATTask.LoadTask;
					infoModel.getATEntity().getWork().setMyState(ATTask.LoadTask);
					infoModel.getLog().info("loadNum = " + infoModel.getAmbulance().getLoadtimes());
					if (infoModel.getAmbulance().getLoadtimes() > 2) {
						ambulance.underMissionTarget().getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, infoModel.getTime() + 60);
						infoModel.getATEntity().getWork().setNowWorkingOn(null);
						AmbUtile.finish(ambulance);
						target = null;
					}
					else
					{
						infoModel.getLog().info(" I am loading target " + target.getID() + "number of ambulances near " + nowWorking.size() + "max : " + max + "my id " + ambulance.getID());
						AmbUtile.sendStatueMsgBySay(2, false, infoModel.getAmbulance(), infoModel.getAmbulance().last_task);
						AmbUtile.sendInfoMsg(infoModel.getAmbulance().getID().getValue(), 0, AmbUtile.getMyBusyCycles(target, target.getPositionPair(), ATTask.LoadTask, infoModel.getAmbulance()), ambulance);
						task = new LoadTask(target, infoModel.getTime());
					}

				}
				else
				{
					///TODO
					//now
					infoModel.getLog().info("^^ I should not load the target because someone else is going to, I am ignoring the target, target is: " + target.getID() + "^^");
					task = null;
					target.getRescueInfo().setIgnoredUntil(IgnoreReason.ImNotLoader, 10000);
					ambulance.setUnderMissionTarget(null);
					ambulance.target = null;
					ambulance.log().info("id= " + ambulance.getID() + " target =" + ambulance.target);
					ambulance.centerRecommendedTarget = null;
					target = null;
				}
			}
		}
	}

	private void loadingInjured() {
		infoModel.getLog().info("^^ I have loaded the target. in loadinginjured function. target is: " + target.getID() + "^^");

		if (target.getHP() == 0) {
			infoModel.getLog().info("^^ my target was dead and I am unloading it^^");
			if (infoModel.getAmbulance().last_task == ATTask.UnloadTask)
				infoModel.getAmbulance().setUnloadedtimes(infoModel.getAmbulance().getLoadtimes() + 1);
			else
				infoModel.getAmbulance().setUnloadedtimes(1);
			if (infoModel.getAmbulance().getUnloadedtimes() < 2) {
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				infoModel.getAmbulance().last_task = ATTask.UnloadTask;
				infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);

				task = new UnloadTask(target, infoModel.getTime());
				target.getRescueInfo().setIgnoredUntil(IgnoreReason.DeadHuman, 1000);
				AmbUtile.sendMessage(2, ambulance, target);
				AmbUtile.finish(ambulance);
				target = null;
				infoModel.getATEntity().getWork().setTarget(null, null);
			}

			else {
				target = null;
				infoModel.getAmbulance().setUnderMissionTarget(null);
			}

		}
		else if (infoModel.getModel().refuges().size() == 0 && infoModel.getATEntity().getPositionArea() instanceof Road) {

			infoModel.getLog().info("^^ target and I are in ROad in no refuge map . I am unloading, target is: " + target.getID() + "^^");
			infoModel.getLog().info(" ambulance id " + ambulance.getID() + " target" + ambulance.target + infoModel.getAmbulance().target);
			if (infoModel.getAmbulance().last_task == ATTask.UnloadTask)
				infoModel.getAmbulance().setUnloadedtimes(infoModel.getAmbulance().getLoadtimes() + 1);
			else
				infoModel.getAmbulance().setUnloadedtimes(1);
			if (infoModel.getAmbulance().getUnloadedtimes() < 2) {
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				infoModel.getAmbulance().last_task = ATTask.UnloadTask;
				infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);

				task = new UnloadTask(target, infoModel.getTime());
				target.getRescueInfo().setIgnoredUntil(IgnoreReason.InRefuge, 1000);
				AmbUtile.finish(ambulance);
				target = null;
			}
			else {
				target = null;
				infoModel.getAmbulance().setUnderMissionTarget(null);
			}
		}
		else if (infoModel.getATEntity().getPositionArea() instanceof Refuge) {
			infoModel.getLog().info("^^ target and I are in refuge . I am unloading, target is: " + target.getID() + "^^");
			if (infoModel.getAmbulance().last_task == ATTask.UnloadTask)
				infoModel.getAmbulance().setUnloadedtimes(infoModel.getAmbulance().getLoadtimes() + 1);
			else
				infoModel.getAmbulance().setUnloadedtimes(1);

			if (infoModel.getAmbulance().getUnloadedtimes() < 2) {
				infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
				infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
				infoModel.getAmbulance().last_task = ATTask.UnloadTask;
				infoModel.getATEntity().getWork().setMyState(ATTask.UnloadTask);

				task = new UnloadTask(target, infoModel.getTime());
				target.getRescueInfo().setIgnoredUntil(IgnoreReason.InRefuge, 1000);
				AmbUtile.sendMessage(1, ambulance, target);
				AmbUtile.finish(ambulance);
				infoModel.getLog().info(" I have unload task");
				infoModel.getLog().info(" ambulance id " + ambulance.getID() + " target" + ambulance.target + infoModel.getAmbulance().target);
				target = null;
			}
			else {
				target = null;
				infoModel.getAmbulance().setUnderMissionTarget(null);
			}

		}

		else if (infoModel.getModel().refuges().size() == 0) {
			infoModel.getLog().info("^^ I am taking target to Road, target is:" + target.getID() + "^^");
			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.GoToRoad;
			infoModel.getATEntity().getWork().setMyState(ATTask.GoToRoad);

			task = new GoToRoadTask(null, infoModel.getTime());
			int busy = AmbUtile.getMyBusyCycles(target, target.getPositionPair(), ATTask.GoToRoad, infoModel.getAmbulance());
			AmbUtile.sendInfoMsgIfRequire(2, busy, infoModel.getAmbulance());
		}

		else {
			infoModel.getLog().info("^^ I am taking target to refuge, target is:" + target.getID() + "^^");
			int busy = AmbUtile.getMyBusyCycles(target, target.getPositionPair(), ATTask.GoToRefugeTask, infoModel.getAmbulance());
			AmbUtile.sendInfoMsgIfRequire(2, busy, infoModel.getAmbulance());
			AmbUtile.sendStatueMsgBySay(2, false, infoModel.getAmbulance(), ATTask.GoToRefugeTask);
			if (infoModel.getAmbulance().isFull())
				AmbUtile.sendTaskAckMsg(1, infoModel.getAmbulance(), target);
			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.GoToRefugeTask;
			infoModel.getATEntity().getWork().setMyState(ATTask.GoToRefugeTask);

			task = new GoToRefugeTask(null, infoModel.getTime());
		}

	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub

	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}

class IdComparator implements Comparator<AmbulanceTeam> {

	@Override
	public int compare(AmbulanceTeam o1, AmbulanceTeam o2) {
		if (o1.getID().getValue() > o2.getID().getValue())
			return -1;
		return 1;
	}

	// return -1 : go to the top. 

}
