


package sos.ambulance_v2.decision.states;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import rescuecore2.misc.Pair;
import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.base.AmbulanceConstants.ATstates;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.decision.controller.AmbulancesAndMeasurments;
import sos.ambulance_v2.decision.controller.CenterDecision;
import sos.ambulance_v2.tools.GraphUsage;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.GasStation;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.types.StandardMove;
import sos.base.util.SOSActionException;
import sos.search_v2.tools.cluster.ClusterData;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class SelfDecide extends SOSIState<AmbulanceInformationModel> {

	private static final String string = null;

	public GraphUsage gu;

	CenterDecision centerDecide = new CenterDecision(infoModel.getAmbulance());
	ClusterData myCluster;

	ArrayList<Pair<Human, Integer>> betterThanMine;
	ArrayList<Pair<Human, Integer>> civilianCosts;
	ArrayList<AmbulanceTeam> ambulances;
	ArrayList<AmbulanceTeam> freeAmbAgents;
	ArrayList<Human> civilians;
	ArrayList<Human> ALL;
	ArrayList<Pair<ClusterData, Float>> mapClusters;

	boolean centerIsRight = false;
	boolean MyCLusterIsDone = false;
	boolean hasEnoughAT = false;
	boolean choosed;

	Human target;

	int numberOfAmb;
	int timeToRefuge;
	int undermissionCost;
	int centerCost;
	int moveDistance;
	int targetCost;
	int workingOn;
	int timeToResMe;
	int myClusterIndex;
	int myCivClusterIndex;

	public SelfDecide(AmbulanceInformationModel infoModel) {

		super(infoModel);

		gu = new GraphUsage(infoModel.getAgent());
		centerDecide.myAmbs = new ArrayList<Pair<AmbulanceTeam, Integer>>();
		centerDecide.myBestAmbs = new ArrayList<Pair<AmbulanceTeam, Integer>>();
		civilianCosts = new ArrayList<Pair<Human, Integer>>();
		ambulances = new ArrayList<AmbulanceTeam>();
		freeAmbAgents = new ArrayList<AmbulanceTeam>();
		civilians = new ArrayList<Human>();
		betterThanMine = new ArrayList<Pair<Human, Integer>>();
		ALL = new ArrayList<Human>();

	}

	@Override
	
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {

		if (infoModel.getAgent().messageSystem.type == Type.NoComunication || infoModel.getAgent().messageSystem.type == Type.LowComunication)
			return null;

		infoModel.getLog().info("^^ SELFDECIDE STATE^^");

		// Has both Undermission and CenterRecommended and are valid.

		// TODO TESTTTTTTTTTTTTTTTTTTTTTTTTTTTT

		if (infoModel.getAmbulance().isLoadingInjured())
			return null;

		if (AmbUtile.isValiadToDecideUndermission(infoModel.getAmbulance().underMissionTarget(), infoModel.getTime(), infoModel.getATEntity()) && (infoModel.getAmbulance().last_task == ATTask.RescueTask))
			return null;

		if (AmbUtile.isValiadToDecide(infoModel.getAmbulance().underMissionTarget(), infoModel.getTime(), false, infoModel.getATEntity()) && AmbUtile.isValiadToDecide(infoModel.getAmbulance().centerRecommendedTarget, infoModel.getTime(), false, infoModel.getATEntity())
				&& infoModel.getAmbulance().last_task == ATTask.MoveTask && !infoModel.getAmbulance().isLoadingInjured())
		{

			// TODO TESTTTTTTTTTTTTTTTTTTTTTTTTTTTT

			//undermissionCost = (((infoModel.getAmbulance().underMissionTarget().getBuriedness() + 1) * moveDistance + infoModel.getAmbulance().underMissionTarget().getHP())) / (infoModel.getAmbulance().underMissionTarget()
			//	.getDamage() + 1);

			numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, infoModel.getAmbulance().underMissionTarget(), infoModel.getAmbulance().underMissionTarget().getRescueInfo().getDeathTime());
			moveDistance = infoModel.getModel().getDistance(infoModel.getATEntity().getPositionArea(), infoModel.getAmbulance().underMissionTarget().getAreaPosition());
			undermissionCost = numberOfAmb / (AmbUtile.numberOfFreeAmbs(infoModel.getATEntity()) + 1) + infoModel.getAmbulance().underMissionTarget().getBuriedness() + (infoModel.getAmbulance().underMissionTarget().getHP() * infoModel.getAmbulance().underMissionTarget().getRescueInfo().getDeathTime()) + moveDistance;

			//centerCost = (((infoModel.getAmbulance().centerRecommendedTarget.getBuriedness() + 1) * moveDistance + infoModel.getAmbulance().centerRecommendedTarget.getHP())) / (infoModel.getAmbulance().centerRecommendedTarget
			//	.getDamage() + 1);

			moveDistance = infoModel.getModel().getDistance(infoModel.getATEntity().getPositionArea(), infoModel.getAmbulance().centerRecommendedTarget.getAreaPosition());
			numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, infoModel.getAmbulance().underMissionTarget(), infoModel.getAmbulance().underMissionTarget().getRescueInfo().getDeathTime());
			centerCost = numberOfAmb / (AmbUtile.numberOfFreeAmbs(infoModel.getATEntity()) + 1) + infoModel.getAmbulance().centerRecommendedTarget.getBuriedness() + (infoModel.getAmbulance().centerRecommendedTarget.getHP() * infoModel.getAmbulance().centerRecommendedTarget.getRescueInfo().getDeathTime()) + moveDistance;

			if (centerCost < undermissionCost) {
				infoModel.getLog().info("** Center better than Undermission : " + infoModel.getAmbulance().centerRecommendedTarget.getID() + "**");
				infoModel.getATEntity().getWork().setTarget(infoModel.getAmbulance().centerRecommendedTarget, this);
				infoModel.getAmbulance().target = infoModel.getAmbulance().centerRecommendedTarget;
			}

			else
			{
				infoModel.getLog().info("** Undermission better than Center :" + infoModel.getAmbulance().underMissionTarget().getID() + "**");
				infoModel.getATEntity().getWork().setTarget(infoModel.getAmbulance().underMissionTarget(), this);
				infoModel.getAmbulance().target = infoModel.getAmbulance().underMissionTarget();
			}

			infoModel.getLog().info("** End of SELFDECIDE. Amb has a target " + infoModel.getAmbulance().target.getID());
			return null;
		}

		// No CenterRecom. Undermission Valid

		if (AmbUtile.isValiadToDecideUndermission(infoModel.getAmbulance().underMissionTarget(), infoModel.getTime(), infoModel.getATEntity()))
		{
			infoModel.getLog().info("** NO Center Recom. Undermission Valid**" + infoModel.getAmbulance().underMissionTarget().getID());

			infoModel.getATEntity().getWork().setTarget(infoModel.getAmbulance().underMissionTarget(), this);
			infoModel.getAmbulance().target = infoModel.getAmbulance().underMissionTarget();

			infoModel.getLog().info("** End of SELFDECIDE. Amb has a target " + infoModel.getAmbulance().target.getID());
			return null;
		}

		// No UnderMission. Center Valid.

		if (AmbUtile.isValiadToDecide(infoModel.getAmbulance().centerRecommendedTarget, infoModel.getTime(), false, infoModel.getATEntity())) {
			infoModel.getLog().info("** NO Undermission. CenterRecom Valid**" + infoModel.getAmbulance().centerRecommendedTarget.getID());
			infoModel.getATEntity().getWork().setTarget(infoModel.getAmbulance().centerRecommendedTarget, this);
			infoModel.getAmbulance().target = infoModel.getAmbulance().centerRecommendedTarget;

			infoModel.getLog().info("** End of SELFDECIDE. Amb has a target " + infoModel.getAmbulance().target.getID());
			return null;
		}

		if (infoModel.getTime() < 10)
			return null;

		// TODO TESTTTTTTTTTTTTTTTTTTTTTTTTTTTT

		//		if (infoModel.getAmbulance().threelast_task == ATTask.RescueTask || infoModel.getAmbulance().threelast_task == ATTask.UnloadTask || infoModel.getAmbulance().twolast_task == ATTask.RescueTask || infoModel.getAmbulance().twolast_task == ATTask.UnloadTask || infoModel.getAmbulance().last_task == ATTask.UnloadTask || infoModel.getAmbulance().last_task == ATTask.RescueTask)
		//			return null;

		myCluster = infoModel.getModel().searchWorldModel.getClusterData();

		infoModel.getLog().info(" ^^ Entering SELFDECIDE for decision making ^^ ");
		infoModel.getAmbulance().setUnderMissionTarget(null);
		infoModel.getAmbulance().centerRecommendedTarget = null;
		set();

		return null;
	}

	private void set() {
		findValidAmbulancesAndCivilians(false, infoModel.getATEntity());

		infoModel.getLog().info(" target size is " + civilians.size());
		priorityOfTargets();

		setAmbAndCivs();

		if (infoModel.getAmbulance().target == null) {
			if (AmbUtile.checkCLuster(infoModel.getAmbulance()) && civilianCosts.size() > 0) {
				infoModel.getLog().info("Ultimately I choosed a target");
				infoModel.getAmbulance().target = civilianCosts.get(0).first();
				infoModel.getATEntity().getWork().setTarget(civilianCosts.get(0).first(), this);
			}
		}

	}

	public void findValidAmbulancesAndCivilians(boolean isCenter, AmbulanceTeam at) {

		freeAmbAgents.clear();
		civilians.clear();
		numberOfAmb = 0;

		for (AmbulanceTeam ambulance : infoModel.getModel().ambulanceTeams()) {
			if (!AmbUtile.validFreeAgent(ambulance))
				continue;

			freeAmbAgents.add(ambulance);
		}

		ALL.clear();
		ALL.addAll(infoModel.getModel().civilians());

		if (infoModel.getModel().time() < 90 || (infoModel.getModel().getLastAfterShockTime() > 1 && infoModel.getModel().time() < 210))
		{
			ALL.addAll(infoModel.getModel().agents());
		}

		// Checking if a Human is valid for deciding. If it has enough ambulances, it is not valid.

		for (Human civ : ALL) {

			ambulances.clear();

			if (civ == null) {
				infoModel.getLog().info("civ was not valid null "+ civ.getID());
				continue;
			}

			
			if (!civ.isDamageDefined()){
				infoModel.getLog().info("Damage "+ civ.getID() + civ.getDamage());
				
				continue;
			}
			
			if (!civ.isHPDefined()){
				infoModel.getLog().info("HP "+ civ.getID());
				continue;
			}
			
			if (!civ.isPositionDefined()){
				infoModel.getLog().info("Position "+ civ.getID());
				continue;
			}
				
			if (!civ.isBuriednessDefined()){
				infoModel.getLog().info("Bur "+ civ.getID());
				continue;
			}
			
			if (!civ.isAlive()){
				infoModel.getLog().info("Not Alive "+ civ.getID());
				continue;
			}
			

			
			if (!isCenter && !AmbUtile.isReallyReachable(civ, true)) {
				infoModel.getLog().info("civ was not valid not reachable" + civ.getID());
				continue;
			}

			if (civ.getBuriedness() == 0 && civ.getHP() == 10000) {
				infoModel.getLog().info("buriedness 0 HP 10000" + civ.getID());
				continue;
			}

			if (civ.getAreaPosition() instanceof Refuge) {
				infoModel.getLog().info("civ was not valid instanceof refuge" + civ.getID());
				continue;
			}

			if (civ.getHP() == 0) {
				infoModel.getLog().info("civ was not valid hp = 0" + civ.getID());
				continue;
			}

			if (!isCenter && civ.getRescueInfo().getIgnoredUntil() > infoModel.getTime()) {
				infoModel.getLog().info("ignore until" + civ.getID());
				continue;
			}

			if (!isCenter && infoModel.getTime() - civ.updatedtime() < 3) {
				infoModel.getLog().info("updated time " + civ.getID());
				continue;
			}

			if (civ instanceof Civilian && civ.getAreaPosition() instanceof Road && civ.model().refuges().isEmpty()) {
				infoModel.getLog().info("road  " + civ.getID());
				continue;
			}

			if (civ.getAreaPosition() instanceof Building && ((Building) civ.getAreaPosition()).getFieryness() > 0 && ((Building) civ.getAreaPosition()).getFieryness() < 4)
			{
				infoModel.getLog().info("in fire " + civ.getID());
				continue;
			}
			
			if (!(civ instanceof Civilian) && civ.getBuriedness() == 0){
				infoModel.getLog().info("not civ buriedness 0 " + civ.getID());

				continue;
			}
//
//			if (!AmbUtile.isValiadToDecide(civ, infoModel.getTime(), isCenter, infoModel.getATEntity()) || civ.getRescueInfo().getDeathTime() < infoModel.getTime() + 5) {
//				infoModel.getLog().warn("civ " + civ + "invalid");
//				continue;
//			}

			if (civ.getRescueInfo().getDeathTime() > 330){
				infoModel.getLog().info("Death time high " + civ.getID() + " " + civ.getRescueInfo().getDeathTime());
				continue;
			}

			
			

			
			if (infoModel.getTime() < 20 && !infoModel.getAgent().getMyClusterData().getClusterArea().contains(civ.getX(), civ.getY()))
			{
				if(civ.getAreaPosition() instanceof GasStation ){
					infoModel.getATEntity().getWork().setTarget(infoModel.getAmbulance().underMissionTarget(), this);
					infoModel.getLog().log("*****NEar GasStation*****");
				}
				infoModel.getLog().info("Civ not in my Cluster"+ civ.getID());
				continue;

			}
			
			
			
			
			
			// Skipping Fiery Civs. 

			for (AmbulanceTeam amb : infoModel.getModel().ambulanceTeams()) {

				if (amb.getWork() == null || amb.getWork().getTarget() == null) {
					continue;
				}

				else if (amb.getWork().getTarget().getID().getValue() == civ.getID().getValue() && amb.getWork().getCurrentState() == ATstates.RESCUE) {
					{
						ambulances.add(amb);
					}
				}
			}
			infoModel.getLog().info("amb size" + ambulances.size());

			if (infoModel.getModel().refuges().size() == 0) {
				infoModel.getLog().info("refuge 0");
				timeToRefuge = 0;
			}
			else

				timeToRefuge = civ.getRescueInfo().getTimeToRefuge();

			int deathTime = civ.getRescueInfo().getDeathTime();
			numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, civ, deathTime);

			if (ambulances.size() >= numberOfAmb) {
				infoModel.getLog().info("** SET : Civ has enough ATs. I don't have to decide for it" + "ambulances size" + ambulances.size() + "numberOfAmb" + numberOfAmb + "civ" + civ.getID() + "**");
				continue;
			}
			//AmbulanceDecision bestamb= new AmbulanceDecision();
			//bestamb.getBestAmbulanceForr(target);

			civilians.add(civ);
			infoModel.getLog().info("** SET : Civilian is valid for deciding. " + civ.getID());

		}

	}

	private void setAmbAndCivs() {

		Human myCiv = null;
		infoModel.getLog().info("** setAmbAndCiv **");
		choosed = false;

		for (Pair<Human, Integer> civil : civilianCosts) {

			if (choosed)
			{
				infoModel.getLog().info("** CIv is chosen. Let go." + myCiv);
				continue;
			}

			myCiv = civil.first();

			if (infoModel.getModel().refuges().size() == 0)
				timeToRefuge = 0;
			else
				timeToRefuge = civil.first().getRescueInfo().getTimeToRefuge();

			workingOn = 0;
			timeToResMe = 0;

			if (!AmbUtile.isValiadToDecide(civil.first(), infoModel.getTime(), false, infoModel.getATEntity()))
			{
				infoModel.getLog().info("** AmbCiv: Civ was not valid**");
				continue;
			}

			for (AmbulancesAndMeasurments amb : centerDecide.AmbulancesAndMeasur) {
				if (amb.getAmbulance().getWork() == null || amb.getAmbulance().getWork().getTarget() == null)
					continue;
				if (amb.getAmbulance().getWork().getTarget().getID().getValue() == civil.first().getID().getValue() && amb.getAmbulance().getWork().getCurrentState() == ATstates.RESCUE)
					workingOn++;
			}

			centerDecide.myAmbs.clear();

			for (AmbulancesAndMeasurments ambulance : centerDecide.AmbulancesAndMeasur) {

				moveDistance = infoModel.getAgent().model().getDistance(ambulance.getAmbulance().getAreaPosition(), myCiv.getAreaPosition());

				if (infoModel.getModel().refuges().size() == 0)
					timeToResMe = moveDistance + myCiv.getBuriedness();
				else
				{
					sos.base.move.Path path = infoModel.getAgent().move.getPathFromTo(java.util.Collections.singleton(myCiv.getAreaPosition()), infoModel.getModel().refuges(), StandardMove.class);
					int timeToRefuge = sos.tools.Utils.getSampleTimeToTarget(path);
					timeToResMe = moveDistance + myCiv.getBuriedness() + timeToRefuge;
				}

				centerDecide.IWillBeRes = timeToResMe;
				centerDecide.myAmbs.add(new Pair<AmbulanceTeam, Integer>(ambulance.getAmbulance(), centerDecide.IWillBeRes));
			}

			Collections.sort(centerDecide.myAmbs, new AmbComparator());

			if (centerDecide.myAmbs.size() == 0)
			{
				infoModel.getLog().info("** Traget has no amb. I'm choosing it: " + myCiv.getID());
				infoModel.getAmbulance().target = myCiv;
				infoModel.getATEntity().getWork().setTarget(myCiv, this);
				infoModel.getAmbulance().setUnderMissionTarget(null);
				choosed = true;
			}

			else {
				int deathTime = SimpleDeathTime.getEasyLifeTime(myCiv.getHP(), myCiv.getDamage(), infoModel.getTime());

				hasEnoughAT = false;
				for (int i = 1; !hasEnoughAT; i++) {
					if (timeToRefuge + 7 + myCiv.getBuriedness() / i < deathTime) {
						hasEnoughAT = true;
					}
					numberOfAmb = i;
				}

				for (int i = 0; i < numberOfAmb - workingOn; i++) {

					if (centerDecide.myAmbs.get(i).first().getAgent().getID().getValue() == infoModel.getAgent().getID().getValue()) {
						infoModel.getLog().info("** I'm in target's best ambs. Choosing it: " + myCiv.getID());
						infoModel.getAmbulance().target = civil.first();
						infoModel.getATEntity().getWork().setTarget(civil.first(), this);
						infoModel.getAmbulance().setUnderMissionTarget(null);
						choosed = true;
						return;
					}

					else
						choosed = false;

				}
			}

		}
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
	}

	
	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}


	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}


	@Override
	public void hear(String header, DataArrayList data,
			SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	
	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

	public void priorityOfTargets() {
		civilianCosts.clear();

		infoModel.getLog().info("** Priority of Targets **");

		targetCost = 0;
		//		mapClusters = AmbUtile.ClustersPriority(infoModel.getAmbulance());
		//		myClusterIndex = mapClusters.indexOf(myCluster);

		for (Human civil : civilians) {
			if (!AmbUtile.isValiadToDecide(civil, infoModel.getTime(), false, infoModel.getATEntity()))
				continue;

			for(Building fire : infoModel.getModel().buildings()){
				if(fire.getFieryness()>0){
					if(civil.getAreaPosition() == fire.getAreaPosition()){
						infoModel.getLog().log("fiery place eyy babaa");
						continue;
					}
					
				}
			}
			if (civil.getRescueInfo().getIgnoredUntil() > infoModel.getTime())
				continue;

			// TODO TESTTTTTTTTTTTTTTTTTTTTTTTTTTTT

			//			 targetCost = (((civil.getBuriedness() + 1) * moveDistance + civil.getHP())) / (civil
			//					.getDamage() + 1);

			//			myCivClusterIndex = mapClusters.indexOf(civil.model().searchWorldModel.getClusterData(civil));
			if (infoModel.getModel().refuges().size() == 0) {
				infoModel.getLog().info("refuge 0");
				timeToRefuge = 0;
			}
			else

				timeToRefuge = civil.getRescueInfo().getTimeToRefuge();

			numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, civil, civil.getRescueInfo().getDeathTime());
			moveDistance = infoModel.getModel().getDistance(infoModel.getATEntity().getPositionArea(), civil.getAreaPosition());
			targetCost = numberOfAmb / (AmbUtile.numberOfFreeAmbs(infoModel.getATEntity()) + 1) + civil.getBuriedness() + (civil.getHP() * civil.getRescueInfo().getDeathTime()) + moveDistance;//

			// TODO TESTTTTTTTTTTTTTTTTTTTTTTTTTTTT

			//			if (myCivClusterIndex < myClusterIndex)
			//				targetCost /= myClusterIndex - myCivClusterIndex;
			//
			//			else if (myCivClusterIndex > myClusterIndex)
			//				targetCost *= mapClusters.size() - (myCivClusterIndex - myClusterIndex + 1);

			if (!(civil instanceof Civilian) && (infoModel.getTime() < 80 || infoModel.getModel().getLastAfterShockTime() > 1))
				targetCost /= 3f;
			

			civilianCosts.add(new Pair<Human, Integer>(civil, targetCost));

			infoModel.getLog().info("** PriorityOfTarget : " + civil.getID() + "Cost:" + targetCost + "**");
		}

		Collections.sort(civilianCosts, new CostComparator2());

	}


}

class AmbComparator implements Comparator<Pair<AmbulanceTeam, Integer>> {

	// return -1 : go to the top. 

	@Override
	public int compare(Pair<AmbulanceTeam, Integer> o1,
			Pair<AmbulanceTeam, Integer> o2) {
		if (o1.second() < o2.second())
			return -1;
		else if (o1.second() > o2.second())
			return 1;
		return 0;
	}

}

class CostComparator implements Comparator<Pair<Civilian, Integer>> {
	@Override
	public int compare(Pair<Civilian, Integer> h1, Pair<Civilian, Integer> h2) {// Target
		// cost
		int cost1 = h1.second();
		int cost2 = h2.second();

		if (cost1 < cost2 ||
				(cost1 == cost2 && (h1.first().getID().getValue() > h2.first().getID().getValue())))
			return -1;
		else if (h1.second() > h2.second())
			return 1;
		else
			return 1;

	}
}


class CostComparator2 implements Comparator<Pair<Human, Integer>> {
	@Override
	public int compare(Pair<Human, Integer> h1, Pair<Human, Integer> h2) {// Target
		// cost
		int cost1 = h1.second();
		int cost2 = h2.second();

		if (cost1 < cost2 ||
				(cost1 == cost2 && (h1.first().getID().getValue() > h2.first().getID().getValue())))
			return -1;
		else if (h1.second() > h2.second())
			return 1;
		else
			return 1;

	}
}
