package sos.ambulance_v2.decision.states;

import java.util.List;

import sos.ambulance_v2.ATTask;
import sos.ambulance_v2.AmbulanceInformationModel;
import sos.ambulance_v2.base.RescueInfo.IgnoreReason;
import sos.ambulance_v2.decision.AmbUtile;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.StandardEntity;
import sos.base.entities.VirtualCivilian;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.util.SOSActionException;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.decisionMaker.implementations.targets.AreaTarget;
import sos.tools.decisionMaker.implementations.tasks.StandardMoveToAreaTask;

public class VIrtualRescueState extends SOSIState<AmbulanceInformationModel>{
	private SOSTask<?> task = null;
	public VIrtualRescueState(AmbulanceInformationModel infoModel) {
		super(infoModel);
		// TODO Auto-generated constructor stub
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		
		if(infoModel.getAmbulance().virtualTarget == null || infoModel.getAmbulance().virtualTarget.getDeathTime() < infoModel.getTime() + 5)
			return null;
		infoModel.getLog().info("//VIrtual Rescue State\\");
		move();
		return task;
	}
	

	private void move() {
		
		VirtualCivilian virtualCiv = infoModel.getAmbulance().virtualTarget;
		if(infoModel.getATEntity().getPosition() != virtualCiv.getPosition()){
			infoModel.getLog().info("I am moving to my virtual target");
			infoModel.getAmbulance().threelast_task = infoModel.getAmbulance().twolast_task;
			infoModel.getAmbulance().twolast_task = infoModel.getAmbulance().last_task;
			infoModel.getAmbulance().last_task = ATTask.MoveTask;
			task = new StandardMoveToAreaTask(new AreaTarget(virtualCiv.getPosition()), infoModel.getTime());		

		}
		
		else
			
		{
//			for(Human civil : infoModel.getModel().civilians()){
//				
//				if(!AmbUtile.isValiadToDecide(civil, infoModel.getTime(), false,infoModel.getATEntity()) || virtualCiv == null)
//					continue;
//				infoModel.getLog().info("civilian position:" + civil.getPosition().getID().getValue() + "virt position " + virtualCiv.getPosition().getID().getValue());
//				if(civil.getPosition().getID().getValue() == virtualCiv.getPosition().getID().getValue()){
//					infoModel.getLog().info("puting virt target into real target's position");
//					infoModel.getAmbulance().target = civil;
//					continue;
//				}
//			}
			
			boolean done = false;
			for(Human civil : infoModel.getAmbulance().getVisibleEntities(Civilian.class)){
				if (done)
					continue;
				if(isValiadToDecideLow(civil, infoModel.getTime(), false, infoModel.getATEntity())){
					infoModel.getAmbulance().target = civil;
					infoModel.getATEntity().getWork().setTarget(civil, this);
					infoModel.getAmbulance().setUnderMissionTarget(civil);
					infoModel.getLog().info("setting virt in real" + civil.getID());
					done = true;
					continue;
				}
				
			}
			
			task = null;
			infoModel.getAmbulance().virtualTarget = null;
		}
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub
	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}
	
	public boolean isValiadToDecideLow(Human civ, int time, boolean isCenter, AmbulanceTeam amb) {

		if (civ == null) {
			infoModel.getLog().info("civ was not valid null");
			return false;
		}
		
		if (!civ.isDamageDefined() || !civ.isHPDefined() || !civ.isPositionDefined() || !civ.isBuriednessDefined()) {
			infoModel.getLog().info("civ was not valid damage or hp or pos");
			return false;
		}
		if (!isCenter && !AmbUtile.isReallyReachable(civ, true)) {
			infoModel.getLog().info("civ was not valid not reachable" + civ.getID());
			return false;
		}
		if (civ.getBuriedness() == 0 && civ.getHP() == 10000) {
			infoModel.getLog().info("civ was not valid buridness 0 and hp 1000" + civ.getID());
			return false;
		}
		if (civ.getAreaPosition() instanceof Refuge) {
			infoModel.getLog().info("civ was not valid instanceof refuge" + civ.getID());
			return false;
		}
//		if (!isCenter && civ.getPosition() instanceof AmbulanceTeam) {
//			//	logUtile.log("civ was not valid instanceof ambulance" + civ.getID());
//			return false;
//		}

		if (civ.getHP() == 0) {
			infoModel.getLog().info("civ was not valid hp = 0" + civ.getID());
			return false;
		}

		if (civ.getRescueInfo().getIgnoredUntil() > time && civ.getRescueInfo().getIgnoreReason() == IgnoreReason.WillDie) {
			infoModel.getLog().info("civ was not valid ignored" + civ.getID());
			return false;
		}


		return true;
	}

}
