package sos.fire_v2;

import java.awt.Font;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import javax.swing.JTextArea;

import rescuecore2.geometry.GeometryTools2D;
import rescuecore2.worldmodel.EntityID;
import sos.base.SOSConstant;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.FireBrigade;
import sos.base.entities.Refuge;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.util.SOSActionException;
import sos.base.util.SOSGeometryTools;
import sos.base.util.geom.ShapeInArea;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.base.util.sosLogger.SOSLoggerSystem.OutputType;
import sos.fire_v2.Saba_States.Decide;
import sos.fire_v2.Saba_States.LastTask;
import sos.fire_v2.Saba_States.SFireZone;
import sos.fire_v2.Saba_States.Sort;
import sos.fire_v2.Saba_States.sabaMove;
import sos.fire_v2.base.AbstractFireBrigadeAgent;
import sos.fire_v2.base.worldmodel.FirePrecompute;
import sos.fire_v2.decision.FBFeedbackFactory;
import sos.fire_v2.decision.FireDecisionMaker;
import sos.fire_v2.decision.FireInformationModel;
import sos.fire_v2.decision.tasks.ExtinguishTask;
import sos.fire_v2.position.Positioing;
import sos.search_v2.agentSearch.FireBrigadeSearch;
import sos.search_v2.tools.cluster.BuridBlockSearchCluster;
import sos.search_v2.worldModel.SearchWorldModel;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.implementations.targets.AreaTarget;
import sos.tools.decisionMaker.implementations.tasks.StandardMoveToAreaTask;

public class FireBrigadeAgent extends AbstractFireBrigadeAgent {

	public FireDecisionMaker FDK;
	Priority p;
	EntityID currentBuilding;
	public Sort sorted;
	public Building searchBuildingTarget = null;
	public LastTask last_task = LastTask.NONE;
	public int lastSearchCycle = 0;

	/***** SABA *****/
	public SFireZone MyZone = null;
	public SFireZone lastZone = null;
	public Boolean TheresFireInMyCluster = false;
	public Boolean ThereWasFireInMyCluster = false;
	public ArrayList<Building> SearchfireZoneSide = null;
	public ArrayList<Building> SearchIntrruptSide = null;
	public Building IntrruptBuilding = null;
	public SOSLoggerSystem Saba_Log;
	public Decide decide;
	public Building FoundFireBuildingInMyCluster = null;

	public JTextArea mySelectHistory = new JTextArea() {
		private static final long serialVersionUID = 1L;

		@Override
		public void append(String str) {
			if (!SOSConstant.IS_CHALLENGE_RUNNING)
				super.append(str);
		};
	};

	private void newThink() throws SOSActionException {
		FDK.lastAct = "NONE";
		FDK.lastState = "NONE";
		SOSTask<?> decide = FDK.decide();
		if (decide != null) {
			FDK.lastAct = decide.getClass().getSimpleName();
			decide.execute(this);
		}
		FDK.lastAct = "SEARCH";
//		search();
		randomWalk();
	}

	public Positioing positioning;

	@Override
	protected void prepareForThink() {
		super.prepareForThink();
		for (Building b : getVisibleEntities(Building.class)) {
			b.setChecked(true);
		}
		FoundFireBuildingInMyCluster = null;
		ThereWasFireInMyCluster = TheresFireInMyCluster;
		TheresFireInMyCluster = false;
		for (Building b : getMyClusterData().getBuildings()) {
			if (!isValid(b))
				continue;
			
			Saba_Log.info("FireBrigadeAgent: Building " + b.getID() + " was in my cluster and was on fire so TheresFireInMyCluster becomes true. ");
			TheresFireInMyCluster = true;
			FoundFireBuildingInMyCluster = b;
			break;

		}
		
	}

	public boolean isValid(Building b) {
		
		
		
		if(b instanceof Refuge){
			return false;
		}
		
		if (b.getTemperature() < 20) {
		//	Saba_Log.info("FBA: Temp less than 20. Skip");
			return false;
		}

		if (b.getTemperature() < 30 && b.getFieryness() > 3)
		{
		//	Saba_Log.info("FBA: Temp less than 30 and fireyness more than 3. Skip");
			return false;
		}
//		if (b.getFieryness() <= 8 && b.getFieryness() > 3) {
//		//	Saba_Log.info("FBA: Burned. Skip");
//			return false;
//		}
		if(b.isIgnored && time() -(b.ExtinguishEnded+b.ExtinguishStarted) >10){
			Saba_Log.info("FBA: Building "+ b.getID() + "was skipped more than 10 cycles ago. False the isIgnored. ");
			b.isIgnored = false;
		}
		if(b.ExtinguishEnded!=0 || b.ExtinguishStarted!=0){
		Saba_Log.info("FBA: isValid was called for building " + b.getID());
		Saba_Log.info("FBA: Extinguish Ended: " + b.ExtinguishEnded);
		Saba_Log.info("FBA: Extinguish Started: " + b.ExtinguishStarted);
		}
		if(!getVisibleEntities(Building.class).contains(b) && b.ExtinguishEnded>10){
			Saba_Log.info("FBA: Building wasn't in visible entity and it has been more than 10 cycles that it's being extinguished. Skip");
			b.isIgnored=true;
			return false;
		}
		
		
		return true;
	}

	@Override
	protected void think() throws SOSActionException {
		super.think();

		//	p.choose(this);

		//Exercise	
		//	if(model().getFireBuildings().isEmpty()
		//		search();
		//	else{
		//		currentBuilding = model().getFireBuildings().get(0).building().getID();
		//		extinguish( currentBuilding , maxPower);
		//		
		//	}
		CreatMapsFireZones(); //TODO change place to think
		newThink();
		/*
		 * Building b=model().buildings().get(0);
		 * Building bu=model().buildings().get(1);
		 * for(Building b2 : model().buildings() )
		 * {
		 * if(b2.getID().getValue() == 19571)
		 * b=b2;
		 * else if(b2.getID().getValue() == 27209 )
		 * bu=b2;
		 * }
		 * if(bu.getTemperature()>18)
		 * extinguish(b, maxPower);
		 * rest();
		 */
	}

	public void log(String st) {
		sosLogger.agent.info(st);
	}

	@Override
	protected void finalizeThink() {
		super.finalizeThink();
	}

	@Override
	protected void preCompute() {
		super.preCompute();
		Saba_Log = new SOSLoggerSystem(me(), "SabaFire/", true, OutputType.File, true);
		sosLogger.addToAllLogType(Saba_Log);
		p = new Priority();
		mySelectHistory.setFont(new Font("serif", Font.ITALIC, 12));
		maxWater = config.getIntValue(MAX_WATER_KEY);
		maxDistance = config.getIntValue(MAX_DISTANCE_KEY);
		maxPower = config.getIntValue(MAX_POWER_KEY);
		viewDistance = config.getIntValue(VIEW_DISTANCE);

		model().setOwner(this);

		FirePrecompute preCompute = new FirePrecompute(model());

		model().setMapSides();
		model().setSideIslands();
		preCompute.execute();

		positioning = new Positioing(this, model());

		log("MAP NAME ::::> " + getMapInfo().getRealMapName());

		/////SEARCH////////////////////////
		SearchWorldModel<FireBrigade> ss = new SearchWorldModel<FireBrigade>(model());
		model().searchWorldModel = ss;
		newSearch = new FireBrigadeSearch<FireBrigade>(this, ss, new BuridBlockSearchCluster<FireBrigade>(this, model().fireBrigades()));

		////////////SELECT//////
		model().estimatedModel.world = model();
		FDK = new FireDecisionMaker(this, new FBFeedbackFactory());
		sorted = new Sort(FDK.getInfoModel());
		decide = new Decide(FDK.getInfoModel());
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		super.hear(header, data, dynamicBitArray, sender, channel);
	}

	@Override
	protected void thinkAfterExceptionOccured() throws SOSActionException {
//		try {
//			// search();
//		} catch (SOSActionException e) {
//			throw e;
//		} catch (Exception e) {
//			e.printStackTrace();
//		}
		randomWalk(true);
	}

	//***********SABA***********\

	public int myPlaceInGroup() {
		ArrayList<FireBrigade> firebirgadesInTheSameCluster = getVisibleEntities(FireBrigade.class);
		if (firebirgadesInTheSameCluster.isEmpty())
			return -1;

		firebirgadesInTheSameCluster.add(me());
		sorted.sortID(firebirgadesInTheSameCluster);

		return firebirgadesInTheSameCluster.indexOf(me());

	}

	public Vector<SFireZone> MapsAllFireZones = new Vector<SFireZone>();

	public void CreatMapsFireZones() {
		MapsAllFireZones.clear();
		sorted.setBuildings();
		for (Building b : sorted.validBuildings) {

			if (checkRepeated(b))
				continue;
			if (!isValid(b))
				continue;

			SFireZone creatingFireZone = new SFireZone(this);
			creatingFireZone.FireZoneBuildings.add(b);
			creatingFireZone.addBuildingToFireZone(b);			
			creatingFireZone.setFireZoneNeighbours();
			creatingFireZone.setSize();

			for (Building b2 : creatingFireZone.FireZoneBuildings){
				b2.MyFireZone = creatingFireZone;
			}
			
			MapsAllFireZones.add(creatingFireZone);

		}
		sorted.callFireZoneWeight(MapsAllFireZones);
		Collections.sort(MapsAllFireZones, new FireZoneWeightComparator());

	}

	public boolean checkRepeated(Building b) {

		for (SFireZone f1 : MapsAllFireZones){
			if (f1.FireZoneBuildings.contains(b))
				return true;
			if(f1.others.contains(b))
				return true;
		}
		return false;
	}

	public SOSTask<?> GoToMax(Building b, FireInformationModel infoModel) {
		for (Area r : b.getRoadsInSight()) {
			if (SOSGeometryTools.distance(r, infoModel.getLastSelectedBuilding())
			> maxDistance)
				continue;

			last_task = LastTask.sabaMove;
			infoModel.Saba_Log.info("-----"+r.getID());
			return new StandardMoveToAreaTask(new AreaTarget(r), infoModel.getTime());
		}
		return null;
	}

	public SOSTask<?> canExtinguish(FireInformationModel infoModel) {

		infoModel.Saba_Log.info("FBA: Extinguish >_< ");
		//		if(!( IntrruptBuilding != null &&  IntrruptBuilding.getID() == infoModel.getLastSelectedBuilding().getID()))
		//				updateFireZone();
		last_task = LastTask.ExtinguishTask;
		
		if(infoModel.getLastSelectedBuilding().ExtinguishStarted ==0)
			infoModel.getLastSelectedBuilding().ExtinguishStarted = infoModel.getTime();
		infoModel.getLastSelectedBuilding().ExtinguishEnded++;
		infoModel.getLog().info(infoModel.getLastSelectedBuilding().ExtinguishEnded);
		return new ExtinguishTask(infoModel.getLastSelectedBuilding(), null, infoModel.getTime());

	}

	public SOSTask<?> correctExtinguish(Building target, FireInformationModel infoModel) {

		double d = GeometryTools2D.getDistance(me().getPositionPoint(), target.getPositionPoint());
		if (getVisibleEntities(Building.class).contains(target)) {

			if (d <= (maxDistance)) {

				Building b = null;

				if (infoModel.getEntity().getAreaPosition() instanceof Building) {
					b = (Building) (infoModel.getEntity().getAreaPosition());
				}

				if (b != null && b.getFieryness() > 0 && b.getFieryness() < 4) {
					infoModel.Saba_Log.info("Extinguish: In visible and in max, GoToMax was called!");
					GoToMax(b, infoModel);
				}
				infoModel.Saba_Log.info("Extinguish: In visible and in max, canExtinguish was called!");
				return canExtinguish(infoModel);
			}
			//TOO VISIBLE HASTI TOO MAX NISTI!
			infoModel.Saba_Log.info("Extinguish: In visible and not in max, GoToMax was called!");
			GoToMax(target, infoModel);

		}

		//TOO VISIBLE NISTI

		if (d < maxDistance) {
			//TOO MAX HAST

			for (ShapeInArea n1 : target.getSearchAreas()) {
				//REACHABLE HAST
				Area n = n1.getArea();
				if (n.isReallyReachable(true)) {
					infoModel.Saba_Log.info("Extinguish: Not in visible and in max, one of the neighbours was reachable so move to visible!");
					return new StandardMoveToAreaTask(new AreaTarget(n), infoModel.getTime());

				}
			}

			//REACHABLE NIST
			infoModel.Saba_Log.info("Extinguish: Not in visible and in max, none was reachable so EXTINGUISH! (>.<) ");
			return canExtinguish(infoModel);
		}
		
		infoModel.Saba_Log.info("Extinguish: Not in visible and not in max, move to target!");
		last_task=LastTask.sabaMove;
		return new sabaMove(null, infoModel.getTime(), target);

	}

}

class FireZoneWeightComparator implements Comparator<SFireZone> {

	@Override
	public int compare(SFireZone f1, SFireZone f2) {
		if (f1.getFireZoneWeight() > f2.getFireZoneWeight())
			return -1;

		if (f1.getFireZoneWeight() < f2.getFireZoneWeight())
			return 1;

		return 0;

	}
}
