package sos.fire_v2.Saba_States;

import java.util.List;

import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageXmlConstant;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.blocks.MessageBlock;
import sos.base.message.structure.channel.Channel;
import sos.base.move.types.StandardMove;
import sos.base.util.SOSActionException;
import sos.fire_v2.decision.FireInformationModel;
import sos.fire_v2.decision.tasks.FireRestTask;
import sos.fire_v2.decision.tasks.goToRefuge;
import sos.tools.Utils;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;


public class IAmHurtState extends SOSIState<FireInformationModel>  {
	
	private static final int MIN_DEATH_TIME = 6;
	private static final int MIN_HP = 3000;
	private static final int MAX_DAMAGE = 100;
	private static final int MIN_TIME_TO_RESCUE = 3;
	private int deathTime;
	int timeToRefuge = 6;

	public IAmHurtState(FireInformationModel infoModel) {
		super(infoModel);
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		
		
		if(!isValid()){
			infoModel.getLog().info("$$$$$ Skipped from IAmHurtState $$$$$");
			return null;
		}
		
		 if(infoModel.getModel().refuges().isEmpty())
			 return null;
		 
			 
		 timeToRefuge = getTimeToNearestRefuge();
		if( deathTime > (timeToRefuge+infoModel.getTime()+15) ){
			
			if (infoModel.self().model().refuges().isEmpty())
			{
				infoModel.Saba_Log.info("No Water: No refuge map. ");
				infoModel.self().last_task = LastTask.goToHydrant;
//				if(!(infoModel.self().me().getAreaPosition() instanceof Hydrant)){
//					return new goToHydrant(null, infoModel.getTime(),infoModel);
//				}
				return new FireRestTask(null,infoModel.getTime());
			}

			else {
				infoModel.Saba_Log.info("Going to refuge because of no water");
				infoModel.self().last_task = LastTask.goToRefuge;
				if (!(infoModel.self().me().getAreaPosition() instanceof Refuge)) {

					return new goToRefuge(null, infoModel.getTime());
				}
				return new FireRestTask(null, infoModel.getTime());
			}}
		if(infoModel.self().last_task == LastTask.SabaSearchTask){
			if (infoModel.self().model().refuges().size() == 0)
			{
				infoModel.Saba_Log.info("No Water: No refuge map. ");
				infoModel.self().last_task = LastTask.goToHydrant;
//				if(!(infoModel.self().me().getAreaPosition() instanceof Hydrant)){
//					return new goToHydrant(null, infoModel.getTime(),infoModel);
//				}
				return new FireRestTask(null,infoModel.getTime());
			}

			else {
				infoModel.Saba_Log.info("Going to refuge because of no water");
				infoModel.self().last_task = LastTask.goToRefuge;
				if (!(infoModel.self().me().getAreaPosition() instanceof Refuge)) {

					return new goToRefuge(null, infoModel.getTime());
				}
				return new FireRestTask(null, infoModel.getTime());
			}
			
			
			
			
	}
			return null;
	}

	public boolean isValid() {


		if( iWillDieInNextCycles()){
			infoModel.getLog().info("agent is going to die");
        	infoModel.self().messageBlock=new MessageBlock(MessageXmlConstant.HEADER_DEAD_AGENT);
        	infoModel.self().messageBlock.addData(MessageXmlConstant.DATA_AGENT_INDEX,((Human)infoModel.self().me()).getAgentIndex());
        	infoModel.self().messages.add(infoModel.self().messageBlock);
			return true;
		}
		else if(infoModel.self().me().getBuriedness() > 0)
			return true;
		else if(infoModel.self().me().getDamage() ==0 )
			return false;
		else if ((deathTime -  infoModel.getTime()) < MIN_DEATH_TIME || (infoModel.self().me().getHP()/infoModel.self().me().getDamage() < MIN_HP/MAX_DAMAGE))
			return true;
		else if (deathTime -  infoModel.getTime() > 20 && infoModel.self().me().getDamage() > 30)
			return true;
		else if ( deathTime - infoModel.getTime() < 20 )
			return true;
		else 
			return false;
	}
	
	private boolean iWillDieInNextCycles(){
		deathTime = ((Human)infoModel.self().me()).getRescueInfo().getDeathTime();
		int aliveTime = deathTime-infoModel.getTime();
		int rescueTime = (infoModel.getModel().refuges().isEmpty())?1000 : getTimeToNearestRefuge()+infoModel.self().me().getBuriedness();
		if(aliveTime < 4 && aliveTime < rescueTime)
			return true;
		return false;
	}
	
	

	@Override
	public void giveFeedbacks(List feedbacks) {
	}

	@Override
	public void skipped() {

	}

	@Override
	public void overTaken() {

	}


	public int getTimeToNearestRefuge() {
		sos.base.move.Path path = infoModel.getAgent().move.getPathFromTo(java.util.Collections.singleton(infoModel.getEntity().getAreaPosition()), infoModel.getModel().refuges(), StandardMove.class);
		return Utils.getSampleTimeToTarget(path);
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public String getName() {
		return this.getClass().getName();
	}
	
	
}
