package sos.police_v2;

import java.awt.Point;
import java.awt.Shape;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Vector;

import rescuecore2.misc.Pair;
import rescuecore2.worldmodel.EntityID;
import sos.ambulance_v2.base.AmbulanceConstants.ATstates;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Blockade;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.PoliceForce;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.types.PoliceMove;
import sos.base.util.SOSActionException;
import sos.base.util.TimeNamayangar;
import sos.police_v2.base.AbstractPoliceForceAgent;
import sos.police_v2.base.clearablePointToReachable.GeoClearPointReachablity;
import sos.police_v2.base.clearablePointToReachable.GeoDegreeClearPointToReachable;
import sos.police_v2.interupts.HeavyStockedHandlerState;
import sos.police_v2.interupts.PoliceAbstractIntruptState;
import sos.police_v2.interupts.ReachableCivilianState;
import sos.police_v2.interupts.ReachableRefOrAgentState;
import sos.police_v2.interupts.StockHandlerState;
import sos.police_v2.preCompute.PoliceForceTask;
import sos.police_v2.preCompute.Task;
import sos.police_v2.states.AfterShockAgentsState;
import sos.police_v2.states.AfterShockAmbulanceState;
import sos.police_v2.states.AfterShockRefugeState;
import sos.police_v2.states.CiriticalCivilianReachableity;
import sos.police_v2.states.CiviReachableState;
import sos.police_v2.states.CivilReachable;
import sos.police_v2.states.ClearHighwayState;
import sos.police_v2.states.ConnectAgentState;
import sos.police_v2.states.DangerCivilian;
import sos.police_v2.states.FireReachableState;
import sos.police_v2.states.FirstCheckState;
import sos.police_v2.states.HydrantState;
import sos.police_v2.states.MeStuck;
import sos.police_v2.states.OpenCIviliansWIthoutBurridness;
import sos.police_v2.states.OpenClusterFire;
import sos.police_v2.states.OpenStuckAgent;
import sos.police_v2.states.PoliceAbstractState;
import sos.police_v2.states.PoliceSearch;
import sos.police_v2.states.RefugeState;
import sos.police_v2.states.VisibleHydrantState;
import sos.police_v2.tool.FireZoneManager;
import sos.tools.decisionMaker.definitions.commands.SOSTask;

/**
 * SOS police force agent.
 */
public class PoliceForceAgent extends AbstractPoliceForceAgent {
	public boolean BeginCivReach = false;
	private ArrayList<PoliceAbstractState> states = new ArrayList<PoliceAbstractState>();
	private ArrayList<PoliceAbstractIntruptState> intruptStates = new ArrayList<PoliceAbstractIntruptState>();
	private HashMap<AmbulanceTeam, Human> ambulanceTeamTarget;
	public Blockade lastTarget = null;
	//public Precompute precompute;
	public HashMap<PoliceForceTask, ArrayList<Task<? extends StandardEntity>>> agent_task = new HashMap<PoliceForceTask, ArrayList<Task<? extends StandardEntity>>>();
	public boolean precomputeflag = false;
	public Civilian assignCivil = null;
	public Shape myClusterShape;
	SOSTask currentTask;
	public boolean preIsDone = false;
	public Vector buildzone;
	HashSet mycluster;
	int policeInMyRoadZone;
	public Vector<Road> RoadsZone;
	private FireZoneManager fireZoneManager;
	private int lastShockTime = 0;
	public int hydrantCounter = 0;
	public Building searchBuildingTarget;
	public int MaxHydrantInt = 0;
	public FireReachableState fireReachableState;
	public ConnectAgentState connectAgentState;
	private Pair<Integer, Integer> numberOfCiviliansOpening;

	//clusterdata.getbuilding
	@Override
	protected void preCompute() {
		super.preCompute();// it's better to call first!
		PoliceConstants.STANDARD_OF_MAP = (int) model().getBounds().getWidth() / 200000;
		policeInMyRoadZone = 0;
		log.debug("Standard of Map:", PoliceConstants.STANDARD_OF_MAP);
		TimeNamayangar tm = new TimeNamayangar("Police PreCompute");
		model().makeIslands();
		model().makeHighways();
		clearableToPoint = new GeoClearPointReachablity(this);
		degreeClearableToPoint = new GeoDegreeClearPointToReachable(this);
		setNumberOfCiviliansOpening(new Pair<Integer, Integer>(0, 0));
		setLastShockTime(0);
		//TODO

		doStatePrecomputes();
		sosLogger.base.consoleInfo("PolicePreCompute " + tm.stop());
		ambulanceTeamTarget = new HashMap<AmbulanceTeam, Human>();
		clearWidth = getConfig().getIntValue("clear.repair.rad", 1250);
		myClusterShape = model().searchWorldModel.getClusterData().getConvexShape();

		buildzone = new Vector<Building>();
		mycluster = newSearch.getSearchWorld().getMapClusterType().getMyClusterBuildings();
		RoadsZone = makeRoadsZone(newSearch.getSearchWorld().getClusterData().getBuildings());
		setFireZoneManager(new FireZoneManager(model(), RoadsZone, this));
		makeStates();
		makeIntruptsStates();

	}

	private void makeStates() {

		states.add(new AfterShockAmbulanceState(this));
		states.add(new AfterShockRefugeState(this));
		states.add(new AfterShockAgentsState(this));

		states.add(new HydrantState(this));
		connectAgentState = new ConnectAgentState(this);
		states.add(connectAgentState);
		states.add(new CiriticalCivilianReachableity(this));

		states.add(new OpenStuckAgent(this));
		states.add(new ClearHighwayState(this));
		states.add(new OpenClusterFire(this));
		fireReachableState = new FireReachableState(this);
		states.add(fireReachableState);

		states.add(new CiviReachableState(this));
		//states.add(new FireSearchState(this));
		states.add(new OpenCIviliansWIthoutBurridness(this));
		states.add(new RefugeState(this));
		states.add(new VisibleHydrantState(this));
		
		states.add(new PoliceSearch(this));
		states.add(new DangerCivilian(this));
		

	}

	private void makeIntruptsStates() {
		//TODO
		intruptStates.add(new MeStuck(this));
		intruptStates.add(new StockHandlerState(this));
		intruptStates.add(new HeavyStockedHandlerState(this));
		intruptStates.add(new FirstCheckState(this));
		intruptStates.add(new ReachableRefOrAgentState(this));
		intruptStates.add(new ReachableCivilianState(this));
		intruptStates.add(new CivilReachable(this));

	}

	private void doStatePrecomputes() {
		ArrayList<PoliceAbstractState> allStates = new ArrayList<PoliceAbstractState>(intruptStates);
		allStates.addAll(states);
		for (PoliceAbstractState state : allStates) {
			long start = System.currentTimeMillis();
			state.precompute();
			long time = (System.currentTimeMillis() - start);
			boolean consoleLog = time > 10 || !(state instanceof PoliceAbstractIntruptState);
			log.info(state + " got:" + time + "ms", consoleLog);
		}

	}

	@Override
	protected void prepareForThink() {
		super.prepareForThink();
		if (time() == 100 || time() == 150 || time() == 200 || time() == 250)
			setNumberOfCiviliansOpening(new Pair<Integer, Integer>(0, time()));
		for (Building b : getVisibleEntities(Building.class)) {
			b.setChecked(true);
		}
	}

	@Override
	protected void think() throws SOSActionException {
		super.think();

		doIntrupt();
		doStateAct();
		walk();

	}

	public int myPlaceInGroup() {
		ArrayList<PoliceForce> AmbulanceTeamsInTheSameCluster = getVisibleEntities(PoliceForce.class);
		if (AmbulanceTeamsInTheSameCluster.isEmpty())
			return -1;

		AmbulanceTeamsInTheSameCluster.add(me());
		Collections.sort(AmbulanceTeamsInTheSameCluster, new IdComparator());

		return AmbulanceTeamsInTheSameCluster.indexOf(me());

	}

	private void doBugState() throws SOSActionException {
		//		try {
		//			getState(SearchState.class).act();
		//		} catch (SOSActionException e1) {
		//			throw e1;
		//		} catch (Exception e) {
		//			sosLogger.agent.fatal(e);
		//		}

		walk();
		log.warn("NO ACT");
		problemRest("Exception occured");
	}

	private void walk() throws SOSActionException {
		PriorityQueue<Blockade> blockadesInRange = model().getBlockadesInRange(me().getX(), me().getY(), clearDistance);
		Blockade selectedBlock = null;
		if (blockadesInRange.size() > 0)
			selectedBlock = blockadesInRange.remove();
		if (selectedBlock != null) {
			log.info("clear");
			clear(new Point(selectedBlock.getX(), selectedBlock.getY()));
		}
		ArrayList<Area> dest = new ArrayList<Area>();
		for (Blockade blockade : model().blockades()) {
			dest.add(blockade.getAreaPosition());
			log.info("dest" + blockade.getID());
		}
		if (dest.size() > 0) {
			log.info("move in think" + dest.size());
			move.move(dest, PoliceMove.class);
		}

	}

	private void doStateAct() throws SOSActionException {
		TimeNamayangar tm = new TimeNamayangar();
		for (PoliceAbstractState lastState : states) {
			log.info("^^In State: " + lastState);
			try {
				tm.reset();
				tm.start();
				super.lastState = lastState + "";
				lastState.act();
				tm.finish();
				sosLogger.act.debug(lastState + " do nothing but take " + tm);
			} catch (SOSActionException e) {
				sosLogger.act.debug(lastState + "doing " + e.getMessage() + " take " + tm);
				throw e;
			} catch (Exception e) {
				sosLogger.error("State:" + lastState + " failed...");
				sosLogger.error(e);
			}
			log.info("state" + lastState + " finished^^");
		}
	}

	private void doIntrupt() throws SOSActionException {
		for (PoliceAbstractIntruptState intruptState : intruptStates) {
			lastState = intruptState.getClass().getSimpleName();
			try {
				log.trace(intruptState + " ^^Interuprt Eligibility?");
				if (intruptState.canMakeIntrupt()) {
					log.trace(intruptState + " ^^TRUE");
					intruptState.act();
				}
			} catch (SOSActionException e) {
				sosLogger.act.debug(lastState + "doing " + e.getMessage());
				throw e;
			} catch (Exception e) {
				sosLogger.error("State:" + lastState + " failed...");
				sosLogger.error(e);
			}
		}
	}

	@Override
	protected void thinkAfterExceptionOccured() throws SOSActionException {
		log.warn("An unhandeled exception occured.... act as bug state");
		doBugState();
	}

	@Override
	protected void finalizeThink() {
		super.finalizeThink();
	}

	@SuppressWarnings("unchecked")
	public <T extends PoliceAbstractState> T getState(Class<T> stateClass) {
		for (PoliceAbstractState state : states) {
			if (stateClass.isInstance(state))
				return (T) state;
		}
		return null;
	}

	public HashMap<AmbulanceTeam, Human> getAmbulanceTeamTarget() {
		return ambulanceTeamTarget;
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		super.hear(header, data, dynamicBitArray, sender, channel);
		for (PoliceAbstractState state : states) {
			state.hear(header, data, dynamicBitArray, sender, channel);
		}

		if (header.equalsIgnoreCase(HEADER_AMBULANCE_INFO)) {
			ATstates state = convertStateIndexToState(data.get(DATA_AT_STATE));
			AmbulanceTeam at = model().ambulanceTeams().get(data.get(DATA_AMBULANCE_INDEX));
			if (state != ATstates.SEARCH && data.get(DATA_ID) != 0) {
				Human hmn = (Human) model().getEntity(new EntityID(data.get(DATA_ID)));
				getAmbulanceTeamTarget().put(at, hmn);
			}
		}

	}

	public ATstates convertStateIndexToState(int stateIndex) {
		switch (stateIndex) {
		case 0:
			return ATstates.MOVE_TO_TARGET;
		case 1:
			return ATstates.RESCUE;
		case 2:
			return ATstates.MOVE_TO_REFUGE;
		case 3:
			return ATstates.SEARCH;
		default:
			return null;
		}
	}

	public ArrayList<Task<? extends StandardEntity>> getTasksOf(PoliceForceTask policeForceTask) {
		return agent_task.get(policeForceTask);
	}

	private Vector<Road> makeRoadsZone(HashSet<Building> myClusterBuildings) {
		Vector<Road> temp = new Vector<Road>();
		Road road = null;

		for (Building building : myClusterBuildings) {
			if (!buildzone.contains(building))
				buildzone.add(building);
			for (sos.base.entities.Area area : building.getRoadsInSight()) {
				if (area != null)
					if (area instanceof Road) {
						road = (Road) area;
						if (!temp.contains(road))
							temp.add(road);
					}
			}
		}
		for (PoliceForce policeForce : model().policeForces()) {
			if (isMyNearZone(newSearch.getSearchWorld().getMapClusterType().getCluster(policeForce).getBuildings())) {
				for (Building building : newSearch.getSearchWorld().getClusterData(policeForce).getBuildings()) {
					//					if (!buildzone.contains(building))
					//						buildzone.add(building);
					for (sos.base.entities.Area area : building.getRoadsInSight()) {
						if (area != null)
							if (area instanceof Road) {
								road = (Road) area;
								if (!temp.contains(road))
									temp.add(road);
							}
					}
				}
			}
		}

		return temp;
	}

	private boolean isMyNearZone(HashSet<Building> zonecluster) {
		for (Building b1 : zonecluster)
			for (Building b2 : b1.realNeighbors_Building())
				if (mycluster.contains(b2))
					return true;
		return false;
	}

	public FireZoneManager getFireZoneManager() {
		return fireZoneManager;
	}

	public void setFireZoneManager(FireZoneManager fireZoneManager) {
		this.fireZoneManager = fireZoneManager;
	}

	public Pair<Integer, Integer> getNumberOfCiviliansOpening() {
		return numberOfCiviliansOpening;
	}

	public void setNumberOfCiviliansOpening(Pair<Integer, Integer> numberOfCiviliansOpening) {
		this.numberOfCiviliansOpening = numberOfCiviliansOpening;
	}

	public int getLastShockTime() {
		return lastShockTime;
	}

	public void setLastShockTime(int lastShockTime) {
		this.lastShockTime = lastShockTime;
	}

}

class IdComparator implements Comparator<Human> {

	@Override
	public int compare(Human h1, Human h2) {
		if (h1.getID().getValue() < h2.getID().getValue())
			return 1;
		else if (h1.getID().getValue() > h2.getID().getValue())
			return -1;

		return 0;
	}
}
