package sos.police_v2.SSstate;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import rescuecore2.geometry.Line2D;
import rescuecore2.geometry.Point2D;
import rescuecore2.geometry.Vector2D;
import rescuecore2.misc.Pair;
import rescuecore2.worldmodel.EntityID;
import sos.base.entities.Area;
import sos.base.entities.Blockade;
import sos.base.entities.Civilian;
import sos.base.entities.Edge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.Path;
import sos.base.move.types.PoliceMove;
import sos.base.util.SOSActionException;
import sos.base.util.blockadeEstimator.AliGeometryTools;
import sos.police_v2.PoliceConstants;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.SSTask.ClearBlockade;
import sos.police_v2.SSTask.ClearPoint;
import sos.police_v2.SSTask.Move_path;
import sos.police_v2.clearableBlockadeToReachable.GeoPathReachablity;
import sos.police_v2.clearableBlockadeToReachable.ReachablingUsingReachablityTool;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.search_v2.tools.cluster.ClusterData;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class CivilianOpening extends SOSIState<PoliceInformationModel> {
	private Civilian targetcivilian;
	ArrayList<Pair<Civilian, Integer>> civiliansCosts;
	private SOSTask<?> task = null;
	ReachablingUsingReachablityTool reachablityTool;
	protected final GeoPathReachablity reachableWithEdge;

	public CivilianOpening(PoliceInformationModel infoModel) {
		
		super(infoModel);
		targetcivilian = null;
		reachablityTool = new ReachablingUsingReachablityTool(((PoliceForceAgent)infoModel.getAgent()));
		reachableWithEdge = new GeoPathReachablity(((PoliceForceAgent)infoModel.getAgent()));

	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {

		if (targetcivilian != null && civilianReachableDone(targetcivilian))
		{
			targetcivilian.setIsReallyReachable(true);
			targetcivilian = null;
		}
		if (!SSUtils.isValidCivilian(targetcivilian, ((PoliceForceAgent) infoModel.getAgent())))
			targetcivilian = chooseBestCivilian();
		if (targetcivilian != null)
		{
			Pair ep = getEntrancePoint(targetcivilian.getAreaPosition());
			if (ep == null)
			{
				targetcivilian = null;
			}
			moveToPoint(ep);
		}
		return task;
	}
	
	protected void moveToPoint(Pair<? extends Area, Point2D> destinations) throws SOSActionException {
		ArrayList<Pair<? extends Area, Point2D>> dests = new ArrayList<Pair<? extends Area, Point2D>>();
		dests.add(destinations);
		moveToPoint(dests);
	}
	
	protected void moveToPoint(Collection<Pair<? extends Area, Point2D>> destinations) throws SOSActionException {

		if (destinations.isEmpty()) {
			moveToARandomArea();
		}

		Path path = ((PoliceForceAgent)infoModel.getAgent()).move.getPathToPoints(destinations, PoliceMove.class);
		move(path);
	}
	
	public void moveToARandomArea() throws SOSActionException {

		Road randomDST = infoModel.getModel().roads().get((int) (Math.random() * (infoModel.getModel().roads().size() - 1)));
		ArrayList<Area> area = new ArrayList<Area>();
		area.add(randomDST);
		Path path = ((PoliceForceAgent)infoModel.getAgent()).move.getPathTo(area, PoliceMove.class);
		move(path);

	}
	
	public void move(Path path) throws SOSActionException {
		if (path == null) {
			moveToARandomArea();
		}
		justMakeReachable(path);
		task = new Move_path(path, null, infoModel.getTime());
	}
	
	private void justMakeReachable(Path path) throws SOSActionException {
		if (!PoliceConstants.IS_NEW_CLEAR) {

			ArrayList<Blockade> blocks = new ArrayList<Blockade>();

			ArrayList<Blockade> blockadeEdge = reachableWithEdge.getBlockingBlockadeOfPath(path);
			blocks.addAll(blockadeEdge);
			//
			if (blocks.isEmpty()) {
				blocks.addAll(reachablityTool.getBlockingBlockadeOfPath(path));
				//	ArrayList<Blockade> blockadeEdge = clearableBlockadeToFullReachable.getBlockingBlockadeOfPath(path);
			}

			Blockade clearBlock = chooseBestBlockade(blocks);

			if (clearBlock != null)
				task = new ClearBlockade(clearBlock, null, infoModel.getTime());
		} else {
			Point next = ((PoliceForceAgent)infoModel.getAgent()).degreeClearableToPoint.nextPointToClear(path, true, true);
			if (next != null) {
				task = new ClearPoint(next, null, infoModel.getTime());
			}
		}
	}
	
	protected Blockade chooseBestBlockade(List<Blockade> list) {
		if (list == null) {
			return null;
		}
		Blockade best = null;
		double bestValue = 0;
		for (Blockade blockade : list) {
			if (SSUtils.isValid(blockade)) {
				int value = (int) (1000000 / Math.max(infoModel.getPolice().getPositionPoint().distance(blockade.getCenteroid()), 1));
				if (value > bestValue) {
					bestValue = value;
					best = blockade;
				}
			}
		}
		return best;
	}

	private boolean civilianReachableDone(Civilian assignCivil)
	{
		if (((PoliceForceAgent) infoModel.getAgent()).location().equals(assignCivil.getPosition()))
			return true;
		if (assignCivil.isReallyReachable(true))
			return true;
		if (assignCivil.getPosition().isReallyReachable(true))
			return true;
		return ((PoliceForceAgent) infoModel.getAgent()).location().getNeighbours().contains(assignCivil.getPosition()) && ((PoliceForceAgent) infoModel.getAgent()).location().isBlockadesDefined() && ((PoliceForceAgent) infoModel.getAgent()).location().getBlockades().isEmpty();
	}

	private Civilian chooseBestCivilian()
	{

		ArrayList<Civilian> civis = new ArrayList<Civilian>();
		ClusterData myclusterData = ((PoliceForceAgent) infoModel.getAgent()).newSearch.getSearchWorld().getClusterData();
		double myclusterscore = ((PoliceForceAgent) infoModel.getAgent()).newSearch.getRemainingJobScorer().remainingJobScore(myclusterData);
		Vector clusterDatas = new Vector();
		clusterDatas.add(myclusterData);

		for (Iterator iterator = ((PoliceForceAgent) infoModel.getAgent()).newSearch.getSearchWorld().getAllClusters().iterator(); iterator.hasNext();)
		{
			ClusterData data = (ClusterData) iterator.next();
			if (myclusterscore < ((PoliceForceAgent) infoModel.getAgent()).newSearch.getRemainingJobScorer().remainingJobScore(data))
				clusterDatas.add(data);
			for (Iterator iterator1 = infoModel.getModel().civilians().iterator(); iterator1.hasNext();)
			{
				Civilian civ = (Civilian) iterator1.next();
				for (Iterator iterator2 = clusterDatas.iterator(); iterator2.hasNext();)
				{
					ClusterData clusterData = (ClusterData) iterator2.next();
					if (civ.isPositionDefined() && clusterData.getClusterArea().contains(civ.getX(),civ.getY()))
						civis.add(civ);
				}

			}

		}

		Civilian best = null;
		best = getBestUnreachableCivilians(civis);
		return best;
	}

	private Civilian getBestUnreachableCivilians(ArrayList<Civilian> cives)
	{
		Civilian best = null;
		civiliansCosts = new ArrayList<Pair<Civilian, Integer>>();
		int cost = 0;
		for (Civilian civ : cives) {

			if (!SSUtils.isValidCivilian(civ, ((PoliceForceAgent) infoModel.getAgent())))
				continue;
			int moveDistance = infoModel.getAgent().model().getDistance(infoModel.getATEntity().getAreaPosition(), civ.getAreaPosition());
			cost = (civ.getBuriedness() + civ.getHP()) * (moveDistance + 1) * 2;
			civiliansCosts.add(new Pair<Civilian, Integer>(civ, cost));

		}

		Collections.sort(civiliansCosts, new CostComparator());

		return civiliansCosts.get(0).first();
	}

	public Pair<Area, Point2D> getEntrancePoint(Area dest) {
		if (dest == null) {
			return null;
		}
		Path path = ((PoliceForceAgent) infoModel.getAgent()).move.getPathTo(Collections.singleton(dest), PoliceMove.class);
		ArrayList<EntityID> pathArray = path.getIds();
		if (pathArray.isEmpty()) {
			return null;
		}

		Edge inComingEdge;
		if (pathArray.size() == 1) {
			inComingEdge = dest.getEdgeTo(((PoliceForceAgent) infoModel.getAgent()).location());
		} else {
			inComingEdge = dest.getEdgeTo(pathArray.get(pathArray.size() - 2));
		}
		if (inComingEdge == null)
			return new Pair<Area, Point2D>(dest, dest.getPositionPoint());
		Line2D wallLine = inComingEdge.getLine();// new Line2D(edge.getStartX(), edge.getStartY(), edge.getEndX() - edge.getStartX(), edge.getEndY() - edge.getStartY());

		Vector2D offset;
		if (AliGeometryTools.havecorrectDirection(dest)) {
			offset = wallLine.getDirection().getNormal().normalised().scale(10);
		} else {
			offset = wallLine.getDirection().getNormal().normalised().scale(-10);
		}
		Point2D destXY = inComingEdge.getMidPoint().plus(offset);
		return new Pair<Area, Point2D>(dest, destXY);
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub

	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub

	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub

	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}

class CostComparator implements Comparator<Pair<Civilian, Integer>> {
	@Override
	public int compare(Pair<Civilian, Integer> h1, Pair<Civilian, Integer> h2) {// Target
		// cost
		int cost1 = h1.second();
		int cost2 = h2.second();

		if (cost1 < cost2 ||
				(cost1 == cost2 && (h1.first().getID().getValue() > h2.first().getID().getValue())))
			return -1;
		else if (h1.second() > h2.second())
			return 1;
		else
			return 1;

	}
}
