package sos.police_v2.SSstate;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.FireBrigade;
import sos.base.entities.Human;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.Path;
import sos.base.move.types.PoliceMove;
import sos.police_v2.SSTask.Move_path;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class CloseAgents extends SOSIState<PoliceInformationModel> {

	public CloseAgents(PoliceInformationModel infoModel) {
		super(infoModel);
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) {
		
		infoModel.getLog().info("//CLose Agents State\\");
		

		for (Human hum : infoModel.self().getVisibleEntities(Human.class)) {
			if (!(hum instanceof FireBrigade || hum instanceof AmbulanceTeam)){
				infoModel.getLog().info("target not instance of fire or human");
				continue;
			}
			
			if (Point.distance(infoModel.getPolice().getX(), infoModel.getPolice().getY(), hum.getX(), hum.getY()) > infoModel.self().clearDistance){
				infoModel.getLog().info("distance > myclear distance");
				continue;
			}
			
			if (!SSUtils.isReachableTo(hum, infoModel.self())) {
				List<? extends StandardEntity> targets = Arrays.asList(hum);
				ArrayList<Pair<? extends Area, Point2D>> reachableAreas = new ArrayList<Pair<? extends Area, Point2D>>();
				for (StandardEntity entity : targets) {
					if (entity == null) {
						infoModel.getLog().info("entity = null");
						continue;
					}
					
					if (entity.getPositionPair() == null) {
						infoModel.getLog().info("entity positionpair = null");
						continue;
					}
					
					if (!SSUtils.isReachableTo(entity, infoModel.self())) {
						infoModel.getLog().info("is no reachable to");
						reachableAreas.addAll(SSUtils.getReachableAreasPair(entity));
					}
				}
				
				
				if (!reachableAreas.isEmpty()) {
					infoModel.getLog().info("reachable areas is not empty");
					Road randomDST = infoModel.getModel().roads().get((int) (Math.random() * (infoModel.getModel().roads().size() - 1)));
					ArrayList<Area> area = new ArrayList<Area>();
					area.add(randomDST);
					Path path = infoModel.getAgent().move.getPathTo(area, PoliceMove.class);
					infoModel.getLog().info("new move path to" + path.getDestination());
					return new Move_path(path, null, infoModel.getTime());
					
				}

			}
			return null;

		}

		return null;
	}

	@Override
	public void giveFeedbacks(List feedbacks) {
	}

	@Override
	public void skipped() {
	}

	@Override
	public void overTaken() {
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
	}

	@Override
	public String getName() {
		return "close agents";
	}

	@Override
	public void taken() {
		super.taken();
	}
}
