package sos.police_v2.SSstate;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import sos.base.entities.Building;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.police_v2.SSTask.Search_task;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.search_v2.tools.SearchTask;
import sos.search_v2.tools.cluster.ClusterData;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.decisionMaker.implementations.targets.AreaTarget;
import sos.tools.decisionMaker.implementations.tasks.StandardMoveToAreaTask;

public class LowComOpenSearch extends SOSIState<PoliceInformationModel> {

	public LowComOpenSearch(PoliceInformationModel infoModel) {
		super(infoModel);
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) {
		
		infoModel.getLog().info("// LowComOpenSEarch \\");
		if (infoModel.getAgent().messageSystem.type != Type.LowComunication)
			return null;
		SearchTask task = infoModel.self().newSearch.fireSearchTask();
		if (task != null && !task.getArea().isEmpty()) {
				return new Search_task(task, null, infoModel.getTime());
		}
		Building best = SSUtils.openInClusterCivilian(infoModel.self().model().searchWorldModel.getClusterData(),infoModel.self());
		if (best != null)
			return new StandardMoveToAreaTask(new AreaTarget(best), infoModel.getTime());

		task = infoModel.self().newSearch.blockSearchTask();
		if (task != null && !task.getArea().isEmpty()) {
			if (task instanceof SearchTask)
				return new Search_task(task, null, infoModel.getTime());
		}
		task = infoModel.self().newSearch.civilianSearchTask();
		if (task != null && !task.getArea().isEmpty()) {
				return new Search_task(task, null, infoModel.getTime());
		}

		ArrayList<ClusterData> cds = new ArrayList<ClusterData>(infoModel.getModel().searchWorldModel.getAllClusters());
		Collections.sort(cds, new Comparator<ClusterData>() {

			@Override
			public int compare(ClusterData o1, ClusterData o2) {
				double o1s = infoModel.self().newSearch.getRemainingJobScorer().remainingJobScore(o1);
				double o2s = infoModel.self().newSearch.getRemainingJobScorer().remainingJobScore(o2);
				if (o2s > o1s)
					return 1;
				if (o2s < o1s)
					return -1;
				return 0;
			}
		});
		int i = 0;
		for (ClusterData cd : cds) {
			if (i++ == 5)
				break;
			 best=SSUtils.openInClusterCivilian(cd,infoModel.self());
			if (best != null)
				return new StandardMoveToAreaTask(new AreaTarget(best), infoModel.getTime());
		}

		task = infoModel.self().newSearch.combinedSearchTask();
		if (task != null)
				return new Search_task(task,null,infoModel.getTime());

		return null;
	}

	@Override
	public void giveFeedbacks(List feedbacks) {
	}

	@Override
	public void skipped() {
	}

	@Override
	public void overTaken() {
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
	}

	@Override
	public String getName() {
		return "ImHurtState";
	}

	@Override
	public void taken() {
		super.taken();
	}
}
