package sos.police_v2.SSstate;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.Path;
import sos.base.move.types.PoliceMove;
import sos.base.sosFireZone.SOSEstimatedFireZone;
import sos.base.sosFireZone.SOSRealFireZone;
import sos.base.util.SOSActionException;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.SSTask.Move_path;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.police_v2.preCompute.PoliceForceTask;
import sos.police_v2.preCompute.Task;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class OpeningAgent extends SOSIState<PoliceInformationModel>{

	Task<? extends StandardEntity> target = null;
	ArrayList<Task<? extends StandardEntity>> doneAgent;
	
	
	public OpeningAgent(PoliceInformationModel infoModel) {
		super(infoModel);
		doneAgent = new ArrayList<Task<? extends StandardEntity>>();
		// TODO Auto-generated constructor stub
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		for (PoliceForceTask police : ((PoliceForceAgent)infoModel.getAgent()).model().getPoliceForSpecialTask()) {
			if (police.getRealEntity().equals(infoModel.getPolice())) {
				return null;
			}
		}
		
		if (((PoliceForceAgent)infoModel.getAgent()).preIsDone)
			return null;
		
		infoModel.getLog().info("// Opening agent\\");
		
		if (target != null){
			infoModel.getLog().info("target was not null");
			return work();
		}
		setTarget();
		if (target != null) {
			return work();
		}

	return null;
	}
	
	public SOSTask<?>  work() throws SOSActionException{
		if (!isTargetStillValid()) {
			doneAgent.add(target);
			infoModel.getLog().info("done agent . add" + target.getFinalValue());
			target = null;
			return null;
		}
		
		if (target != null) {
			if (isDoneToClearAgent()) {
				doneAgent.add(target);
				infoModel.getLog().info("done agent sec . add" + target.getFinalValue());
				target = null;
				return null;
			}
			infoModel.getLog().info("make reachable to"  );
			return makeReachableTo(Arrays.asList(target.getRealEntity()));
		}
		return null;
	}
	
	public SOSTask<?> makeReachableTo(List<? extends StandardEntity> targets) throws SOSActionException {
		ArrayList<Pair<? extends Area, Point2D>> reachableAreas = new ArrayList<Pair<? extends Area, Point2D>>();
		for (StandardEntity entity : targets) {
			if (entity == null) {
				infoModel.getLog().info("entity = null");
				continue;
			}
			if (entity.getPositionPair() == null) {
				infoModel.getLog().info("entity positionpair = null");
				continue;
			}
			if (!SSUtils.isReachableTo(entity,infoModel)) {
				reachableAreas.addAll(SSUtils.getReachableAreasPair(entity));
			}
		}
		
		
		if (!reachableAreas.isEmpty()) {
			infoModel.getLog().info("reachable Areas is not empty");
			if (reachableAreas.isEmpty()) {
				Road randomDST = infoModel.getModel().roads().get((int) (Math.random() * (infoModel.getModel().roads().size() - 1)));
				ArrayList<Area> area = new ArrayList<Area>();
				area.add(randomDST);
				Path path = infoModel.getMove().getPathTo(area, PoliceMove.class);
				infoModel.getLog().info("new move path to" + path.getDestination());
				return new Move_path(path, null, infoModel.getTime());

			}

			Path path = infoModel.getMove().getPathToPoints(reachableAreas, PoliceMove.class);
			infoModel.getLog().info("new move path2 to" + path.getDestination());
			return new Move_path(path, null, infoModel.getTime());
			
		}
		return null;
	}
	
	private boolean isDoneToClearAgent() {
		infoModel.getLog().info("in isDoneToClean function");
		if (SSUtils.isReachableTo(target.getRealEntity(),infoModel)) {
			infoModel.getLog().info("is reahable to");
			return true;
		}
		if (infoModel.getPolice().getAreaPosition().equals(target.getRealEntity())) {
			infoModel.getLog().info("police area is equal to target");
			return true;
		}
		if (target.getRealEntity().isReallyReachable(true)) {
			infoModel.getLog().info("target is realy reachable");
			return true;
		}
		return false;
	}
	
	private boolean isTargetStillValid() {
		infoModel.getLog().info("is Target Still valid function");
		if (target.isDone())
		{
			infoModel.getLog().info("target is done");
			return false;
		}
		Human human = (Human) target.getRealEntity();

		if (human.isBuriednessDefined())
			if (human.getBuriedness() != 0) {
				infoModel.getLog().info("burridness is not 0");
				return false;
			}
		if (!human.isPositionDefined()) {
			infoModel.getLog().info("position not defined");
			return false;
		}
		if (human.getAreaPosition() instanceof Refuge){
			infoModel.getLog().info("in refuge");
			return false;
		}
		if (!(human.getAreaPosition() instanceof Building)){
			infoModel.getLog().info("not in building");
			return false;
		}
		if (doneAgent.contains(target))
		{
			infoModel.getLog().info("in doagent" );
			return false;
		}
		return true;
	}
	
	private void setTarget() {
		infoModel.getLog().info("in setTarget function");
		ArrayList<Task<? extends StandardEntity>> tasks =	((PoliceForceAgent)infoModel.getAgent()).getTasksOf(
				((PoliceForceAgent)infoModel.getAgent()).model().getPoliceTasks(infoModel.getPolice()));
		for (Task<? extends StandardEntity> task : tasks) {
			if (!(task.getRealEntity() instanceof Human))
			{
				infoModel.getLog().info("not instance of human");
				continue;
			}
			if (task.isDone()){
				infoModel.getLog().info("task is done");
				continue;
			}
			if (doneAgent.contains(task)){
				infoModel.getLog().info("do agent contains task");
				continue;
			}
			Human human = (Human) task.getRealEntity();

			if (human.isBuriednessDefined())
				if (human.getBuriedness() != 0) {
					continue;
				}
			if (!human.isPositionDefined()) {
				continue;
			}
			if (human.getAreaPosition() instanceof Refuge)
				continue;
			if (!(human.getAreaPosition() instanceof Building))
				continue;
			if (isInDangersPlace(human)) {
				target = task;
				break;
			}
		}
	}
	
	private boolean isInDangersPlace(Human human) {
		int distance = getDistanceToFire(human);
		if (distance < (infoModel.getModel().getDiagonalOfMap() / 10)){
			infoModel.getLog().info("human is in danger place");
			return true;
		}
		return false;
	}
	
	private int getDistanceToFire(Human human) {
		int distance = Integer.MAX_VALUE;
		for (Pair<ArrayList<SOSRealFireZone>, SOSEstimatedFireZone> pair : infoModel.getModel().getFireSites()) {
			if (pair.second().isDisable())
				continue;
			for (Building out : pair.second().getOuter()) {
				distance = (int) Math.min(distance, Point.distance(human.getX(), human.getY(), out.getX(), out.getY()));
			}
		}
		return distance;
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
	
	}
	

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}
