package sos.police_v2.SSstate;
 /// Saina: from heavystuckHandler
import java.awt.Point;
import java.awt.geom.Path2D;
import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;

import sos.base.entities.Area;
import sos.base.entities.Blockade;
import sos.base.entities.Edge;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.Path;
import sos.base.util.SOSActionException;
import sos.base.worldGraph.Node;
import sos.base.worldGraph.WorldGraphEdge;
import sos.police_v2.PoliceConstants;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.SSTask.ClearBlockade;
import sos.police_v2.SSTask.ClearPoint;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.tools.GraphEdge;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.geometry.SOSGeometryTools;

public class ReallyStuck extends SOSIState<PoliceInformationModel> {
	protected boolean isFinished = false;
	private SOSTask<?> task = null;
	public ReallyStuck(PoliceInformationModel infoModel) {
		super(infoModel);
		// TODO Auto-generated constructor stub
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		if (!PoliceConstants.IS_NEW_CLEAR) 
			clearNearestBlockade();
		else {
			isFinished = false;
			ArrayList<Blockade> stockdBlock = findStockBlockade2();
			Blockade best = chooseBestBlockade(stockdBlock);
			if (best != null)
				task = new ClearPoint(new Point(best.getX(), best.getY()), null, infoModel.getTime());
			else
				isFinished = true;
		}
		// TODO Auto-generated method stub
		return task;
	}
	

		
	public void clearNearestBlockade() throws SOSActionException {
		PriorityQueue<Blockade> blockadesInRange = ((PoliceForceAgent)infoModel.getAgent()).model().getBlockadesInRange(infoModel.getPolice().getX(), infoModel.getPolice().getY(),((PoliceForceAgent)infoModel.getAgent()).clearDistance);
		Blockade selectedBlock = null;
		if (!blockadesInRange.isEmpty())
			selectedBlock = blockadesInRange.remove();
		if (selectedBlock != null)
			task = new ClearBlockade(selectedBlock, null, infoModel.getTime());
	}
	
	public ArrayList<Blockade> findStockBlockade2() {

		ArrayList<Area> position = new ArrayList<Area>();
		Path2D geopath = findGeoPathInMove(position);
		ArrayList<Blockade> list = new ArrayList<Blockade>();
		for (Area ar : position) {
			if (ar.isBlockadesDefined())
				for (Blockade blockade : ar.getBlockades()) {
					if (SOSGeometryTools.haveIntersection(blockade.getShape(), geopath)) {
						if (SSUtils.isValid(blockade))
							list.add(blockade);
						}
					}
			}
		return list;
	}
		
	public Path2D findGeoPathInMove(ArrayList<Area> position) {

		Path2D geopath = new Path2D.Double();
		geopath.moveTo(infoModel.getPolice().getX(), infoModel.getPolice().getY());
		Path path = ((PoliceForceAgent)infoModel.getAgent()).informationStacker.getLastMovePath();

		GraphEdge[] allMoveEdges = path.getEdges();
		position.add(path.getSource().first());
		if (allMoveEdges == null) {
			geopath.lineTo(path.getDestination().second().getX(), path.getDestination().second().getY());
		} else {
			int index = 2;
			for (int i = 0; i < allMoveEdges.length; i++) {

				GraphEdge edge = allMoveEdges[i];
				short tail;
				short head;
				if (isInEdge(edge.getHeadIndex())) {
					head = edge.getHeadIndex();
					tail = edge.getTailIndex();
				} else {
					tail = edge.getHeadIndex();
					head = edge.getTailIndex();
				}
				Node start = infoModel.getModel().nodes().get(head);
				Node end = infoModel.getModel().nodes().get(tail);

				geopath.lineTo(start.getPosition().getX(), start.getPosition().getY());
				geopath.lineTo(end.getPosition().getX(), end.getPosition().getY());

				if (edge instanceof WorldGraphEdge) {
					Area area = infoModel.getModel().areas().get(((WorldGraphEdge) edge).getInsideAreaIndex());
					if (area.getLastSenseTime() == infoModel.getTime())
						position.add(area);
					index--;
					if (index == 0)
						break;
				}
			}
			if (position.size() <= 2) {
				if (path.getDestination().first().getLastSenseTime() == infoModel.getTime())
					position.add(path.getDestination().first());
				geopath.lineTo(path.getDestination().second().getX(), path.getDestination().second().getY());
			}
		}
		return geopath;
	}	
	
	private boolean isInEdge(short headIndex) {

		for (int i = 0; i < infoModel.getPolice().getAreaPosition().getPassableEdges().length; i++) {
			Edge ed = infoModel.getPolice().getAreaPosition().getPassableEdges()[i];
			if (headIndex == ed.getNodeIndex())
				return true;
		}
		return false;
	}
	
	protected Blockade chooseBestBlockade(List<Blockade> list) {
		if (list == null) {
			return null;
		}
		Blockade best = null;
		double bestValue = 0;
		for (Blockade blockade : list) {
			if (SSUtils.isValid(blockade)) {
				int value = (int) (1000000 / Math.max(infoModel.getPolice().getPositionPoint().distance(blockade.getCenteroid()), 1));

				if (value > bestValue) {
					bestValue = value;
					best = blockade;
				}
			}
		}
		return best;
	}
		
		

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}
