package sos.police_v2.SSstate;

import java.awt.Point;
import java.awt.geom.Path2D;
import java.util.ArrayList;
import java.util.List;

import sos.base.entities.Area;
import sos.base.entities.Blockade;
import sos.base.entities.Edge;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.Path;
import sos.base.util.SOSActionException;
import sos.base.worldGraph.Node;
import sos.base.worldGraph.WorldGraphEdge;
import sos.police_v2.PoliceConstants;
import sos.police_v2.SSTask.ClearPoint;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.tools.GraphEdge;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;
import sos.tools.geometry.SOSGeometryTools;

public class Stuck extends SOSIState<PoliceInformationModel>{

	short index;
	boolean isReachable;
	Node[] allMove;
	Point me,p1,p2,select = null;
	public Stuck(PoliceInformationModel infoModel) {
		super(infoModel);
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		infoModel.getLog().info("//StuckState\\");
		if(! SSUtils.isStocked(infoModel.self())){
			infoModel.getLog().info("police is not stucked");
			  return null;
		}
		
		if (!PoliceConstants.IS_NEW_CLEAR) {
			ArrayList<Area> position = new ArrayList<Area>();
			Path2D geopath = findGeoPathInMove(position);

			ArrayList<Blockade> list = new ArrayList<Blockade>();
			for (Area ar : position) {
				if (ar.isBlockadesDefined())
					for (Blockade blockade : ar.getBlockades()) {
						if (SOSGeometryTools.haveIntersection(blockade.getShape(), geopath)) {
							if (SSUtils.isValid(blockade))
								list.add(blockade);
						}
					}
			}
			Blockade best = chooseBestBlockade(list);

			if (best != null){
				infoModel.getLog().info("new clear point to" + new Point(best.getX(), best.getY()) );
				return new ClearPoint(new Point(best.getX(), best.getY()),null,infoModel.getTime());
			}
		} else {


			Path path = infoModel.self().informationStacker.getLastMovePath();
			index = 0;
			isReachable = false;
			allMove = path.getNodes();
			select = null;
			
			Point result = infoModel.self().degreeClearableToPoint.nextPointToClear(path, false, true);

			if (result != null){
				infoModel.getLog().info("clear poin task to " + result.getLocation());
				return new ClearPoint(result,null,infoModel.getTime());
			}
		}
	
		return null;
			
	}

	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void hear(String header, DataArrayList data,
			SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

	public Path2D findGeoPathInMove(ArrayList<Area> position) {
		infoModel.getLog().info(" In find geoParhinMove function");
		Path2D geopath = new Path2D.Double();
		geopath.moveTo(infoModel.getPolice().getX(), infoModel.getPolice().getY());
		Path path = infoModel.self().informationStacker.getLastMovePath();

		GraphEdge[] allMoveEdges = path.getEdges();
		position.add(path.getSource().first());
		if (allMoveEdges == null) {
			geopath.lineTo(path.getDestination().second().getX(), path.getDestination().second().getY());
		} else {
			int index = 2;
			for (int i = 0; i < allMoveEdges.length; i++) {

				GraphEdge edge = allMoveEdges[i];
				short tail;
				short head;
				boolean check=false;
				for (int k = 0; k < infoModel.getPolice().getAreaPosition().getPassableEdges().length; k++) {
					Edge ed = infoModel.getPolice().getAreaPosition().getPassableEdges()[k];
					if (edge.getHeadIndex() == ed.getNodeIndex()){
						check=true;
					break;}
				}
				check= false;
				if (check) {
					head = edge.getHeadIndex();
					tail = edge.getTailIndex();
				} else {
					tail = edge.getHeadIndex();
					head = edge.getTailIndex();
				}
				Node start = infoModel.getModel().nodes().get(head);
				Node end = infoModel.getModel().nodes().get(tail);

				geopath.lineTo(start.getPosition().getX(), start.getPosition().getY());
				geopath.lineTo(end.getPosition().getX(), end.getPosition().getY());

				if (edge instanceof WorldGraphEdge) {
					Area area = infoModel.getModel().areas().get(((WorldGraphEdge) edge).getInsideAreaIndex());
					if (area.getLastSenseTime() == infoModel.getTime())
						position.add(area);
					index--;
					if (index == 0)
						break;
					//					break;
				}
			}
			if (position.size() <= 2) {
				if (path.getDestination().first().getLastSenseTime() == infoModel.getTime())
					position.add(path.getDestination().first());
				geopath.lineTo(path.getDestination().second().getX(), path.getDestination().second().getY());
			}
		}
		return geopath;

	}


	protected Blockade chooseBestBlockade(List<Blockade> list) {
		infoModel.getLog().info("in chooseBestBlockade fnction");
		if (list == null) 
			return null;
		
		Blockade best = null;
		double bestValue = 0;
		for (Blockade blockade : list) {
			if (SSUtils.isValid(blockade)) {
				int value = (int) (1000000 / Math.max(infoModel.getPolice().getPositionPoint().distance(blockade.getCenteroid()), 1));
				if (value > bestValue) {
					bestValue = value;
					best = blockade;
				}
			}
		}
		return best;
	}
	
}
