package sos.police_v2.SSstate;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;

import rescuecore2.misc.Pair;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.PoliceForce;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.SOSBitArray;
import sos.base.message.structure.blocks.DataArrayList;
import sos.base.message.structure.channel.Channel;
import sos.base.move.types.PoliceMove;
import sos.base.sosFireZone.SOSAbstractFireZone;
import sos.base.sosFireZone.SOSEstimatedFireZone;
import sos.base.sosFireZone.SOSRealFireZone;
import sos.base.sosFireZone.util.ConvexHull_arr_New;
import sos.base.util.SOSActionException;
import sos.police_v2.PoliceConstants;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.decision.PoliceInformationModel;
import sos.police_v2.decision.SSUtils;
import sos.police_v2.preCompute.PoliceForceTask;
import sos.tools.decisionMaker.definitions.commands.SOSTask;
import sos.tools.decisionMaker.definitions.feedback.SOSFeedback;
import sos.tools.decisionMaker.implementations.stateBased.SOSEventPool;
import sos.tools.decisionMaker.implementations.stateBased.events.SOSEvent;
import sos.tools.decisionMaker.implementations.stateBased.states.SOSIState;

public class UpdateFire extends SOSIState<PoliceInformationModel>{
	private Road lasttarget;
	public ArrayList<Road> roadsToClearNearFireSide;
	public UpdateFire(PoliceInformationModel infoModel) {
		super(infoModel);

		// TODO Auto-generated constructor stub
	}

	@Override
	public SOSTask<?> decide(SOSEventPool eventPool) throws SOSActionException {
		if (!isValidForMe()) {
			lasttarget = null;
			return null;
		}
		ArrayList<SOSAbstractFireZone> firezones = new ArrayList<SOSAbstractFireZone>();

		for (Pair<ArrayList<SOSRealFireZone>, SOSEstimatedFireZone> pfirezone : infoModel.getAgent().fireSiteManager.getFireSites()) {
			SOSEstimatedFireZone firezone = pfirezone.second();
			if (firezone.getAllBuildings().size() < 100 && !firezone.isDisable()) {
				if (!getRoadsToClearNearFireSide(firezone).isEmpty()) {
					firezones.add(firezone);
				}

			}
		}
		
		if (firezones.isEmpty()) {
			lasttarget = null;
			return null;
		}

		SOSAbstractFireZone myFireZone = getBestFireZone(firezones);

		if (myFireZone == null) {
			lasttarget = null;
			return null;
		}

		updateFireZone(myFireZone);
		

		return null;
	}
	
	private void updateFireZone(SOSAbstractFireZone myFireZone) throws SOSActionException {
		if (myFireZone == null)
			return;

		roadsToClearNearFireSide = getRoadsToClearNearFireSide(myFireZone);
		priorityRoads(roadsToClearNearFireSide);
		if (lasttarget != null && roadsToClearNearFireSide.contains(lasttarget))
			
			SSUtils.makeReachableTo(Arrays.asList(lasttarget),infoModel);
		for (Road road : roadsToClearNearFireSide) {
			lasttarget = road;
			SSUtils.makeReachableTo(Arrays.asList(road),infoModel);
		}
		lasttarget = null;

	}
	
	private void priorityRoads(ArrayList<Road> roadsToClearNearFireSide) {
		Collections.sort(this.roadsToClearNearFireSide, new Comparator<Road>() {

			@Override
			public int compare(Road o1, Road o2) {
				if (((PoliceForceAgent)infoModel.getAgent()).model().getHighways().contains(o1) == ((PoliceForceAgent)infoModel.getAgent()).model().getHighways().contains(o2)) {
					float weighto1 = infoModel.getMove().getWeightTo(o1, o1.getX(), o1.getY(), PoliceMove.class) * getRoadScore(o1);
					float weighto2 = infoModel.getMove().getWeightTo(o2, o2.getX(), o2.getY(), PoliceMove.class) * getRoadScore(o2);

					if (weighto1 < weighto2)
						return -1;
					if (weighto1 > weighto2)
						return 1;
					return 0;
				}
				if (((PoliceForceAgent)infoModel.getAgent()).model().getHighways().contains(o1))
					return -1;
				if (((PoliceForceAgent)infoModel.getAgent()).model().getHighways().contains(o2))
					return 1;
				return 0;
			}


			private float getRoadScore(Road road) {
				return 1;
			}
		});
	}
	private SOSAbstractFireZone getBestFireZone(ArrayList<SOSAbstractFireZone> firezones) {
		for (SOSAbstractFireZone firezone : firezones) {
			if (firezone.getAllBuildings().size() < 100 && !firezone.isDisable()) {
				return firezone;
			}
		}
		return null;
	}
	
	public boolean isValidForMe() {
		for (PoliceForceTask police : ((PoliceForceAgent)infoModel.getAgent()).model().getPoliceForSpecialTask()) {
			if (police.getRealEntity().equals(infoModel.getPolice()))
				return true;
		}
		int updaterState = infoModel.getModel().policeForces().size() / 10;
		if (updaterState >= infoModel.getModel().policeForces().size())
			return false;
		ArrayList<PoliceForce> clonepolices = new ArrayList<PoliceForce>(infoModel.getModel().policeForces());
		Collections.sort(clonepolices, new Comparator<PoliceForce>() {

			@Override
			public int compare(PoliceForce o1, PoliceForce o2) {
				return ((PoliceForceAgent)infoModel.getAgent()).model().getPoliceTasks(o1).getJobDone() - ((PoliceForceAgent)infoModel.getAgent()).model().getPoliceTasks(o2).getJobDone();
			}
		});
		for (int i = 0; i < updaterState; i++) {
			if (clonepolices.get(i).equals(infoModel.getPolice()))
				return true;
		}
		return false;
	}
	
	private ArrayList<Road> getRoadsToClearNearFireSide(SOSAbstractFireZone myFireZone) {


		ConvexHull_arr_New fireZoneConvex = getConvexOfFireZone(myFireZone);

		ConvexHull_arr_New worldConvex = new ConvexHull_arr_New(infoModel.getModel().buildings()).getScaleConvex(.8f);

		HashSet<Road> inFireZone = new HashSet<Road>();
		Collection<Road> roads =  infoModel.getModel().getObjectsInRectangle(fireZoneConvex.getShape().getBounds(), Road.class);
		for (Road road : roads)
			if (fireZoneConvex.contains(road.getX(), road.getY()))
				inFireZone.add(road);

		ConvexHull_arr_New fireZoneConvex_OutSide = fireZoneConvex.getScaleConvex(1.5f);
		ArrayList<Road> selectedRoads = new ArrayList<Road>();
		roads = infoModel.getModel().getObjectsInRectangle(fireZoneConvex_OutSide.getShape().getBounds(), Road.class);

		HashSet<Road> mapsides = new HashSet<Road>();
		FOR: for (Road road : roads) {
			if (!fireZoneConvex_OutSide.contains(road.getX(), road.getY()))
				continue;

			if (inFireZone.contains(road))
				continue;

			for (Area ne : road.getNeighbours()) {
				if (ne instanceof Building)
					continue FOR;
				if (road.getSOSGroundArea() < PoliceConstants.VERY_SMALL_ROAD_GROUND_IN_MM)
					continue FOR;
				if (SSUtils.isReachableTo(road,((PoliceForceAgent)infoModel.getAgent())))
					continue FOR;
			}
			if (worldConvex.contains(road.getX(), road.getY())) {
				mapsides.add(road);
				continue;
			}
			selectedRoads.add(road);

		}
		if (selectedRoads.isEmpty()) {
			ConvexHull_arr_New fireZoneConvex_Inside = fireZoneConvex.getScaleConvex(.8f);
			for (Road road : inFireZone) {
				if (fireZoneConvex_Inside.contains(road.getX(), road.getY()))
					continue;
				if (worldConvex.contains(road.getX(), road.getY())) {
					mapsides.add(road);
					continue;
				}
				selectedRoads.add(road);

			}
		}

		if (selectedRoads.isEmpty()) {
			selectedRoads.addAll(mapsides);
		}

		return selectedRoads;
	}
	
	private ConvexHull_arr_New getConvexOfFireZone(SOSAbstractFireZone myFireZone) {
		ArrayList<Building> fiery = new ArrayList<Building>();
		for (Building b : myFireZone.getFires()) {
			if (b.virtualData[0].isBurning())
				fiery.add(b);
		}
		ConvexHull_arr_New convex = new ConvexHull_arr_New(fiery);
		return convex;
	}


	@Override
	public void giveFeedbacks(List<SOSFeedback> feedbacks) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipped() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void overTaken() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected void handleEvent(SOSEvent sosEvent) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void hear(String header, DataArrayList data, SOSBitArray dynamicBitArray, StandardEntity sender, Channel channel) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

}
