package sos.police_v2.decision;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import rescuecore2.geometry.GeometryTools2D;
import rescuecore2.geometry.Line2D;
import rescuecore2.geometry.Point2D;
import rescuecore2.geometry.Vector2D;
import rescuecore2.misc.Pair;
import rescuecore2.worldmodel.EntityID;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.SOSAgent;
import sos.base.SOSWorldModel;
import sos.base.entities.Area;
import sos.base.entities.Blockade;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Edge;
import sos.base.entities.Human;
import sos.base.entities.PoliceForce;
import sos.base.entities.Refuge;
import sos.base.entities.Road;
import sos.base.entities.StandardEntity;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.move.Path;
import sos.base.move.types.PoliceMove;
import sos.base.move.types.PoliceReachablityMove;
import sos.base.reachablity.PoliceReachablity;
import sos.base.reachablity.Reachablity;
import sos.base.reachablity.Reachablity.ReachablityState;
import sos.base.reachablity.tools.SOSArea;
import sos.base.reachablity.tools.Utility;
import sos.base.util.SOSActionException;
import sos.base.util.SOSGeometryTools;
import sos.base.util.blockadeEstimator.AliGeometryTools;
import sos.base.util.information_stacker.act.MoveAction;
import sos.base.util.information_stacker.act.StockMoveAction;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.police_v2.PoliceConstants;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.SSTask.Move_path;
import sos.police_v2.preCompute.Task;
import sos.search_v2.tools.cluster.ClusterData;
import sos.search_v2.worldModel.SearchBuilding;
import sos.tools.decisionMaker.definitions.commands.SOSTask;

public class SSUtils {

	public static boolean isValid(Blockade b) {
		if (b == null)
			return false;

		PoliceForceAgent policeForceAgent = (PoliceForceAgent) b.getAgent();
		double blockadeDistance = getBlockadeDistance(b);
		if (b.getLastSenseTime() < policeForceAgent.time() - 2)
			return false;
		if (blockadeDistance >= policeForceAgent.clearDistance)
			return false;
		return true;
	}
	
	public static  boolean isAgentvalid(PoliceForce agent){
	if(agent.getHP() == 0 )
		return false;
	
	return true;
	
}
	
	
	
	public static boolean isValidHuman(Human human) {
		if (human == null)
			return false;
		if (!human.isPositionDefined())
			return false;
		if (!human.isHPDefined())
			return false;
		if (human.getHP() < 100)
			return false;
		if (!human.isDamageDefined())
			return false;
		if (human.isBuriednessDefined() && human.getBuriedness() == 0)
			return false;
		if (human.getPositionArea() instanceof Refuge)
			return false;

		return true;
	}
	
	public static ArrayList<Blockade> getClearableBlockades(Road blockArea, Point2D point1, Point2D point2) {
	SOSLoggerSystem log = blockArea.getAgent().sosLogger.agent;
	log.debug("getClearableBlockades(Road " + blockArea + ", Point2D" + point1 + ", Point2D " + point2);
	ArrayList<Blockade> clearableBlockades = PoliceReachablity.clearableBlockades(blockArea, point1, point2);
	if (clearableBlockades.isEmpty() && Reachablity.isReachable(blockArea, point1, point2) == ReachablityState.Close) {
		blockArea.getAgent().sosLogger.info("there is an error in communication between police reachablity and reachablity!!! it should bo solve!!!now using police another reachablity util!!! ");
		Edge clearAreaEdge = new Edge(point1, point2);
		if (blockArea.isBlockadesDefined() && blockArea.getLastSenseTime() > blockArea.model().time() - 2) {
			for (Blockade blockade : blockArea.getBlockades()) {
				if (haveIntersect(blockade.getExpandedBlock(), clearAreaEdge)) {
					clearableBlockades.add(blockade);
				}
			}
		}
	}
	return clearableBlockades;
}
	
	private static boolean haveIntersect(SOSArea expandedBlock, Edge clearAreaEdge) {
		java.awt.geom.Area area = new java.awt.geom.Area(expandedBlock.getShape());
		area.intersect(new java.awt.geom.Area(clearAreaEdge.getShape()));
		boolean haveIntersectGeom = !area.isEmpty();
		boolean haveIntersectotherReachablity = false;
		for (Edge e : expandedBlock.getEdges()) {
			if (Utility.getIntersect(e, clearAreaEdge) != null) {
				haveIntersectotherReachablity = true;
				break;
			}
		}
		if (haveIntersectotherReachablity != haveIntersectGeom)
			System.err.println("IntersectotherReachablity:" + haveIntersectotherReachablity + " but IntersectGeom:" + haveIntersectGeom);
		return haveIntersectotherReachablity;
	}
	
	public static boolean isEntrance(Area area) {
	if (area instanceof Building)
		return false;
	for (Area nei : area.getNeighbours()) {
		if (nei instanceof Building)
			return true;
	}
	return false;
}
	
	public static ArrayList<ClusterData> getNeighberCluster(SOSWorldModel model, ClusterData select) {
	//		System.err.println(" getNeighberCluster TO POLICE UTILS CHECK Nashode hanoz ");
	ArrayList<ClusterData> result = new ArrayList<ClusterData>();
	for (ClusterData cluster : model.searchWorldModel.getAllClusters()) {
		if (cluster.equals(select))
			continue;
		if (cluster.isCoverer())
			continue;
		if (isNeighberCluster(select, cluster))
			result.add(cluster);
	}
	return result;
}
	
	private static boolean isNeighberCluster(ClusterData select, ClusterData cluster) {
		long distance = (long) Point.distance(select.getX(), select.getY(), cluster.getX(), cluster.getY());
		Rectangle rect1 = select.getBounds();
		long d1 = (long) Point.distance(rect1.getWidth(), rect1.getHeight(), 0, 0);
		d1 = d1 / 2;
		Rectangle rect2 = cluster.getBounds();
		long d2 = (long) Point.distance(rect2.getWidth(), rect2.getHeight(), 0, 0);
		d2 = d2 / 2;
		return distance <= (d1 + d2) * 0.85f;
	}
	
	public static boolean isValidHealtyCivilian(Civilian civ, PoliceForceAgent owner) {
		if (owner == null) {
			System.err.println("Owner is null!!!! why?????????");
		}
		SOSLoggerSystem log = owner.log;
		if (civ == null) {
			log.debug("NULL CIVILIAN CALLED TO CHECK ");
			return false;
		}
		if (civ.isUnkonwnCivilian()) {
			log.debug(civ + " is not valid beacuse it is UnknownCivilian");
			return false;
		}
		if (!civ.isPositionDefined()) {
			log.debug(civ + " is not valid beacuse it hasn't a defined position");
			return false;
		}
		if (civ.isReallyReachable(false) || (owner.isReachableTo(civ.getPositionPair()))) {
			log.debug(civ + " is not valid beacuse it is Reacable...");
			return false;
		}
		if (civ.getDamage() != 0) {
			log.debug(civ + " is not valid because got damage");
			return false;
		}

		if (!(civ.getAreaPosition() instanceof Building)) {
			log.debug(civ + "is not in building now");
			return false;
		}
		
		if (civ.getAreaPosition() instanceof Refuge)
			return false;

		log.debug(civ + " is valid Healty !");
		return true;
	}
	
	public static ArrayList<Pair<? extends Area, Point2D>> getReachableAreasPair(StandardEntity entity) {
	if (entity == null) {
		System.err.println("The pased entity is null in PoliceUtils.getReachableAreas????WHY??");
		return new ArrayList<Pair<? extends Area, Point2D>>();
	}
	ArrayList<Pair<? extends Area, Point2D>> allAreas = new ArrayList<Pair<? extends Area, Point2D>>();
	allAreas.add(entity.getPositionPair());
	if (entity instanceof Human) {
		for (Edge ed : ((Human) entity).getImReachableToEdges()) {
			Area target = entity.model().areas().get(ed.getMyAreaIndex());
			allAreas.add(new Pair<Area, Point2D>(target, ed.getMidPoint()));
		}
	}



	return allAreas;
}
	

	
	public static Blockade chooseBestBlockade(List<Blockade> list, PoliceForceAgent agent) {
		if (list == null) {
			return null;
		}
		Blockade best = null;
		double bestValue = 0;
		for (Blockade blockade : list) {
			if (isValid(blockade)) {
				int value = (int) (1000000 / Math.max(agent.me().getPositionPoint().distance(blockade.getCenteroid()), 1));
				if (value > bestValue) {
					bestValue = value;
					best = blockade;
				}
			}
		}
		return best;
	}
	
	public static boolean isReachableTo(StandardEntity se,PoliceInformationModel infoModel) {
		if (se instanceof Human) {
			Human hum = (Human) se;
			if (infoModel.getAgent().getVisibleEntities(Human.class).contains(hum) && hum.isPositionDefined() && hum.getAreaPosition().isBlockadesDefined()) {
				for (Blockade blockade : hum.getAreaPosition().getBlockades()) {
					if (blockade.getShape().contains(hum.getPositionPoint().toGeomPoint()))
						return false;
				}
			}
			if(!hum.getImReachableToEdges().isEmpty())
				return infoModel.getMove().getMoveType(PoliceReachablityMove.class).isReallyReachableTo(hum.getImReachableToEdges());

		}
		return !infoModel.getMove().isReallyUnreachableXYPolice(se.getPositionPair());
		
	}
	
	public static void makeReachableTo(List<? extends StandardEntity> targets,PoliceInformationModel infoModel) throws SOSActionException {
		ArrayList<Pair<? extends Area, Point2D>> reachableAreas = new ArrayList<Pair<? extends Area, Point2D>>();
		for (StandardEntity entity : targets) {
			if (entity == null) {
				continue;
			}
			if (entity.getPositionPair() == null) {
				continue;
			}
			if (!SSUtils.isReachableTo(entity,infoModel)) {
				reachableAreas.addAll(getReachableAreasPair(entity));
			}
		}
		if (!reachableAreas.isEmpty()) {
			moveToPoint(reachableAreas,infoModel);
		}
	}
	
	public static int getDistance(StandardEntity source, StandardEntity goal) {
		return getDistance(source.getLocation().first(), source.getLocation().second(), goal.getLocation().first(), goal.getLocation().second());

	}
	
	public static int getDistance(Point2D source, Point2D destination) {
	return getDistance((int) source.getX(), (int) source.getY(), (int) destination.getX(), (int) destination.getY());
}
	
	public static int getDistance(Task<?> source, Task<?> destination) {

		return getDistance(source.getX(), source.getY(), destination.getX(), destination.getY());
	}
	
	public static int getDistance(int x1, int y1, int x2, int y2) {

		//TODO Make it better
		int dx = x1 - x2;
		int dy = y1 - y2;
		return (int) (Math.sqrt(dx * dx + dy * dy) / PoliceConstants.DISTANCE_UNIT);
	}
	
	public static SOSTask<?> moveToPoint(Collection<Pair<? extends Area, Point2D>> dests,PoliceInformationModel infoModel) throws SOSActionException {
	if (dests.isEmpty()) {
		Road randomDST = infoModel.getModel().roads().get((int) (Math.random() * (infoModel.getModel().roads().size() - 1)));
		ArrayList<Area> area = new ArrayList<Area>();
		area.add(randomDST);
		Path path = infoModel.getMove().getPathTo(area, PoliceMove.class);
		return new Move_path(path, null, infoModel.getTime());
	}

	Path path = infoModel.getMove().getPathToPoints(dests, PoliceMove.class);
	return new Move_path(path, null, infoModel.getTime());
}

	/////////////////
	public static double getBlockadeDistance(Blockade b) {
		if (b == null)
			return Double.MAX_VALUE;

		PoliceForceAgent policeForceAgent = (PoliceForceAgent) b.getAgent();
		Point2D agentLocation = policeForceAgent.getPositionPoint();

		double bestDistance = Double.MAX_VALUE;
		List<Line2D> blockadelines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
		for (Line2D line : blockadelines) {
			Point2D closest = GeometryTools2D.getClosestPointOnSegment(line, agentLocation);
			double distance = GeometryTools2D.getDistance(agentLocation, closest);
			if (bestDistance > distance) {
				bestDistance = distance;
			}
		}
		return bestDistance;
	}

	////////////////////////////////
	public static boolean isStocked(PoliceForceAgent agent) {
		if (agent.time() < agent.FREEZE_TIME) {
			return false;
		}
		if (agent.informationStacker.getInformations(1).getAct() instanceof MoveAction && !(agent.informationStacker.getInformations(1).getAct() instanceof StockMoveAction)) {
			int lastLocationDistance = getDistance(agent.me().getPositionPair().second(), agent.informationStacker.getInformations(1).getPositionPair().second());
			if (lastLocationDistance < PoliceConstants.STOCK_DISTANCE) {
				return true;
			}
		}
		return false;
	}

	///////////////////////
	public static boolean isReachableTo(StandardEntity SEntity, PoliceForceAgent agent) {
		if (SEntity instanceof Human) {
			Human hum = (Human) SEntity;
			if (agent.getVisibleEntities(Human.class).contains(hum) && hum.isPositionDefined() && hum.getAreaPosition().isBlockadesDefined()) {
				for (Blockade blockade : hum.getAreaPosition().getBlockades()) {
					if (blockade.getShape().contains(hum.getPositionPoint().toGeomPoint()))
						return false;
				}
			}
			
			if (!hum.getImReachableToEdges().isEmpty())
				return agent.move.getMoveType(PoliceReachablityMove.class).isReallyReachableTo(hum.getImReachableToEdges());

		}
		return !agent.move.isReallyUnreachableXYPolice(SEntity.getPositionPair());

	}
	

	
	
	
	////////////check validity of civilians (in road)
	public static boolean isValidCivilianInRoad(Civilian civ, PoliceForceAgent agent) {

		if (civ == null) {
			return false;
		}
		if (civ.isUnkonwnCivilian()) {
			return false;
		}
		if (!civ.isPositionDefined()) {
			return false;
		}
		if (civ.isReallyReachable(false) || (agent.isReachableTo(civ.getPositionPair()))) {
			return false;
		}
		if (civ.getHP() != 0)
			return false;

		if (civ.getDamage() != 0)
			return false;
		

		return true;
	}
////////////
	//////////// check validity of civilians (without damage)
	public static boolean isValidCivilian(Civilian civ, PoliceForceAgent agent) {

		if (civ == null) {
			return false;
		}
		if (civ.isUnkonwnCivilian()) {
			return false;
		}
		if (!civ.isPositionDefined()) {
			return false;
		}
		if (civ.isReallyReachable(false) || (agent.isReachableTo(civ.getPositionPair()))) {
			return false;
		}
		if (!(civ.getAreaPosition() instanceof Building)) {
			return false;
		}
		
		if (!(civ.getAreaPosition() instanceof Refuge)) {
			return false;
		}
		return true;
	}
	////////////
	
	//////////// check validity of civilians (without damage)
	public static boolean isValidCivilianNotReady(Civilian civ, PoliceForceAgent agent) {

		if (civ == null) {
			return false;
		}
		if (civ.isUnkonwnCivilian()) {
			return false;
		}
		if (!civ.isPositionDefined()) {
			return false;
		}
		if (civ.isReallyReachable(false) || (agent.isReachableTo(civ.getPositionPair()))) {
			return false;
		}
		if (civ.getDamage() != 0) {
			return false;
		}
		if (civ.getBuriedness() != 0) {
			return false;
		}
		if (!(civ.getAreaPosition() instanceof Building)) {
			return false;
		}
		return true;
	}
	////////////
	
	public static boolean missionComplete(PoliceForceAgent agent) {
		if (agent.location().equals(agent.assignCivil.getPosition()))
			return true;
		if (agent.assignCivil.isReallyReachable(true))
			return true;
		if (agent.assignCivil.getPosition().isReallyReachable(true))
			return true;
		if (agent.location().getNeighbours().contains(agent.assignCivil.getPosition()) && agent.location().isBlockadesDefined() && agent.location().getBlockades().isEmpty())
			return true;
		return false;
	}
/////////////
	public static Pair<Area, Point2D> getEntrancePoint(Area dest,PoliceForceAgent agent) {
		if (dest == null) {
			return null;
		}
		Path path = agent.move.getPathTo(Collections.singleton(dest), PoliceMove.class);
		ArrayList<EntityID> pathArray = path.getIds();
		if (pathArray.isEmpty()) {
			return null;
		}

		Edge inComingEdge;
		if (pathArray.size() == 1) {
			inComingEdge = dest.getEdgeTo(agent.location());
		} else {
			inComingEdge = dest.getEdgeTo(pathArray.get(pathArray.size() - 2));
		}
		if (inComingEdge == null)
			return new Pair<Area, Point2D>(dest, dest.getPositionPoint());
		Line2D wallLine = inComingEdge.getLine();// new Line2D(edge.getStartX(), edge.getStartY(), edge.getEndX() - edge.getStartX(), edge.getEndY() - edge.getStartY());

		Vector2D offset;
		if (AliGeometryTools.havecorrectDirection(dest)) {
			offset = wallLine.getDirection().getNormal().normalised().scale(10);
		} else {
			offset = wallLine.getDirection().getNormal().normalised().scale(-10);
		}
		Point2D destXY = inComingEdge.getMidPoint().plus(offset);
		return new Pair<Area, Point2D>(dest, destXY);
	}
	
///////////
	public static boolean missionComplete(Civilian assignCivil,PoliceForceAgent agent) {
		if (agent.location().equals(assignCivil.getPosition()))
			return true;
		if (assignCivil.isReallyReachable(true))
			return true;
		if (assignCivil.getPosition().isReallyReachable(true))
			return true;
		if (agent.location().getNeighbours().contains(assignCivil.getPosition()) && agent.location().isBlockadesDefined() && agent.location().getBlockades().isEmpty())
			return true;
		return false;
	}
///////////////
	public static Civilian assignNewCivilian(PoliceForceAgent agent) {

		ArrayList<Civilian> validCivilians = new ArrayList<Civilian>();
		ArrayList<ClusterData> helpClusters = new ArrayList<ClusterData>();
		ArrayList<ClusterData> neighberCluster = getNeighberCluster(agent.model(), agent.model().searchWorldModel.getClusterData());
		HashMap<ClusterData, ArrayList<Civilian>> clustersCivilians = new HashMap<ClusterData, ArrayList<Civilian>>();

		validCivilians.clear();
		if (agent.messageSystem.type == Type.LowComunication || agent.messageSystem.type == Type.NoComunication) {
			final ClusterData myCluster = agent.model().searchWorldModel.getClusterData();
			ArrayList<ClusterData> openCivilianClusterList = new ArrayList<ClusterData>();
			openCivilianClusterList.add(myCluster);
			ArrayList<ClusterData> cloneClusters = new ArrayList<ClusterData>(agent.model().searchWorldModel.getAllClusters());
			Collections.sort(cloneClusters, new Comparator<ClusterData>() {

				@Override
				public int compare(ClusterData o1, ClusterData o2) {
					double o1s = SOSGeometryTools.distance(myCluster.getX(), myCluster.getY(), o1.getX(), o1.getY());
					double o2s = SOSGeometryTools.distance(myCluster.getX(), myCluster.getY(), o2.getX(), o2.getY());
					if (o1s > o2s)
						return 1;
					if (o1s < o2s)
						return -1;
					return 0;

				}
			});
			int i = 0;
			for (ClusterData cd : cloneClusters) {
				openCivilianClusterList.add(cd);
				i++;
				if (i == 2)
					break;
			}
			for (Civilian civ : agent.model().civilians()) {
				for (ClusterData clusterData : openCivilianClusterList) {
					if (civ.isPositionDefined() && clusterData.getClusterArea().contains(civ.getX(),civ.getY()))
						validCivilians.add(civ);
				}
			}
		} 
		else {

			helpClusters.clear();
			ClusterData myCluster = agent.model().searchWorldModel.getClusterData();
			double myCLusterScore = agent.newSearch.getRemainingJobScorer().remainingJobScore(myCluster);
			for (ClusterData cd : neighberCluster) {
				if (myCLusterScore < agent.newSearch.getRemainingJobScorer().remainingJobScore(cd))
					helpClusters.add(cd);
			}
			ArrayList<ClusterData> openCivilianClusterList = makeListOfValidClusters(agent,helpClusters);
			for (ClusterData cluster : openCivilianClusterList){
				if(cluster == null || clustersCivilians.get(cluster) == null)
					continue;
				validCivilians.addAll(clustersCivilians.get(cluster));
			}

		}
		Civilian best = getBestUnreachableCivilians(validCivilians,agent);

		return best;
}
	/////////////////
	public static  ArrayList<ClusterData> makeListOfValidClusters(PoliceForceAgent agent,ArrayList<ClusterData> helpClusters) {
		ArrayList<ClusterData> result = new ArrayList<ClusterData>();
		ClusterData mycluster = agent.model().searchWorldModel.getClusterData();
		result.add(mycluster);
		HashMap<ClusterData, ArrayList<Civilian>> clustersCivilians = new HashMap<ClusterData, ArrayList<Civilian>>();
		clustersCivilians.clear();
		clustersCivilians.put(mycluster, new ArrayList<Civilian>());
		for (ClusterData cluster : helpClusters)
			clustersCivilians.put(cluster, new ArrayList<Civilian>());
		for (Civilian civilian : agent.model().civilians()) {
			if (!isValidCivilian(civilian, agent, true))
				continue;
			if (mycluster.getBuildings().contains(civilian.getPositionArea())) {
				clustersCivilians.get(mycluster).add(civilian);
				continue;
			}
			for (ClusterData cluster : helpClusters) {
				if (cluster.getBuildings().contains(civilian.getPositionArea()))
					clustersCivilians.get(cluster).add(civilian);
			}
		}
		for (ClusterData cluster : helpClusters)
			if (clustersCivilians.get(cluster).size() > 5)
				result.add(cluster);
		return result;
	}
	
	public static boolean isValidCivilian(Civilian civ, SOSAgent<?> owner, boolean checkReachablity) {
	
	if (owner == null) {
		System.err.println("Owner is null!!!! why?????????");
	}
	SOSLoggerSystem log = owner.sosLogger.agent;
	if (civ == null) {
		log.debug(" civ is not valid beacuse it is null");
		return false;
	}
	if (civ.isUnkonwnCivilian()) {
		log.debug(civ + " is not valid beacuse it is UnknownCivilian");
		return false;
	}
	if (!civ.isPositionDefined()) {
		log.debug(civ + " is not valid beacuse it hasn't a defined position");
		return false;
	}
	if (!(civ.getAreaPosition() instanceof Building)) {
		log.debug(civ + " is not valid beacuse it isn't in building");
		return false;

	}

	if ((civ.getAreaPosition() instanceof Refuge)) {
		log.debug(civ + " is not valid beacuse it isn't in building");
		return false;
	}

	if (civ.isReallyReachable(false) || (checkReachablity && !owner.move.isReallyUnreachableXYPolice(civ.getPositionPair()))) {
		log.debug(civ + " is not valid beacuse it is Reacable...");
		return false;
	} else if (!checkReachablity)
		log.debug("Reachablity not checked!");

	if (civ.getHP() == 0) {
		log.debug(civ + " is not valid beacuse it is dead...");
		return false;
	}
	if (civ.getDamage() > 200) {
		log.debug(civ + " is not valid beacuse it's damage is bigger that 200...");
		return false;
	}
	
	int deathTime = SimpleDeathTime.getEasyLifeTime(civ.getHP(), civ.getDamage(), owner.time());

	if (deathTime - owner.model().time() < civ.getBuriedness() / 3 + 5) {
		int dieCycle = civ.getRescueInfo().getDeathTime();
		log.debug(civ + " is not valid beacuse it will die soon...dieCycle= " + dieCycle);
		return false;
	}

	log.debug(civ + " is valid!");
	return true;
}

	/////////////////////////
	private static Civilian getBestUnreachableCivilians(ArrayList<Civilian> cives,PoliceForceAgent agent) {
		if (cives.size() == 0)
			return null;
		Civilian best = null;
		ClusterData myCluster = agent.model().searchWorldModel.getClusterData();
		int diagonal = myCluster.getDiagonalOfCluster();
		ArrayList<Pair<Civilian, Integer>> scoreList = new ArrayList<Pair<Civilian, Integer>>();
		ArrayList<Pair<Civilian, Integer>> FindMax = new ArrayList<Pair<Civilian, Integer>>();
		int maxCost = 0;
		int cost;
		for (Civilian civ : cives) {
			if (!isValidCivilian(civ, agent, true))
				continue;
			if(civ.getDamage() == 0)
				cost=Integer.MAX_VALUE;
			else
			cost = (civ.getBuriedness() * civ.getHP()) / civ.getDamage();
			FindMax.add(new Pair<Civilian,Integer>(civ,cost));
			if(cost > maxCost)
				maxCost = cost;
		}
		
		for (Civilian civ : cives) {
			int indexCiv = findIndexCost(civ,FindMax);
			if (!isValidCivilian(civ, agent, true))
				continue;
			if (myCluster.getBuildings().contains(civ.getAreaPosition())) {
				int score = 1000;
				int dis = (int) Point.distance(civ.getX(), civ.getY(), agent.me().getX(), agent.me().getY());
				score += ((diagonal - dis) / diagonal) * 100;
				//score += (200 - (civ.getRescueInfo().getDeathTime() - model().time())) * 10;
				score += 200 + maxCost - FindMax.get(indexCiv).second();
				//TODO
//				if ((civ.getRescueInfo().getDeathTime() - agent.model().time()) < civ.getBuriedness() + 15)
//					score += 2000;
				scoreList.add(new Pair<Civilian, Integer>(civ, score));
			}
		}
		Collections.sort(scoreList, new PairComparator());

		Pair<Civilian, Integer> maxScoreCivilian = null;

		if (scoreList.size() > 0) {
			maxScoreCivilian = scoreList.get(0);
		}
		
		for (Civilian civ : cives) {
			int indexCiv = findIndexCost(civ,FindMax);
			if (!isValidCivilian(civ, agent, true))
				continue;
			if (!myCluster.getBuildings().contains(civ.getPositionArea())) {
				int score = 0;
				if (maxScoreCivilian != null) {
					int dis = (int) Point.distance(maxScoreCivilian.first().getX(), maxScoreCivilian.first().getY(), civ.getX(), civ.getY());
					score = ((diagonal - dis) / diagonal) * 100;
					//score += (100 - (civ.getRescueInfo().getDeathTime() - model().time())) * 10;
					score += 100 + maxCost - FindMax.get(indexCiv).second();
				} 
				
				else {
					int dis = (int) Point.distance(myCluster.getX(), myCluster.getY(), civ.getX(), civ.getY());
					score = ((diagonal - dis) / diagonal) * 100;
					int deathTime = SimpleDeathTime.getEasyLifeTime(civ.getHP(), civ.getDamage(), agent.time());

					score += (100 - (deathTime - agent.model().time())) * 10;
				}
				scoreList.add(new Pair<Civilian, Integer>(civ, score));

			}
		}
		Collections.sort(scoreList, new PairComparator());
		if (scoreList.size() > 0)
			best = scoreList.get(0).first();
		return best;
	}
	
	private static int findIndexCost(Civilian civ, ArrayList<Pair<Civilian, Integer>> findMax) {
		for (int i = 0 ; i < findMax.size() ; i++){
			if(findMax.get(i).first().getID().getValue() == civ.getID().getValue())
					return i;
		}
		return 0;
		
			
	}
	////////////
	public static class PairComparator implements Comparator<Pair<Civilian, Integer>> {

		@Override
		public int compare(Pair<Civilian, Integer> o1, Pair<Civilian, Integer> o2) {
			if (o1.second() > o2.second())
				return -1;
			else if (o2.second() > o1.second())
				return 1;
			return 0;
		}

	}
	
//	public static void setAgentTarget(PoliceForceAgent police) {
//		ArrayList<Task<? extends StandardEntity>> tasks = state.getTasksOf(model().getPoliceTasks(agent.me()));
//		for (Task<? extends StandardEntity> task : tasks) {
//			if (!(task.getRealEntity() instanceof Human))
//				continue;
//			if (task.isDone())
//				continue;
//			if (doneAgent.contains(task))
//				continue;
//			Human human = (Human) task.getRealEntity();
//
//			if (human.isBuriednessDefined())
//				if (human.getBuriedness() != 0) {
//					continue;
//				}
//			if (!human.isPositionDefined()) {
//				continue;
//			}
//			if (human.getAreaPosition() instanceof Refuge)
//				continue;
//			if (!(human.getAreaPosition() instanceof Building))
//				continue;
//			if (isInDangersPlace(human)) {
//				police.sett= task;
//				break;
//			}
//		}
//	}
	////////////////
	public static Building openInClusterCivilian(ClusterData cd,PoliceForceAgent agent){
		Building best=null;
		long bestmove=Long.MAX_VALUE;
		for (Building b : cd.getBuildings()) {
			SearchBuilding s = agent.model().searchWorldModel.getSearchBuilding(b);
			if (s.getValidCivilianCountInLowCom() == 0)
				continue;
			if (!s.isReallyUnReachableInLowCom(true))
				continue;
			
			long tmpMoveScore = agent.move.getWeightTo(s.getRealBuilding(), s.getRealBuilding().getX(), s.getRealBuilding().getY(), PoliceMove.class);
			
			if (best == null || bestmove > tmpMoveScore) {
				best = s.getRealBuilding();
				bestmove = tmpMoveScore;
			}
			
		}	
		return best;
	}
	
	public static java.awt.geom.Area getClearArea(Human agent, int targetX, int targetY,
			int clearLength, int clearRad) {
		clearLength = clearLength - 200;
		clearRad = clearRad - 200;
		Vector2D agentToTarget = new Vector2D(targetX - agent.getX(), targetY
				- agent.getY());

		if (agentToTarget.getLength() > clearLength)
			agentToTarget = agentToTarget.normalised().scale(clearLength);

		Vector2D backAgent = (new Vector2D(agent.getX(), agent.getY()))
				.add(agentToTarget.normalised().scale(-450));
		Line2D line = new Line2D(backAgent.getX(), backAgent.getY(),
				agentToTarget.getX(), agentToTarget.getY());

		Vector2D dir = agentToTarget.normalised().scale(clearRad);
		Vector2D perpend1 = new Vector2D(-dir.getY(), dir.getX());
		Vector2D perpend2 = new Vector2D(dir.getY(), -dir.getX());

		Point2D points[] = new Point2D[] {
				line.getOrigin().plus(perpend1),
				line.getEndPoint().plus(perpend1),
				line.getEndPoint().plus(perpend2),
				line.getOrigin().plus(perpend2) };
		int[] xPoints = new int[points.length];
		int[] yPoints = new int[points.length];
		for (int i = 0; i < points.length; i++) {
			xPoints[i] = (int) points[i].getX();
			yPoints[i] = (int) points[i].getY();
		}
		return new java.awt.geom.Area(new Polygon(xPoints, yPoints, points.length));
	}

	
	
}