package sos.police_v2.states;

import java.awt.geom.Point2D;
import java.util.ArrayList;

import rescuecore2.misc.Pair;
import sos.base.entities.AmbulanceTeam;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Human;
import sos.base.util.SOSActionException;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.decision.SSUtils;

public class AfterShockAmbulanceState extends PoliceAbstractState {

	ArrayList<Pair<AmbulanceTeam, Integer>> ambulanceScores;
	public AfterShockAmbulanceState(PoliceForceAgent policeForceAgent) {
		super(policeForceAgent);

	}

	@Override
	public void act() throws SOSActionException {
		if(model().getLastAfterShockTime() > model().time() - 5)
			agent.setLastShockTime(model().getLastAfterShockTime());
		if(agent.getLastShockTime() != 0)
			chooseAmbulance();
	}


	private void chooseAmbulance() throws SOSActionException {
		Human target;
		ambulanceScores = new ArrayList<Pair<AmbulanceTeam,Integer>>();
		AmbulanceTeam choosedAmb;
		
		for (AmbulanceTeam amb : model().ambulanceTeams()) {
			int score = 0;
			int timeToBeRescued = 0;
			if (!model().searchWorldModel.getClusterData().getConvexShape().contains(new Point2D.Double(amb.getX(), amb.getY())))
				continue;
			if (!amb.isBuriednessDefined() && (amb.getBuriedness() != 0))
				continue;
			if( amb.getAreaPosition() instanceof Building)
				score = numberOfAtInBuilding(amb.getAreaPosition()) * 100;
			target = agent.getAmbulanceTeamTarget().get(amb);
			if(SSUtils.isValidHuman(target))
				timeToBeRescued = target.getRescueInfo().getDeathTime() - model().time();
			if(timeToBeRescued < 8)
				continue;
			score += 500 - timeToBeRescued;
			if(target.getAreaPosition().equals(amb))
				score += 100;
			
			ambulanceScores.add(new Pair<AmbulanceTeam, Integer>(amb, score));
			if(chooseBestTarget().size() > 0)
				choosedAmb = chooseBestTarget().get(0).first();
			else
				choosedAmb = null;
			if(!choosedAmb.isReallyReachable(true) && !(choosedAmb.getAreaPosition().equals(agent.getPositionPair())))
				makeReachableTo(choosedAmb);			
			
		}
		// TODO Auto-generated method stub
		
	}
	
	private ArrayList<Pair<AmbulanceTeam, Integer>> chooseBestTarget() {
		int max = 0;
		ArrayList<Pair<AmbulanceTeam, Integer>> bestAmb = new ArrayList<Pair<AmbulanceTeam, Integer>>();
		
		for (Pair<AmbulanceTeam, Integer> amb : ambulanceScores) {
			if (amb.second() > max) {
				max = amb.second();
				if (bestAmb.size() != 0)
					bestAmb.clear();
				bestAmb.add(new Pair<AmbulanceTeam, Integer>(amb.first(), max));
			}
		}
		// TODO Auto-generated method stub
		return bestAmb;
		
	}

	private int numberOfAtInBuilding(Area areaPosition) {
		int index = 0;
		for (AmbulanceTeam at : model().ambulanceTeams()) {
			if (at.isPositionDefined() && at.getPositionArea().equals(areaPosition))
				index++;
		}
		return index;
	}

	@Override
	public void precompute() {
		// TODO Auto-generated method stub
		
	}
}
