package sos.police_v2.states;

import java.awt.Shape;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Vector;

import rescuecore2.geometry.Point2D;
import rescuecore2.misc.Pair;
import sos.ambulance_v2.decision.AmbUtile;
import sos.ambulance_v2.tools.GraphUsage;
import sos.ambulance_v2.tools.SimpleDeathTime;
import sos.base.entities.Area;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Human;
import sos.base.entities.Refuge;
import sos.base.message.structure.MessageConstants.Type;
import sos.base.util.SOSActionException;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.decision.SSUtils;
import sos.search_v2.tools.cluster.ClusterData;

public class CiviReachableState extends PoliceAbstractState {
	private GraphUsage gu;
	private Civilian targetcivilian = null;
	Pair<Civilian, Integer> bestCiv = null;
	Shape dangerousArea;


	public CiviReachableState(PoliceForceAgent policeForceAgent) {
		super(policeForceAgent);

	}

	@Override
	public void act() throws SOSActionException {
		dangerousArea = agent.getFireZoneManager().activeFirezoneShapeScale(1.5f);		
		if (targetcivilian != null && civilianReachableDone(targetcivilian)) {
			targetcivilian.setIsReallyReachable(true);
			targetcivilian = null;
		}

		if (!isValidCivilian(targetcivilian)) {
			targetcivilian = chooseBestCivilian();
		}
		
		if (targetcivilian != null) {
			Pair<Area, Point2D> ep = getEntrancePoint(targetcivilian.getAreaPosition());
			if (ep == null) {
				targetcivilian = null;
				return;
			}
			moveToPoint(ep);
		}
	}

	private Civilian chooseBestCivilian() {
		Vector<Civilian> civis = new Vector<Civilian>();
		ClusterData myclusterData = agent.newSearch.getSearchWorld().getClusterData();
		double myclusterscore = agent.newSearch.getRemainingJobScorer().remainingJobScore(myclusterData);
		Vector<ClusterData> clusterDatas = new Vector<ClusterData>();
		clusterDatas.add(myclusterData);

		for (ClusterData data : agent.newSearch.getSearchWorld().getAllClusters()) {
			if (myclusterscore < agent.newSearch.getRemainingJobScorer().remainingJobScore(data)) {
				clusterDatas.add(data);
			}
			
		}
		for (Civilian civ : model().civilians()) {
			if (civ.isHPDefined() && civ.getHP() == 0)
				continue;
			if (civ.isBuriednessDefined() && civ.getBuriedness() == 0)
				continue;

			for (ClusterData clusterData : clusterDatas) {

				if (civ.isPositionDefined() && clusterData.getClusterArea().contains(civ.getX(),civ.getY()))
					civis.add(civ);
			}
		}
		return getBestUnreachableCivilians(civis);

	}

	private boolean isValidCivilian(Civilian target) {
		if (target == null || target.isUnkonwnCivilian())
			return false;
		
		if (target.getAreaPosition() instanceof Refuge)
			return false;
		if (target.isReallyReachable(true) || (agent.isReachableTo(target.getPositionPair())))
		{
			agent.log.info("in is valid civilian and the civ is not valid");
			return false;

		}
		else
		{
			agent.log.info(" Civ is not reachable");
		}

		if (!target.isPositionDefined())
			return false;
		//TODO
		if (target.getHP() == 0 || target.getDamage() > 200)
			return false;
		if (dangerousArea.contains(target.getX(),target.getY()))
			return false;
		if ((target.getAreaPosition() instanceof Building))
			if(((Building) target.getAreaPosition()).getFieryness() == 0)
				return true;
		if (agent.model().time() > 120)
			return true;
		if (SimpleDeathTime.getEasyLifeTime(target.getHP(), target.getDamage(), target.updatedtime()) - agent.model().time() < target.getBuriedness() / 3 + 5) {
			return false;
		}
		return true;
	}

	private boolean civilianReachableDone(Civilian assignCivil) {
		if (agent.location().equals(assignCivil.getPosition()))
			return true;
		if (assignCivil.isReallyReachable(true))
			return true;
		if (assignCivil.getPosition().isReallyReachable(true))
			return true;
		if (agent.location().getNeighbours().contains(assignCivil.getPosition()) && agent.location().isBlockadesDefined() && agent.location().getBlockades().isEmpty())
			return true;
		return false;

	}

	private Civilian getBestUnreachableCivilians(Collection<Civilian> cives) {
		log.info("getBestUnreachableCivilians");
		ArrayList<Pair<Civilian, Integer>> costs = new ArrayList<Pair<Civilian, Integer>>();
		int cost = 0;

		int numberOfAmb = 0;
		int moveDistance;
		int timeToRefuge = 0;
		for (Civilian civilian : cives) {
			if (civilian.isReallyReachable(true))
				continue;
			
			
			if (SSUtils.isValidCivilian(civilian, agent, true)) {
				if (agent.model().refuges().size() == 0) {
					timeToRefuge = 0;
				}
				else
					timeToRefuge = civilian.getRescueInfo().getTimeToRefuge();

				
				moveDistance = agent.model().getDistance(agent.me().getPositionArea(), civilian.getAreaPosition());
				numberOfAmb = AmbUtile.numberOfAmbNeeded(timeToRefuge, civilian, civilian.getRescueInfo().getDeathTime());
				if (agent.messageSystem.type == Type.LowComunication && agent.messageSystem.type == Type.NoComunication) {
					int moveTime = gu.getFoolMoveTimeFromTo(agent.me().getAreaPosition(), civilian.getAreaPosition());
					cost = ((20 + (moveTime)) * civilian.getRescueInfo().getDeathTime()) + 1 / ((civilian.getBuriedness() * (civilian.getHP() * 15)) + 1);

				}
				else
					cost = civilian.getBuriedness() + (civilian.getHP() * civilian.getRescueInfo().getDeathTime()) + moveDistance;

//				cost = civilian.getBuriedness() + (civilian.getHP() * civilian.getRescueInfo().getDeathTime()) + moveDistance;
//				cost = numberOfAmb / (AmbUtile.numberOfFreeAmbs(model().ambulanceTeams().get(0)) + 1) + civilian.getBuriedness() + (civilian.getHP() * civilian.getRescueInfo().getDeathTime()) + moveDistance;
//				cost = (civilian.getBuriedness() + civilian.getHP()) * (moveDistance + 1);
			
			}

			costs.add(new Pair<Civilian, Integer>(civilian, cost));
		}

		return getBestCiv(costs);

	}

	private Civilian getBestCiv(ArrayList<Pair<Civilian, Integer>> costs) {
		int min = 1999999999;
		bestCiv = null;

		for (Pair<Civilian, Integer> civ : costs) {
			if (civ.second() < min) {
				min = civ.second();
				bestCiv = new Pair<Civilian, Integer>(civ.first(), min);
			}
		}
		// TODO Auto-generated method stub
		if (bestCiv == null)
			return null;
		else
			return bestCiv.first();
	}

	private class DistanceComparator implements java.util.Comparator<Civilian> {
		HashMap<Civilian, Integer> civ_cost = new HashMap<Civilian, Integer>();

		@Override
		public int compare(Civilian c1, Civilian c2) {
			Integer d1 = civ_cost.get(c1);
			if (d1 == null) {
				d1 = (int) (SSUtils.getDistance(agent.me(), c2) * (Math.random() * 25 + 75) / 100);
				civ_cost.put(c1, d1);
			}
			Integer d2 = civ_cost.get(c2);
			if (d2 == null) {
				d2 = (int) (SSUtils.getDistance(agent.me(), c2) * (Math.random() * 25 + 75) / 100);
				civ_cost.put(c2, d2);
			}

			double score1 = 1d;
			double score2 = 1d;
			if (c1.getPositionArea() instanceof Building)
				score1 = agent.newSearch.getRemainingJobScorer().getClusterRJS((Building) c1.getAreaPosition());
			if (c2.getPositionArea() instanceof Building)
				score2 = agent.newSearch.getRemainingJobScorer().getClusterRJS((Building) c2.getAreaPosition());

			double finalScore1 = score1 / (d1);
			double finalScore2 = score2 / (d2);
			if (finalScore2 < finalScore1)
				return -1;
			if (finalScore2 > finalScore1)
				return 1;

			return 0;
		}
	}

	@Override
	public void precompute() {

	}

}

class CostComparator implements Comparator<Pair<Human, Integer>> {
	@Override
	public int compare(Pair<Human, Integer> h1, Pair<Human, Integer> h2) {// Target
																		  // cost
		int cost1 = h1.second();
		int cost2 = h2.second();

		if (cost1 < cost2 ||
				(cost1 == cost2 && (h1.first().getID().getValue() > h2.first().getID().getValue())))
			return -1;
		else if (h1.second() > h2.second())
			return 1;
		else
			return 1;

	}
}
