package sos.police_v2.tool;

import java.awt.Point;
import java.util.Collection;
import java.util.Vector;

import sos.base.entities.Building;
import sos.base.entities.Road;
import sos.police_v2.base.worldModel.PoliceWorldModel;

public class FireZone {
	private Point centerofzone;
	private Vector<Building> buildings;
	private PoliceWorldModel model;
	public int pfWorkingNum=0;
	public int firstTaskSize=0;
	private int bordersize = 0;
	
	public FireZone(PoliceWorldModel worldModel) {
		buildings = new Vector<Building>();
		centerofzone = new Point(0, 0);
		model = worldModel;
	}

	public Boolean isContain(Building target) {
		return buildings.contains(target);
	}

	public void addBuild(Building target) {
		if (!buildings.contains(target)) {
			centerofzone.setLocation((centerofzone.getX() * buildings.size() + target.getX()) / (buildings.size() + 1), (centerofzone.getY() * buildings.size() + target.getY()) / (buildings.size() + 1));
			buildings.add(target);
		}
	}

	public Vector<Building> getBorder() {
		Vector<Building> border = new Vector<Building>();
		for (Building building : buildings) {
			if (isBorder(building))
				border.add(building);
		}
		bordersize = border.size();
		return border;

	}
	public boolean isActive(){
		for(Building building:getBorder()){
			if(building.getFieryness()==1||building.getFieryness()==2){
				if((model.time()-building.getLastCycleUpdated())<7)
					return true;
			}
		}
		return false;
	}

	private boolean isBorder(Building building) {
		if (building.getFieryness() == 2 || building.getFieryness() == 1)
			for (Building select : building.realNeighbors_Building()) {
				if (select.getFieryness() == 0) {

					return true;
				}
			}
		return false;
	}

	public Point getCenterOfZone() {
		return centerofzone;
	}

	public Vector<Road> getBestToReachable(Vector<Road> policezone) {
		Vector<Road> best = new Vector<Road>();
		Road temp;
		for (Building building : getBorder()) {
			temp = chooseisrange(building);
			if (temp != null) {
				if (policezone.contains(temp))
					best.add(temp);
			}
		}
		return best;

	}

	private Road chooseisrange(Building target) {
		Collection<Road> roads = model.getObjectsInRange(target, 25000, Road.class);
		Road bestroad = null;
		for (Road road : roads) {
			if (bestroad != null) {
				if (sumOfDistance(road, target) > sumOfDistance(bestroad, target)) {
					bestroad = road;
				}
			} else
				bestroad = road;
		}
		return bestroad;

	}

	private double sumOfDistance(Road road, Building target) {
		return (Point.distance(centerofzone.x, centerofzone.y, road.getX(), road.getY()) + Point.distance(target.getX(), target.getY(), road.getX(), road.getY()));
	}

	public int getBorderSize() {
		return bordersize;
	}

	public int getSize() {
		return buildings.size();
	}

	public Vector<Building> getBuilding() {
		return buildings;
	}

}
