package sos.police_v2.tool;

import java.awt.Point;
import java.awt.Shape;
import java.awt.geom.Area;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import rescuecore2.misc.Pair;
import sos.base.SOSWorldModel;
import sos.base.entities.Building;
import sos.base.entities.Civilian;
import sos.base.entities.Road;
import sos.base.sosFireZone.util.ConvexHull_arr_New;
import sos.base.util.SOSGeometryTools;
import sos.base.util.sosLogger.SOSLoggerSystem;
import sos.police_v2.PoliceForceAgent;
import sos.police_v2.base.worldModel.PoliceWorldModel;
import sos.search_v2.tools.cluster.ClusterData;

public class FireZoneManager {
	private Vector<FireZone> fireZones;
	private PoliceWorldModel policeWorldModel;

	private Vector<Road> Policezone;
	private double range;
	private SOSWorldModel fireBrigadeWorldModel;
	public ArrayList<ClusterData> notInFireClusters;
	ArrayList<Pair<FireZone, Integer>> sortSize = new ArrayList<Pair<FireZone, Integer>>();
	int closedRefuges;
	ArrayList<Civilian> clusterCivilians = new ArrayList<Civilian>();

	/************************************/
	public int need0 = 0, need1 = 0, need2 = 0;
	private PoliceForceAgent pf;

	/***********************************/

	public FireZoneManager(PoliceWorldModel model, Vector<Road> zone, PoliceForceAgent pf) {
		this.policeWorldModel = model;
		this.pf = pf;
		Policezone = zone;
		fireZones = new Vector<FireZone>();
		notInFireClusters = new ArrayList<ClusterData>();
		notInFireClusters.addAll(policeWorldModel.searchWorldModel
				.getAllClusters());
		updateFireZone();
		range = Point.distance(0, 0, policeWorldModel.getBounds().getWidth(),
				policeWorldModel.getBounds().getHeight()) / 2;
	}

	public void SetSize() {
		sortSize.clear();
		int z = 0, o = 0, t = 0;
		for (FireZone f : fireZones) {
			int tasksize = makePoliceTaskZone(f).size();
			f.firstTaskSize = tasksize;
			if (tasksize < 5) {
				sortSize.add(new Pair<FireZone, Integer>(f, 0));
				z++;
				continue;
			} else if (tasksize < 10) {
				sortSize.add(new Pair<FireZone, Integer>(f, 1));
				o++;
				continue;
			} else {
				sortSize.add(new Pair<FireZone, Integer>(f, 2));
				t++;
			}
		}
		pf.log.info("sorts = "+ sortSize.size());
		updateFators(z, o, t);
		Collections.sort(sortSize, new Comparator<Pair<FireZone, Integer>>() {

			@Override
			public int compare(Pair<FireZone, Integer> arg0,
					Pair<FireZone, Integer> arg1) {
				// TODO Auto-generated method stub
				if (arg0.second() < arg1.second())
					return -1;
				else
					return 1;
			}
		});

	}

	public void fireZoneLog(SOSLoggerSystem log) {
		for (FireZone f : fireZones) {
			log.info(" firezone = " + f.getCenterOfZone() + " border = "
					+ f.getBorderSize());
		}

	}

	ArrayList<ClusterData> all = new ArrayList<ClusterData>();

	public FireZone getBestForMe(PoliceForceAgent pf) {
		all.clear();
		all.addAll(pf.model().searchWorldModel.getAllClusters());
		SetSize();

		for (Pair<FireZone, Integer> p : sortSize) {

			pf.log.info("------------------new -------------------");
			for (Building b : p.first().getBuilding()) {
				pf.log.info(b.getID());
			}
			pf.log.info("------------------end -------------------");

			int index = getMyIndex(p, pf, all);
			switch (p.second()) {
			case 0:
				p.first().pfWorkingNum = need0;
				if (index <= need0)
					return p.first();
				break;
			case 1:
				p.first().pfWorkingNum = need1;
				if (index <= need1)
					return p.first();
				break;

			case 2:
				p.first().pfWorkingNum = need2;
				if (index <= need2)
					return p.first();
				break;

			default:
				break;
			}

		}
		return null;
	}

	private int getMyIndex(final Pair<FireZone, Integer> p,
			final PoliceForceAgent pf, ArrayList<ClusterData> ac) {

		ArrayList<ClusterData> allclusters = new ArrayList<ClusterData>();
		allclusters.addAll(ac);

		Collections.sort(allclusters, new Comparator<ClusterData>() {

			@Override
			public int compare(ClusterData arg1, ClusterData arg2) {
				float dis1 = (float) SOSGeometryTools.distance(p.first()
						.getCenterOfZone().getX(), p.first().getCenterOfZone()
						.getY(), arg1.getX(), arg1.getY());
				float dis2 = (float) SOSGeometryTools.distance(p.first()
						.getCenterOfZone().getX(), p.first().getCenterOfZone()
						.getY(), arg2.getX(), arg2.getY());

				if (dis1 < dis2)
					return -1;
				else
					return 1;
			}
		});

		int index = allclusters.indexOf(pf.model().searchWorldModel
				.getClusterData());
		pf.log.info("index " + index);

		ArrayList<ClusterData> removes = new ArrayList<ClusterData>();
		for (int i = 0; i < allclusters.size(); i++) {
			if (p.second() == 0)
				if (i > need0)
					break;
			if (p.second() == 1)
				if (i > need1)
					break;
			if (p.second() == 2)
				if (i > need2)
					break;
			removes.add(allclusters.get(i));
		}

		all.removeAll(removes);
		return index;
	}

	public FireZoneManager(SOSWorldModel model, PoliceForceAgent pf) {
		this.pf = pf;
		this.fireBrigadeWorldModel = model;
		fireZones = new Vector<FireZone>();
		updateFireZone();

		range = Point.distance(0, 0, fireBrigadeWorldModel.getBounds()
				.getWidth(), fireBrigadeWorldModel.getBounds().getHeight()) / 2;
	}

	private void megeFirezone(FireZone f1, FireZone f2) {
		for (int i = f2.getBuilding().size() - 1; i > -1; i--) {
			f1.addBuild(f2.getBuilding().elementAt(i));
			f2.getBuilding().elementAt(i).setFireZone(f1);
			f2.getBuilding().remove(i);
		}

	}

	private void mergeFireZoneCheck() {
		for (FireZone f1 : fireZones)
			for (FireZone f2 : fireZones) {
				pf.log.info("hey merge!");
				if (!f1.equals(f2)) {
					pf.log.info("hey merge2!");
					if (f1.getBorder().size() > 0 && f2.getBorder().size() > 0) {
						pf.log.info("hey merge3!");
						double distance;
						double temp;
						distance = Point.distance(f1.getBuilding()
								.firstElement().getX(), f1.getBuilding()
								.firstElement().getY(), f2.getBuilding()
								.firstElement().getX(), f2.getBuilding()
								.firstElement().getY());
						for (Building b1 : f1.getBorder()) {
							for (Building b2 : f2.getBorder()) {

								temp = Point.distance(b1.getX(), b1.getY(),
										b2.getX(), b2.getY());
								if (temp < distance) {
									distance = temp;
								}

							}
						}
						// AZ HAMIN RANGE DAR ADDTOZONE ESTEFADE SHODE
						pf.log.info(" distance =" + distance);
						//	pf.log.warn("f1"+f1.getCenterOfZone() + "f2" +f2.getCenterOfZone());
						if (distance < range / 5) {
							pf.log.info("will merge!");
							megeFirezone(f1, f2);
						}

					}
				}
			}
		for (int i = fireZones.size() - 1; i > -1; i--) {
			if (fireZones.elementAt(i).getSize() == 0) {
				fireZones.remove(i);
			}
		}
	}

	public Shape firezoneShapeScale(float scalepoint) {
		java.awt.geom.Area area = new java.awt.geom.Area();
		for (FireZone zone : fireZones) {
			ConvexHull_arr_New temp = new ConvexHull_arr_New(zone.getBuilding())
					.getScaleConvex(scalepoint);
			area.add(new Area(temp.getShape()));

		}
		return area;

	}

	public Shape activeFirezoneShapeScale(float scalepoint) {
		java.awt.geom.Area area = new java.awt.geom.Area();
		for (FireZone zone : fireZones) {
			if (zone.isActive()) {
				ConvexHull_arr_New temp = new ConvexHull_arr_New(
						zone.getBuilding()).getScaleConvex(scalepoint);
				area.add(new Area(temp.getShape()));

			}

		}
		return area;

	}

	public void updateFireZone() {

		ArrayList<Building> buildings = new ArrayList<Building>();

		buildings.addAll(policeWorldModel.fieryBuildings());

		for (FireZone fz : fireZones) {
			buildings.removeAll(fz.getBuilding());
		}
		for (Building building : buildings) {
			if (!isManagedBefore(building)) {
				addToZone(building);
			}
			else {
				pf.log.info("building = " + building);
			}
			pf.log.info("fsize = " + fireZones.size());
		}
		mergeFireZoneCheck();
	}

	private void updateFators(int z, int o, int t) {

		int onfireClusterSize = policeWorldModel.policeForces().size() - notInFireClusters.size();

		int m = policeWorldModel.policeForces().size();
		closedRefuges = 0;
		int StuckAgents = 3;

		if (policeWorldModel.getLastAfterShockTime() > policeWorldModel.time() - 15)
			pf.setLastShockTime(policeWorldModel.getLastAfterShockTime());

		if (pf.getLastShockTime() != 0 || policeWorldModel.time() < 15)
			closedRefuges = policeWorldModel.refuges().size();

		if (policeWorldModel.time() > 20)
			StuckAgents = 5;

		double s = (m - closedRefuges - onfireClusterSize - StuckAgents) / ((2 * z) + (6 * o) + (12 * t) + (policeWorldModel.ambulanceTeams().size() * 0.4));

		clusterCivilians.clear();
		UpdateClusterCivilian();
		ArrayList<Civilian> unhealthyCivilians = new ArrayList<Civilian>();
		for (Civilian civ : policeWorldModel.civilians()) {
			if (civ == null)
				continue;
//			if (civ.getBuriedness() == 0)
//				continue;
			if (civ.getDamage() == 0)
				continue;
			else {
				unhealthyCivilians.add(civ);
			}
		}

		double s1 = s * 0.8 * policeWorldModel.ambulanceTeams().size() * policeWorldModel.ambulanceTeams().size() * clusterCivilians.size() / unhealthyCivilians.size();

		if (s1 < 1)
			s1 = 1;
		else
			s1 = (int) Math.floor(s1);
		//		for(int i = 0; i < s1; i++){
		//			// mikham in ja begam k ye civil e be sarfe va kone :D
		//			
		//			}

		need0 = (int) Math.floor(2 * s);

		if (need0 < 1)
			need0 = 1;
		need1 = (int) Math.floor((6 * s));
		need2 = (int) Math.floor((12 * s));

		if (policeWorldModel.time() < 150)
		{
			need0 = Math.min(need0, 2);
			need1 = Math.min(need1, 6);
			need2 = Math.min(need2, 10);
		}
		need0 = Math.max(need0, 0);
		need1 = Math.max(need1, 0);
		need2 = Math.max(need2, 0);

	}

	private void UpdateClusterCivilian() {
		for (Civilian civ : policeWorldModel.civilians()) {
			if (policeWorldModel.searchWorldModel
					.getClusterData().getBuildings().contains(civ.getAreaPosition()))
				clusterCivilians.add(civ);
		}
	}

	private boolean isManagedBefore(Building building) {
		if (building.getFireZone() == null)
			return false;
		else
			return true;
	}

	private void addToZone(Building building) {
		building.setFirezonecheck(true);
		for (Building near : building.realNeighbors_Building()) {
			if (near.getFieryness() > 0 && near.getFieryness() < 4)
				if (near.getFireZone() == null) {
					if (!near.isFirezonecheck())
						addToZone(near);
				}
			if (building.getFireZone() == null)
				if (near.getFireZone() != null) {
					near.getFireZone().addBuild(building);
					building.setFireZone(near.getFireZone());
					findCluster(building);

				}

		}
		for (Building near : policeWorldModel.getObjectsInRange(building,
				(int) range / 5, Building.class)) {
			if (near.getFireZone() != null) {
				near.getFireZone().addBuild(building);
				building.setFireZone(near.getFireZone());
				findCluster(building);
			}
		}
		if (building.getFireZone() == null) {
			FireZone temp = new FireZone(policeWorldModel);
			temp.addBuild(building);
			building.setFireZone(temp);
			findCluster(building);
			fireZones.add(temp);
		}
	}

	private void findCluster(Building building) {
		ArrayList<ClusterData> removes = new ArrayList<ClusterData>();
		for (ClusterData c : notInFireClusters) {
			if (c.getBuildings().contains(building))
				removes.add(c);
		}
		notInFireClusters.removeAll(removes);
	}

	public Vector<Road> makePoliceTaskZone() {
		Vector<Road> mustclear = new Vector<Road>();
		for (FireZone fireZone : fireZones) {
			for (Road road : fireZone.getBestToReachable(Policezone)) {
				mustclear.add(road);
			}
		}
		return mustclear;
	}

	public Vector<Road> makePoliceTaskZone(FireZone best) {
		Vector<Road> mustclear = new Vector<Road>();
		for (Road road : best.getBestToReachable(Policezone)) {
			mustclear.add(road);
		}
		return mustclear;
	}

	public ArrayList<Building> getAllFireZoneBorder() {
		ArrayList<Building> result = new ArrayList<Building>();
		// int minDis = Integer.MAX_VALUE;
		for (FireZone z : fireZones)
			result.addAll(z.getBorder());
		return result;
	}

	public Vector<FireZone> getFireZones() {
		return fireZones;
	}
}
